/*
 * Copyright (C) 2013 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#include "config.h"
#include "DFGStackLayoutPhase.h"

#if ENABLE(DFG_JIT)

#include "DFGGraph.h"
#include "DFGPhase.h"
#include "DFGValueSource.h"
#include "Operations.h"

namespace JSC { namespace DFG {

class StackLayoutPhase : public Phase {
    static const bool verbose = false;
    
public:
    StackLayoutPhase(Graph& graph)
        : Phase(graph, "stack layout")
    {
    }
    
    bool run()
    {
        SharedSymbolTable* symbolTable = codeBlock()->symbolTable();

        // This enumerates the locals that we actually care about and packs them. So for example
        // if we use local 1, 3, 4, 5, 7, then we remap them: 1->0, 3->1, 4->2, 5->3, 7->4. We
        // treat a variable as being "used" if there exists an access to it (SetLocal, GetLocal,
        // Flush, PhantomLocal).
        
        BitVector usedLocals;
        
        // Collect those variables that are used from IR.
        bool hasGetLocalUnlinked = false;
        for (BlockIndex blockIndex = m_graph.numBlocks(); blockIndex--;) {
            BasicBlock* block = m_graph.block(blockIndex);
            if (!block)
                continue;
            for (unsigned nodeIndex = block->size(); nodeIndex--;) {
                Node* node = block->at(nodeIndex);
                switch (node->op()) {
                case GetLocal:
                case SetLocal:
                case Flush:
                case PhantomLocal: {
                    VariableAccessData* variable = node->variableAccessData();
                    if (operandIsArgument(variable->local()))
                        break;
                    usedLocals.set(operandToLocal(variable->local()));
                    break;
                }
                    
                case GetLocalUnlinked: {
                    VirtualRegister operand = node->unlinkedLocal();
                    if (operandIsArgument(operand))
                        break;
                    usedLocals.set(operandToLocal(operand));
                    hasGetLocalUnlinked = true;
                    break;
                }
                    
                default:
                    break;
                }
            }
        }
        
        // Ensure that captured variables and captured inline arguments are pinned down.
        // They should have been because of flushes, except that the flushes can be optimized
        // away.
        if (symbolTable) {
            for (int i = symbolTable->captureStart(); i > symbolTable->captureEnd(); i--)
                usedLocals.set(operandToLocal(i));
        }
        if (codeBlock()->usesArguments()) {
            usedLocals.set(operandToLocal(codeBlock()->argumentsRegister()));
            usedLocals.set(operandToLocal(unmodifiedArgumentsRegister(codeBlock()->argumentsRegister())));
        }
        if (codeBlock()->uncheckedActivationRegister() != InvalidVirtualRegister)
            usedLocals.set(operandToLocal(codeBlock()->activationRegister()));
        for (InlineCallFrameSet::iterator iter = m_graph.m_inlineCallFrames->begin(); !!iter; ++iter) {
			InlineCallFrame* inlineCallFrame = *iter;
            if (!m_graph.usesArguments(inlineCallFrame))
                continue;
            
            VirtualRegister argumentsRegister = m_graph.argumentsRegisterFor(inlineCallFrame);
            usedLocals.set(operandToLocal(argumentsRegister));
            usedLocals.set(operandToLocal(unmodifiedArgumentsRegister(argumentsRegister)));
            
            for (unsigned argument = inlineCallFrame->arguments.size(); argument-- > 1;) {
                usedLocals.set(operandToLocal((
                    virtualRegisterForArgument(argument) +
                    inlineCallFrame->stackOffset)));
            }
        }
        
        Vector<unsigned> allocation(usedLocals.size());
        m_graph.m_nextMachineLocal = 0;
        for (unsigned i = 0; i < usedLocals.size(); ++i) {
            if (!usedLocals.get(i)) {
                allocation[i] = UINT_MAX;
                continue;
            }
            
            allocation[i] = m_graph.m_nextMachineLocal++;
        }
        
        for (unsigned i = m_graph.m_variableAccessData.size(); i--;) {
            VariableAccessData* variable = &m_graph.m_variableAccessData[i];
            if (!variable->isRoot())
                continue;
            
            if (operandIsArgument(variable->local())) {
                variable->machineLocal() = variable->local();
                continue;
            }
            
            size_t local = operandToLocal(variable->local());
            if (local >= allocation.size())
                continue;
            
            if (allocation[local] == UINT_MAX)
                continue;
            
            variable->machineLocal() = static_cast<VirtualRegister>(virtualRegisterForLocal(
                allocation[operandToLocal(variable->local())]));
        }
        
        if (codeBlock()->usesArguments()) {
            VirtualRegister argumentsRegister = static_cast<VirtualRegister>(virtualRegisterForLocal(
                allocation[operandToLocal(codeBlock()->argumentsRegister())]));
            RELEASE_ASSERT(
                virtualRegisterForLocal(allocation[
                    unmodifiedArgumentsRegister(
                        codeBlock()->argumentsRegister()).toLocal()])
                == unmodifiedArgumentsRegister(argumentsRegister));
            codeBlock()->setArgumentsRegister(argumentsRegister);
        }
        
        if (codeBlock()->uncheckedActivationRegister() != InvalidVirtualRegister) {
            codeBlock()->setActivationRegister(
                virtualRegisterForLocal(allocation[operandToLocal(codeBlock()->activationRegister())]));
        }

        for (unsigned i = m_graph.m_inlineVariableData.size(); i--;) {
            InlineVariableData data = m_graph.m_inlineVariableData[i];
            InlineCallFrame* inlineCallFrame = data.inlineCallFrame;

            if (m_graph.usesArguments(inlineCallFrame)) {
				inlineCallFrame->argumentsRegister = static_cast<VirtualRegister>(virtualRegisterForLocal(
					allocation[operandToLocal(m_graph.argumentsRegisterFor(inlineCallFrame))]));

				RELEASE_ASSERT(
					virtualRegisterForLocal(allocation[unmodifiedArgumentsRegister(
						m_graph.argumentsRegisterFor(inlineCallFrame)).toLocal()])
					== unmodifiedArgumentsRegister(inlineCallFrame->argumentsRegister));
			}
            
			for (unsigned argument = inlineCallFrame->arguments.size(); argument-- > 1;) {
				ArgumentPosition& position = m_graph.m_argumentPositions[
					data.argumentPositionStart + argument];
				VariableAccessData* variable = position.someVariable();
				ValueSource source;
				if (!variable)
					source = ValueSource(SourceIsDead);
				else {
					source = ValueSource::forFlushFormat(
						variable->machineLocal(), variable->flushFormat());
				}
				inlineCallFrame->arguments[argument] = source.valueRecovery();
			}

			RELEASE_ASSERT(inlineCallFrame->isClosureCall == !!data.calleeVariable);

			if (inlineCallFrame->isClosureCall) {
                VariableAccessData* variable = data.calleeVariable->find();
				ValueSource source = ValueSource::forFlushFormat(
					variable->machineLocal(),
					variable->flushFormat());
				inlineCallFrame->calleeRecovery = source.valueRecovery();
			} else
				RELEASE_ASSERT(inlineCallFrame->calleeRecovery.isConstant());
		}
        
        if (symbolTable) {
            if (symbolTable->captureCount()) {
                unsigned captureStartLocal = allocation[
					operandToLocal(codeBlock()->symbolTable()->captureStart())];
                ASSERT(captureStartLocal != UINT_MAX);
                m_graph.m_machineCaptureStart = virtualRegisterForLocal(captureStartLocal);
            } else
                m_graph.m_machineCaptureStart = virtualRegisterForLocal(0);
        
            // This is an abomination. If we had captured an argument then the argument ends
            // up being "slow", meaning that loads of the argument go through an extra lookup
            // table.
            if (const SlowArgument* slowArguments = symbolTable->slowArguments()) {
            	OwnArrayPtr<SlowArgument> newSlowArguments = adoptArrayPtr(new SlowArgument[symbolTable->parameterCount()]);
                for (size_t i = symbolTable->parameterCount(); i--;) {
                    newSlowArguments[i] = slowArguments[i];
                    VirtualRegister reg = VirtualRegister(slowArguments[i].index);
                    if (operandIsLocal(reg))
                        newSlowArguments[i].index = virtualRegisterForLocal(allocation[operandToLocal(reg)]);
                }
            
                m_graph.m_slowArguments = newSlowArguments.release();
            }
        }

        // Fix GetLocalUnlinked's variable references.
        if (hasGetLocalUnlinked) {
            for (BlockIndex blockIndex = m_graph.numBlocks(); blockIndex--;) {
                BasicBlock* block = m_graph.block(blockIndex);
                if (!block)
                    continue;
                for (unsigned nodeIndex = block->size(); nodeIndex--;) {
                    Node* node = block->at(nodeIndex);
                    switch (node->op()) {
                    case GetLocalUnlinked: {
                        VirtualRegister operand = node->unlinkedLocal();
                        if (operandIsLocal(operand))
                            operand = static_cast<VirtualRegister>(virtualRegisterForLocal(allocation[operandToLocal(operand)]));
                        node->setUnlinkedMachineLocal(operand);
                        break;
                    }
                        
                    default:
                        break;
                    }
                }
            }
        }
        
        return true;
    }
};

bool performStackLayout(Graph& graph)
{
    SamplingRegion samplingRegion("DFG Stack Layout Phase");
    return runPhase<StackLayoutPhase>(graph);
}

} } // namespace JSC::DFG

#endif // ENABLE(DFG_JIT)

