/*
 * Copyright (C) 2009, 2010 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#include "config.h"
#include "Executable.h"

#include "BytecodeGenerator.h"
#include "CodeBlock.h"
#include "DFGDriver.h"
#include "DeferGC.h"
#include "JIT.h"
#include "Operations.h"
#include "Parser.h"
#include <wtf/CurrentTime.h>
#include <wtf/Vector.h>
#include <wtf/text/StringBuilder.h>

namespace JSC {

void ScriptExecutable::installCode(CodeBlock* genericCodeBlock)
{
    RELEASE_ASSERT(genericCodeBlock->ownerExecutable() == this);

    JSGlobalData& vm = *genericCodeBlock->vm();

    ASSERT(vm.heap.isDeferred());

    CodeSpecializationKind kind = genericCodeBlock->specializationKind();

    RefPtr<CodeBlock> oldCodeBlock;

    switch (kind) {
    case CodeForCall:
        m_jitCodeForCall = genericCodeBlock->jitCode();
        m_jitCodeForCallWithArityCheck = MacroAssemblerCodePtr();
        m_jitCodeForCallWithArityCheckAndPreserveRegs = MacroAssemblerCodePtr();
        m_numParametersForCall = genericCodeBlock->numParameters();
        break;
    case CodeForConstruct:
        m_jitCodeForConstruct = genericCodeBlock->jitCode();
        m_jitCodeForConstructWithArityCheck = MacroAssemblerCodePtr();
        m_jitCodeForConstructWithArityCheckAndPreserveRegs = MacroAssemblerCodePtr();
        m_numParametersForConstruct = genericCodeBlock->numParameters();
        break;
    }

    switch (genericCodeBlock->codeType()) {
    case GlobalCode: {
        ProgramExecutable* executable = jsCast<ProgramExecutable*>(this);
        ProgramCodeBlock* codeBlock = static_cast<ProgramCodeBlock*>(genericCodeBlock);

        ASSERT(kind == CodeForCall);

        oldCodeBlock = executable->m_programCodeBlock;
        executable->m_programCodeBlock = codeBlock;
        break;
    }

    case EvalCode: {
        EvalExecutable* executable = jsCast<EvalExecutable*>(this);
        EvalCodeBlock* codeBlock = static_cast<EvalCodeBlock*>(genericCodeBlock);

        ASSERT(kind == CodeForCall);

        oldCodeBlock = executable->m_evalCodeBlock;
        executable->m_evalCodeBlock = codeBlock;
        break;
    }

    case FunctionCode: {
        FunctionExecutable* executable = jsCast<FunctionExecutable*>(this);
        FunctionCodeBlock* codeBlock = static_cast<FunctionCodeBlock*>(genericCodeBlock);

        switch (kind) {
        case CodeForCall:
            oldCodeBlock = executable->m_codeBlockForCall;
            executable->m_codeBlockForCall = codeBlock;
            break;
        case CodeForConstruct:
            oldCodeBlock = executable->m_codeBlockForConstruct;
            executable->m_codeBlockForConstruct = codeBlock;
            break;
        }
        break;
    }
    }

    if (oldCodeBlock)
        oldCodeBlock->unlinkIncomingCalls();

    genericCodeBlock->setInstallTime(WTF::monotonicallyIncreasingTime());

    vm.heap.didInstallCode(this);
}

PassRefPtr<CodeBlock> ScriptExecutable::newCodeBlockFor(
	ExecState* exec, CodeSpecializationKind kind, JSScope* scope, JSObject*& exception)
{
    VM* vm = scope->vm();
    JSGlobalObject* lexicalGlobalObject = exec->lexicalGlobalObject();

    ASSERT(vm->heap.isDeferred());

    if (classInfo() == EvalExecutable::info()) {
        EvalExecutable* executable = jsCast<EvalExecutable*>(this);
        RELEASE_ASSERT(kind == CodeForCall);
        RELEASE_ASSERT(!executable->m_evalCodeBlock);

        if (!lexicalGlobalObject->evalEnabled()) {
        	exception = vm->throwException(exec, createEvalError(exec, "Eval is disabled"));
        	return 0;
        }
        ParserError error;
        RefPtr<EvalNode> evalNode = parse<EvalNode>(vm, m_source, 0, Identifier(), isStrictMode() ? JSParseStrict : JSParseNormal, EvalNode::isFunctionNode ? JSParseFunctionCode : JSParseProgramCode, error);

        if (lexicalGlobalObject->hasDebugger() && !EvalNode::scopeIsFunction)
        	lexicalGlobalObject->debugger()->sourceParsed(exec, m_source.provider(), error.line(), error.message());

        if (!evalNode) {
            exception = error.toErrorObject(lexicalGlobalObject, m_source);
            return 0;
        }

        recordParse(evalNode->features(), evalNode->hasCapturedVariables(), evalNode->lineNo(), evalNode->lastLine(), startColumn());

        RefPtr<EvalCodeBlock> evalCodeBlock = adoptRef(new EvalCodeBlock(executable, scope, source().provider()));
        OwnPtr<BytecodeGenerator> generator(adoptPtr(new BytecodeGenerator(evalNode.get(), scope, evalCodeBlock->symbolTable(), evalCodeBlock.get())));

        error = generator->generate(exec);
        evalNode->destroyData();
        if (error.isValid()) {
            exception = error.toErrorObject(lexicalGlobalObject, m_source);
        	return 0;
        }

        return evalCodeBlock.release();
    }

    if (classInfo() == ProgramExecutable::info()) {
        ProgramExecutable* executable = jsCast<ProgramExecutable*>(this);
        RELEASE_ASSERT(kind == CodeForCall);
        RELEASE_ASSERT(!executable->m_programCodeBlock);

        ParserError error;
        RefPtr<ProgramNode> programNode = parse<ProgramNode>(vm, m_source, 0, Identifier(), isStrictMode() ? JSParseStrict : JSParseNormal, ProgramNode::isFunctionNode ? JSParseFunctionCode : JSParseProgramCode, error);

        if (lexicalGlobalObject->hasDebugger() && !ProgramNode::scopeIsFunction)
        	lexicalGlobalObject->debugger()->sourceParsed(exec, m_source.provider(), error.line(), error.message());

        if (!programNode) {
            exception = error.toErrorObject(lexicalGlobalObject, m_source);
            return 0;
        }
        recordParse(programNode->features(), programNode->hasCapturedVariables(), programNode->lineNo(), programNode->lastLine(), startColumn());

        JSGlobalObject* globalObject = scope->globalObject();

        RefPtr<ProgramCodeBlock> programCodeBlock = adoptRef(new ProgramCodeBlock(executable, GlobalCode, scope, source().provider(), source().startColumn()));
        OwnPtr<BytecodeGenerator> generator(adoptPtr(new BytecodeGenerator(programNode.get(), scope, globalObject->symbolTable(), programCodeBlock.get())));

        error = generator->generate(exec);
        programNode->destroyData();
        if (error.isValid()) {
            exception = error.toErrorObject(lexicalGlobalObject, m_source);
        	return 0;
        }

        return programCodeBlock.release();
    }

    RELEASE_ASSERT(classInfo() == FunctionExecutable::info());
    FunctionExecutable* executable = jsCast<FunctionExecutable*>(this);
    RELEASE_ASSERT(!executable->codeBlockFor(kind));

    exception = 0;
    ParserError error;
    JSGlobalObject* globalObject = scope->globalObject();
    RefPtr<FunctionBodyNode> body = parse<FunctionBodyNode>(
        vm,
        m_source,
		executable->m_parameters.get(),
		executable->name(),
		executable->toStrictness(),
        FunctionBodyNode::isFunctionNode ? JSParseFunctionCode : JSParseProgramCode,
        error
    );

    if (lexicalGlobalObject->hasDebugger() && !FunctionBodyNode::scopeIsFunction)
    	lexicalGlobalObject->debugger()->sourceParsed(exec, m_source.provider(), error.line(), error.message());

    if (!body) {
        exception = error.toErrorObject(lexicalGlobalObject, m_source);
        return 0;
    }
    if (executable->m_forceUsesArguments)
        body->setUsesArguments();
    body->finishParsing(executable->m_parameters, executable->m_name, executable->m_functionNameIsInScopeToggle);
    recordParse(body->features(), body->hasCapturedVariables(), body->lineNo(), body->lastLine(), startColumn());

    SourceProvider* provider = source().provider();
    unsigned sourceOffset = source().startOffset();
    unsigned startColumn = source().startColumn();

    RefPtr<FunctionCodeBlock> functionCodeBlock = adoptRef(new FunctionCodeBlock(executable, FunctionCode, scope, provider, sourceOffset, startColumn, kind == CodeForConstruct));
   	OwnPtr<BytecodeGenerator> generator(adoptPtr(new BytecodeGenerator(body.get(), scope, functionCodeBlock->symbolTable(), functionCodeBlock.get())));
    error = generator->generate(exec);
	body->destroyData();
    if (error.isValid()) {
        exception = error.toErrorObject(lexicalGlobalObject, m_source);
		return 0;
    }

	return functionCodeBlock.release();
}

PassRefPtr<CodeBlock> ScriptExecutable::newReplacementCodeBlockFor(
    CodeSpecializationKind kind)
{
    if (classInfo() == EvalExecutable::info()) {
        RELEASE_ASSERT(kind == CodeForCall);
        EvalExecutable* executable = jsCast<EvalExecutable*>(this);
        RefPtr<EvalCodeBlock> result = adoptRef(new EvalCodeBlock(
            CodeBlock::CopyParsedBlock, *executable->m_evalCodeBlock));
        result->setAlternative(executable->m_evalCodeBlock);
        return result;
    }

    if (classInfo() == ProgramExecutable::info()) {
        RELEASE_ASSERT(kind == CodeForCall);
        ProgramExecutable* executable = jsCast<ProgramExecutable*>(this);
        RefPtr<ProgramCodeBlock> result = adoptRef(new ProgramCodeBlock(
            CodeBlock::CopyParsedBlock, *executable->m_programCodeBlock));
        result->setAlternative(executable->m_programCodeBlock);
        return result;
    }

    RELEASE_ASSERT(classInfo() == FunctionExecutable::info());
    FunctionExecutable* executable = jsCast<FunctionExecutable*>(this);
    RefPtr<FunctionCodeBlock> result = adoptRef(new FunctionCodeBlock(
        CodeBlock::CopyParsedBlock, *executable->codeBlockFor(kind)));
    result->setAlternative(executable->codeBlockFor(kind));
    return result;
}

JSObject* ScriptExecutable::prepareForExecutionImpl(ExecState* exec, JSScope* scope, CodeSpecializationKind kind, CodeBlock*& resultCodeBlock)
{
    VM& vm = exec->vm();
    DeferGCForAWhile deferGC(vm.heap);

    JSObject* exception = 0;
    RefPtr<CodeBlock> codeBlock = newCodeBlockFor(exec, kind, scope, exception);
    resultCodeBlock = codeBlock.get();
    if (!codeBlock) {
        RELEASE_ASSERT(exception);
        if(!exception)
        	exception = createOutOfMemoryError(exec);
        return exception;
    }

#if ENABLE(LLINT)
    LLInt::setEntrypoint(vm, codeBlock.get());
#else
    CompilationResult result = JIT::compile(&vm, codeBlock.get(), JITCompilationMustSucceed);
    RELEASE_ASSERT(result == CompilationSuccessful);
#endif

    if(result == CompilationSuccessful) {
    	installCode(codeBlock.get());
    	return 0;
    }

    return vm.throwException(exec, createOutOfMemoryError(exec));
}

const ClassInfo ExecutableBase::s_info = { "Executable", 0, 0, CREATE_METHOD_TABLE(ExecutableBase) };

void ExecutableBase::destroy(JSCell* cell)
{
    static_cast<ExecutableBase*>(cell)->ExecutableBase::~ExecutableBase();
}

void ExecutableBase::clearCode()
{
#if ENABLE(JIT)
    m_jitCodeForCall.clear();
    m_jitCodeForConstruct.clear();
    m_jitCodeForCallWithArityCheck = MacroAssemblerCodePtr();
    m_jitCodeForConstructWithArityCheck = MacroAssemblerCodePtr();
#endif
    m_numParametersForCall = NUM_PARAMETERS_NOT_COMPILED;
    m_numParametersForConstruct = NUM_PARAMETERS_NOT_COMPILED;
}

#if ENABLE(DFG_JIT)
Intrinsic ExecutableBase::intrinsic() const
{
    if (const NativeExecutable* nativeExecutable = jsDynamicCast<const NativeExecutable*>(this))
        return nativeExecutable->intrinsic();
    return NoIntrinsic;
}
#else
Intrinsic ExecutableBase::intrinsic() const
{
    return NoIntrinsic;
}
#endif

const ClassInfo NativeExecutable::s_info = { "NativeExecutable", &ExecutableBase::s_info, 0, CREATE_METHOD_TABLE(NativeExecutable) };

void NativeExecutable::destroy(JSCell* cell)
{
    static_cast<NativeExecutable*>(cell)->NativeExecutable::~NativeExecutable();
}

#if ENABLE(DFG_JIT)
Intrinsic NativeExecutable::intrinsic() const
{
    return m_intrinsic;
}
#endif

const ClassInfo ScriptExecutable::s_info = { "ScriptExecutable", &ExecutableBase::s_info, 0, CREATE_METHOD_TABLE(ScriptExecutable) };

ScriptExecutable::ScriptExecutable(Structure* structure, JSGlobalData& globalData, const SourceCode& source, bool isInStrictContext)
: ExecutableBase(globalData, structure, NUM_PARAMETERS_NOT_COMPILED)
, m_source(source)
, m_features(isInStrictContext ? StrictModeFeature : 0)
, m_hasCapturedVariables(false)
, m_neverInline(false)
, m_didTryToEnterInLoop(false)
, m_firstLine(-1)
, m_lastLine(-1)
{
}

void ScriptExecutable::destroy(JSCell* cell)
{
    static_cast<ScriptExecutable*>(cell)->ScriptExecutable::~ScriptExecutable();
}

const ClassInfo EvalExecutable::s_info = { "EvalExecutable", &ScriptExecutable::s_info, 0, CREATE_METHOD_TABLE(EvalExecutable) };

EvalExecutable::EvalExecutable(ExecState* exec, const SourceCode& source, bool inStrictContext)
    : ScriptExecutable(exec->vm().evalExecutableStructure.get(), exec->vm(), source, inStrictContext)
{
}

void EvalExecutable::destroy(JSCell* cell)
{
    static_cast<EvalExecutable*>(cell)->EvalExecutable::~EvalExecutable();
}

unsigned EvalExecutable::numVariables()
{
	return m_evalCodeBlock->numVariables();
}

unsigned EvalExecutable::numberOfFunctionDecls()
{
	return m_evalCodeBlock->numberOfFunctionDecls();
}

const ClassInfo ProgramExecutable::s_info = { "ProgramExecutable", &ScriptExecutable::s_info, 0, CREATE_METHOD_TABLE(ProgramExecutable) };

ProgramExecutable::ProgramExecutable(ExecState* exec, const SourceCode& source)
    : ScriptExecutable(exec->vm().programExecutableStructure.get(), exec->vm(), source, false)
{
}

void ProgramExecutable::destroy(JSCell* cell)
{
    static_cast<ProgramExecutable*>(cell)->ProgramExecutable::~ProgramExecutable();
}

const ClassInfo FunctionExecutable::s_info = { "FunctionExecutable", &ScriptExecutable::s_info, 0, CREATE_METHOD_TABLE(FunctionExecutable) };


FunctionExecutable::FunctionExecutable(JSGlobalData& globalData, FunctionBodyNode* node, unsigned startColumn, bool isBuiltinFunction)
    : ScriptExecutable(globalData.functionExecutableStructure.get(), globalData, node->source(), node->isStrictMode())
    , m_forceUsesArguments(node->usesArguments())
    , m_parameters(node->parameters())
, m_isBuiltinFunction(isBuiltinFunction)
    , m_name(node->ident())
    , m_inferredName(node->inferredName().isNull() ? globalData.propertyNames->emptyIdentifier.string() : node->inferredName().string())
    , m_functionNameIsInScopeToggle(node->functionNameIsInScopeToggle())
{
    m_firstLine = node->lineNo();
    m_lastLine = node->lastLine();
    m_startColumn = startColumn;
}

void FunctionExecutable::destroy(JSCell* cell)
{
    static_cast<FunctionExecutable*>(cell)->FunctionExecutable::~FunctionExecutable();
}

inline const char* samplingDescription(JITCode::JITType jitType)
{
    switch (jitType) {
    case JITCode::InterpreterThunk:
        return "Interpreter Compilation (TOTAL)";
    case JITCode::BaselineJIT:
        return "Baseline Compilation (TOTAL)";
    case JITCode::DFGJIT:
        return "DFG Compilation (TOTAL)";
    case JITCode::FTLJIT:
        return "FTL Compilation (TOTAL)";
    default:
        ASSERT_NOT_REACHED();
        return 0;
    }
}

void EvalExecutable::visitChildren(JSCell* cell, SlotVisitor& visitor)
{
    EvalExecutable* thisObject = jsCast<EvalExecutable*>(cell);
    ASSERT_GC_OBJECT_INHERITS(thisObject, &s_info);
    Base::visitChildren(thisObject, visitor);
    if (thisObject->m_evalCodeBlock)
        thisObject->m_evalCodeBlock->visitAggregate(visitor);
}

void EvalExecutable::unlinkCalls()
{
#if ENABLE(JIT)
    if (!m_jitCodeForCall)
        return;
    ASSERT(m_evalCodeBlock);
    m_evalCodeBlock->unlinkCalls();
#endif
}

void EvalExecutable::clearCode()
{
    m_evalCodeBlock.clear();
    Base::clearCode();
}

JSObject* ProgramExecutable::checkSyntax(ExecState* exec)
{
    ParserError error;
    JSGlobalData* globalData = &exec->globalData();
    JSGlobalObject* lexicalGlobalObject = exec->lexicalGlobalObject();
    RefPtr<ProgramNode> programNode = parse<ProgramNode>(globalData, m_source, 0, Identifier(), JSParseNormal, ProgramNode::isFunctionNode ? JSParseFunctionCode : JSParseProgramCode, error);

    if (programNode)
        return 0;
    ASSERT(error.m_type != ParserError::ErrorNone);
    return error.toErrorObject(lexicalGlobalObject, m_source);
}

void ProgramExecutable::unlinkCalls()
{
    if (!m_jitCodeForCall)
        return;
    ASSERT(m_programCodeBlock);
    m_programCodeBlock->unlinkCalls();
}

void ProgramExecutable::visitChildren(JSCell* cell, SlotVisitor& visitor)
{
    ProgramExecutable* thisObject = jsCast<ProgramExecutable*>(cell);
    ASSERT_GC_OBJECT_INHERITS(thisObject, &s_info);
    Base::visitChildren(thisObject, visitor);
    if (thisObject->m_programCodeBlock)
        thisObject->m_programCodeBlock->visitAggregate(visitor);
}

void ProgramExecutable::clearCode()
{
    m_programCodeBlock.clear();
    Base::clearCode();
}

FunctionCodeBlock* FunctionExecutable::baselineCodeBlockFor(CodeSpecializationKind kind)
{
    FunctionCodeBlock* result;
    if (kind == CodeForCall)
        result = m_codeBlockForCall.get();
    else {
        ASSERT(kind == CodeForConstruct);
        result = m_codeBlockForConstruct.get();
    }
    if (!result)
        return 0;
    return static_cast<FunctionCodeBlock*>(result->baselineAlternative());
}

void FunctionExecutable::visitChildren(JSCell* cell, SlotVisitor& visitor)
{
    FunctionExecutable* thisObject = jsCast<FunctionExecutable*>(cell);
    ASSERT_GC_OBJECT_INHERITS(thisObject, &s_info);
    Base::visitChildren(thisObject, visitor);
    visitor.append(&thisObject->m_nameValue);
    if (thisObject->m_codeBlockForCall)
        thisObject->m_codeBlockForCall->visitAggregate(visitor);
    if (thisObject->m_codeBlockForConstruct)
        thisObject->m_codeBlockForConstruct->visitAggregate(visitor);
}

void FunctionExecutable::clearCode()
{
    m_codeBlockForCall.clear();
    m_codeBlockForConstruct.clear();
    Base::clearCode();
}

void FunctionExecutable::unlinkCalls()
{
#if ENABLE(JIT)
    if (!!m_jitCodeForCall) {
        ASSERT(m_codeBlockForCall);
        m_codeBlockForCall->unlinkCalls();
    }
    if (!!m_jitCodeForConstruct) {
        ASSERT(m_codeBlockForConstruct);
        m_codeBlockForConstruct->unlinkCalls();
    }
#endif
}

FunctionExecutable* FunctionExecutable::fromGlobalCode(const Identifier& name, ExecState* exec, Debugger* debugger, const SourceCode& source, JSObject** exception)
{
	JSGlobalData& globalData = exec->globalData();
    JSGlobalObject* lexicalGlobalObject = exec->lexicalGlobalObject();
    ParserError error;
    RefPtr<ProgramNode> program = parse<ProgramNode>(&globalData, source, 0, Identifier(), JSParseNormal, ProgramNode::isFunctionNode ? JSParseFunctionCode : JSParseProgramCode, error);

    if (lexicalGlobalObject->hasDebugger() && !ProgramNode::scopeIsFunction)
    	lexicalGlobalObject->debugger()->sourceParsed(exec, source.provider(), error.line(), error.message());

    if (!program) {
        *exception = error.toErrorObject(lexicalGlobalObject, source);
        return 0;
    }

    // This function assumes an input string that would result in a single anonymous function expression.
    StatementNode* exprStatement = program->singleStatement();
    ASSERT(exprStatement);
    ASSERT(exprStatement->isExprStatement());
    ExpressionNode* funcExpr = static_cast<ExprStatementNode*>(exprStatement)->expr();
    ASSERT(funcExpr);
    ASSERT(funcExpr->isFuncExprNode());
    FunctionBodyNode* body = static_cast<FuncExprNode*>(funcExpr)->body();
    ASSERT(body);
    ASSERT(body->ident().isNull());

    unsigned startColumn = source.startColumn();
    FunctionExecutable* functionExecutable = FunctionExecutable::create(globalData, body, startColumn, false);
    functionExecutable->m_nameValue.set(globalData, functionExecutable, jsString(&globalData, name.ustring()));
    return functionExecutable;
}

String FunctionExecutable::paramString() const
{
    FunctionParameters& parameters = *m_parameters;
    StringBuilder builder;
    for (size_t pos = 0; pos < parameters.size(); ++pos) {
        if (!builder.isEmpty())
            builder.append(", ");
        parameters.at(pos)->toString(builder);
    }
    return builder.toString();
}

}
