/*
 * Copyright (C) 2009, 2010 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#ifndef Executable_h
#define Executable_h

#include "ArityCheckMode.h"
#include "CallData.h"
#include "CodeSpecializationKind.h"
#include "CompilationResult.h"
//#include "DFGPlan.h"
#include "JSFunction.h"
#include "Interpreter.h"
#include "JITCode.h"
#include "JSGlobalObject.h"
#include "RegisterPreservationMode.h"
#include "Nodes.h"
#include "SamplingTool.h"
#include "SourceCode.h"
#include <wtf/DoublyLinkedList.h>
#include <wtf/PassOwnPtr.h>

namespace JSC {

class BuiltinExecutables;
class CodeBlock;
class Debugger;
class EvalCodeBlock;
class FunctionCodeBlock;
class Heap;
class LLIntOffsetsExtractor;
class ProgramCodeBlock;
class JSScope;

enum CompilationKind { FirstCompilation, OptimizingCompilation };

inline bool isCall(CodeSpecializationKind kind)
{
	if (kind == CodeForCall)
		return true;
	ASSERT(kind == CodeForConstruct);
	return false;
}

class ExecutableBase : public JSCell {
	friend class JIT;
	friend class DoublyLinkedList<ExecutableBase>;
	friend class Heap;

private:
    void setPrev(ExecutableBase* prev) { m_prev = prev; }
    void setNext(ExecutableBase* next) { m_next = next; }

    ExecutableBase* prev() const { return m_prev; }
    ExecutableBase* next() const { return m_next; }

	ExecutableBase* m_prev;
	ExecutableBase* m_next;

protected:
	static const int NUM_PARAMETERS_IS_HOST = 0;
	static const int NUM_PARAMETERS_NOT_COMPILED = -1;

	ExecutableBase(JSGlobalData& globalData, Structure* structure, int numParameters)
		: JSCell(globalData, structure)
		, m_prev(0)
		, m_next(0)
		, m_numParametersForCall(numParameters)
		, m_numParametersForConstruct(numParameters)
	{
	}

	void finishCreation(JSGlobalData& globalData)
	{
		Base::finishCreation(globalData);
	}

public:
	typedef JSCell Base;

	static const bool needsDestruction = true;
	static const bool hasImmortalStructure = true;
	static void destroy(JSCell*);

	bool isFunctionExecutable()
	{
		return structure()->typeInfo().type() == FunctionExecutableType;
	}

	bool isHostFunction() const
	{
		ASSERT((m_numParametersForCall == NUM_PARAMETERS_IS_HOST) == (m_numParametersForConstruct == NUM_PARAMETERS_IS_HOST));
		return m_numParametersForCall == NUM_PARAMETERS_IS_HOST;
	}

	static Structure* createStructure(JSGlobalData& globalData, JSGlobalObject* globalObject, JSValue proto) { return Structure::create(globalData, globalObject, proto, TypeInfo(CompoundType, StructureFlags), &s_info); }

	void clearCode();

	static JS_EXPORTDATA const ClassInfo s_info;

protected:
	static const unsigned StructureFlags = 0;
	int m_numParametersForCall;
	int m_numParametersForConstruct;

public:
	static void clearCodeVirtual(ExecutableBase*);

#if ENABLE(JIT)
	PassRefPtr<JITCode> generatedJITCodeForCall()
	{
		ASSERT(m_jitCodeForCall);
		return m_jitCodeForCall;
	}

	PassRefPtr<JITCode> generatedJITCodeForConstruct()
	{
		ASSERT(m_jitCodeForConstruct);
		return m_jitCodeForConstruct;
	}

	PassRefPtr<JITCode> generatedJITCodeFor(CodeSpecializationKind kind)
	{
		if (kind == CodeForCall)
			return generatedJITCodeForCall();
		ASSERT(kind == CodeForConstruct);
		return generatedJITCodeForConstruct();
	}

    MacroAssemblerCodePtr entrypointFor(
        VM& vm, CodeSpecializationKind kind, ArityCheckMode arity, RegisterPreservationMode registers)
    {
        // Check if we have a cached result. We only have it for arity check because we use the
        // no-arity entrypoint in non-virtual calls, which will "cache" this value directly in
        // machine code.
        if (arity == MustCheckArity) {
            switch (kind) {
            case CodeForCall:
                switch (registers) {
                case RegisterPreservationNotRequired:
                    if (MacroAssemblerCodePtr result = m_jitCodeForCallWithArityCheck)
                        return result;
                    break;
                case MustPreserveRegisters:
                    if (MacroAssemblerCodePtr result = m_jitCodeForCallWithArityCheckAndPreserveRegs)
                        return result;
                    break;
                }
                break;
            case CodeForConstruct:
                switch (registers) {
                case RegisterPreservationNotRequired:
                    if (MacroAssemblerCodePtr result = m_jitCodeForConstructWithArityCheck)
                        return result;
                    break;
                case MustPreserveRegisters:
                    if (MacroAssemblerCodePtr result = m_jitCodeForConstructWithArityCheckAndPreserveRegs)
                        return result;
                    break;
                }
                break;
            }
        }
        MacroAssemblerCodePtr result =
            generatedJITCodeFor(kind)->addressForCall(vm, this, arity, registers);
        if (arity == MustCheckArity) {
            // Cache the result; this is necessary for the JIT's virtual call optimizations.
            switch (kind) {
            case CodeForCall:
                switch (registers) {
                case RegisterPreservationNotRequired:
                    m_jitCodeForCallWithArityCheck = result;
                    break;
                case MustPreserveRegisters:
                    m_jitCodeForCallWithArityCheckAndPreserveRegs = result;
                    break;
                }
                break;
            case CodeForConstruct:
                switch (registers) {
                case RegisterPreservationNotRequired:
                    m_jitCodeForConstructWithArityCheck = result;
                    break;
                case MustPreserveRegisters:
                    m_jitCodeForConstructWithArityCheckAndPreserveRegs = result;
                    break;
                }
                break;
            }
        }
        return result;
    }

    static ptrdiff_t offsetOfJITCodeWithArityCheckFor(
        CodeSpecializationKind kind, RegisterPreservationMode registers)
    {
        switch (kind) {
        case CodeForCall:
            switch (registers) {
            case RegisterPreservationNotRequired:
                return OBJECT_OFFSETOF(ExecutableBase, m_jitCodeForCallWithArityCheck);
            case MustPreserveRegisters:
                return OBJECT_OFFSETOF(ExecutableBase, m_jitCodeForCallWithArityCheckAndPreserveRegs);
            }
        case CodeForConstruct:
            switch (registers) {
            case RegisterPreservationNotRequired:
                return OBJECT_OFFSETOF(ExecutableBase, m_jitCodeForConstructWithArityCheck);
            case MustPreserveRegisters:
                return OBJECT_OFFSETOF(ExecutableBase, m_jitCodeForConstructWithArityCheckAndPreserveRegs);
            }
        }
        RELEASE_ASSERT_NOT_REACHED();
        return 0;
    }

    static ptrdiff_t offsetOfNumParametersFor(CodeSpecializationKind kind)
    {
        if (kind == CodeForCall)
            return OBJECT_OFFSETOF(ExecutableBase, m_numParametersForCall);
        ASSERT(kind == CodeForConstruct);
        return OBJECT_OFFSETOF(ExecutableBase, m_numParametersForConstruct);
    }
#endif // ENABLE(JIT)

	bool hasJITCodeForCall() const
	{
		return m_numParametersForCall >= 0;
	}

	bool hasJITCodeForConstruct() const
	{
		return m_numParametersForConstruct >= 0;
	}

	bool hasJITCodeFor(CodeSpecializationKind kind) const
	{
		if (kind == CodeForCall)
			return hasJITCodeForCall();
		ASSERT(kind == CodeForConstruct);
		return hasJITCodeForConstruct();
	}

	// Intrinsics are only for calls, currently.
	Intrinsic intrinsic() const;

	Intrinsic intrinsicFor(CodeSpecializationKind kind) const
	{
		if (isCall(kind))
			return intrinsic();
		return NoIntrinsic;
	}

protected:
#if ENABLE(JIT)
	RefPtr<JITCode> m_jitCodeForCall;
	RefPtr<JITCode> m_jitCodeForConstruct;
	MacroAssemblerCodePtr m_jitCodeForCallWithArityCheck;
	MacroAssemblerCodePtr m_jitCodeForConstructWithArityCheck;
    MacroAssemblerCodePtr m_jitCodeForCallWithArityCheckAndPreserveRegs;
    MacroAssemblerCodePtr m_jitCodeForConstructWithArityCheckAndPreserveRegs;
#endif
};

class NativeExecutable : public ExecutableBase {
	friend class JIT;
	friend class LLIntOffsetsExtractor;
public:
	typedef ExecutableBase Base;

#if ENABLE(JIT)
	static NativeExecutable* create(JSGlobalData& globalData, MacroAssemblerCodeRef callThunk, NativeFunction function, MacroAssemblerCodeRef constructThunk, NativeFunction constructor, Intrinsic intrinsic)
	{
		NativeExecutable* executable;
		if (!callThunk) {
			executable = new (NotNull, allocateCell<NativeExecutable>(globalData.heap)) NativeExecutable(globalData, function, constructor);
			executable->finishCreation(globalData, 0, 0, intrinsic);
		} else {
			executable = new (NotNull, allocateCell<NativeExecutable>(globalData.heap)) NativeExecutable(globalData, function, constructor);
			executable->finishCreation(globalData, JITCode::hostFunction(callThunk), JITCode::hostFunction(constructThunk), intrinsic);
		}
		return executable;
	}
#endif

	static void destroy(JSCell*);

	NativeFunction function() { return m_function; }
	NativeFunction constructor() { return m_constructor; }

    NativeFunction nativeFunctionFor(CodeSpecializationKind kind)
    {
        if (kind == CodeForCall)
            return function();
        ASSERT(kind == CodeForConstruct);
        return constructor();
    }

    static ptrdiff_t offsetOfNativeFunctionFor(CodeSpecializationKind kind)
    {
        if (kind == CodeForCall)
            return OBJECT_OFFSETOF(NativeExecutable, m_function);
        ASSERT(kind == CodeForConstruct);
        return OBJECT_OFFSETOF(NativeExecutable, m_constructor);
    }

	static Structure* createStructure(JSGlobalData& globalData, JSGlobalObject* globalObject, JSValue proto) { return Structure::create(globalData, globalObject, proto, TypeInfo(LeafType, StructureFlags), &s_info); }

	static const ClassInfo s_info;

	Intrinsic intrinsic() const;

protected:
#if ENABLE(JIT)
	void finishCreation(JSGlobalData& globalData, PassRefPtr<JITCode> callThunk, PassRefPtr<JITCode> constructThunk, Intrinsic intrinsic)
	{
		Base::finishCreation(globalData);
		m_jitCodeForCall = callThunk;
		m_jitCodeForConstruct = constructThunk;
		m_intrinsic = intrinsic;
	}
#endif

private:
	NativeExecutable(JSGlobalData& globalData, NativeFunction function, NativeFunction constructor)
		: ExecutableBase(globalData, globalData.nativeExecutableStructure.get(), NUM_PARAMETERS_IS_HOST)
		, m_function(function)
		, m_constructor(constructor)
	{
	}

	NativeFunction m_function;
	NativeFunction m_constructor;

	Intrinsic m_intrinsic;
};

class ScriptExecutable : public ExecutableBase {
public:
	typedef ExecutableBase Base;

	ScriptExecutable(Structure* structure, JSGlobalData& globalData, const SourceCode& source, bool isInStrictContext);

	static void destroy(JSCell*);

	const SourceCode& source() { return m_source; }
	intptr_t sourceID() const { return m_source.providerID(); }
	const String& sourceURL() const { return m_source.provider()->url(); }
	int lineNo() const { return m_firstLine; }
	int lastLine() const { return m_lastLine; }
    unsigned startColumn() const { return m_startColumn; }

	bool usesEval() const { return m_features & EvalFeature; }
	bool usesArguments() const { return m_features & ArgumentsFeature; }
	bool needsActivation() const { return m_hasCapturedVariables || m_features & (EvalFeature | WithFeature | CatchFeature); }
	bool isStrictMode() const { return m_features & StrictModeFeature; }

    void setNeverInline(bool value) { m_neverInline = value; }
    void setDidTryToEnterInLoop(bool value) { m_didTryToEnterInLoop = value; }
    bool neverInline() const { return m_neverInline; }
    bool didTryToEnterInLoop() const { return m_didTryToEnterInLoop; }
    bool isInliningCandidate() const { return !neverInline(); }

    bool* addressOfDidTryToEnterInLoop() { return &m_didTryToEnterInLoop; }

	void unlinkCalls();

    CodeFeatures features() const { return m_features; }

	static const ClassInfo s_info;

    void recordParse(CodeFeatures features, bool hasCapturedVariables, int firstLine, int lastLine, unsigned startColumn)
    {
        m_features = features;
        m_hasCapturedVariables = hasCapturedVariables;
        m_firstLine = firstLine;
        m_lastLine = lastLine;
        m_startColumn = startColumn;
    }

    void installCode(CodeBlock*);
    PassRefPtr<CodeBlock> newCodeBlockFor(ExecState*, CodeSpecializationKind, JSScope*, JSObject*& exception);
    PassRefPtr<CodeBlock> newReplacementCodeBlockFor(CodeSpecializationKind);

    // This function has an interesting GC story. Callers of this function are asking us to create a CodeBlock
	// that is not jettisoned before this function returns. Callers are essentially asking for a strong reference
	// to the CodeBlock. Because the Executable may be allocating the CodeBlock, we require callers to pass in
	// their CodeBlock*& reference because it's safe for CodeBlock to be jettisoned if Executable is the only thing
	// to point to it. This forces callers to have a CodeBlock* in a register or on the stack that will be marked
	// by conservative GC if a GC happens after we create the CodeBlock.
    template <typename ExecutableType>
    JSObject* prepareForExecution(ExecState* exec, JSScope* scope, CodeSpecializationKind kind, CodeBlock*& resultCodeBlock);

private:
    JSObject* prepareForExecutionImpl(ExecState*, JSScope*, CodeSpecializationKind, CodeBlock*& resultCodeBlock);

protected:
	void finishCreation(JSGlobalData& globalData)
	{
		Base::finishCreation(globalData);
		globalData.heap.addExecutable(this); // Balanced by Heap::deleteUnmarkedCompiledCode().

#if ENABLE(CODEBLOCK_SAMPLING)
		if (SamplingTool* sampler = globalData.interpreter->sampler())
			sampler->notifyOfScope(globalData, this);
#endif
	}

	SourceCode m_source;
	CodeFeatures m_features;
	bool m_hasCapturedVariables;
    bool m_neverInline;
    bool m_didTryToEnterInLoop;
	int m_firstLine;
	int m_lastLine;
    unsigned m_startColumn;
};

class EvalExecutable : public ScriptExecutable {
	friend class LLIntOffsetsExtractor;
public:
	typedef ScriptExecutable Base;

	static void destroy(JSCell*);

    EvalCodeBlock* codeBlock()
    {
        return m_evalCodeBlock.get();
    }

	EvalCodeBlock* generatedBytecode()
	{
		ASSERT(m_evalCodeBlock);
		return m_evalCodeBlock.get();
	}

	static EvalExecutable* create(ExecState* exec, const SourceCode& source, bool isInStrictContext)
	{
		EvalExecutable* executable = new (NotNull, allocateCell<EvalExecutable>(*exec->heap())) EvalExecutable(exec, source, isInStrictContext);
		executable->finishCreation(exec->globalData());
		return executable;
	}

#if ENABLE(JIT)
	PassRefPtr<JITCode> generatedJITCode()
	{
		return generatedJITCodeForCall();
	}
#endif
	static Structure* createStructure(JSGlobalData& globalData, JSGlobalObject* globalObject, JSValue proto)
	{
		return Structure::create(globalData, globalObject, proto, TypeInfo(EvalExecutableType, StructureFlags), &s_info);
	}

    DECLARE_INFO;

	void unlinkCalls();

	void clearCode();

    unsigned numVariables();
    unsigned numberOfFunctionDecls();

private:
    friend class ScriptExecutable;
	EvalExecutable(ExecState*, const SourceCode&, bool);

	static void visitChildren(JSCell*, SlotVisitor&);

	RefPtr<EvalCodeBlock> m_evalCodeBlock;
};

class ProgramExecutable : public ScriptExecutable {
	friend class LLIntOffsetsExtractor;
public:
	typedef ScriptExecutable Base;

	static ProgramExecutable* create(ExecState* exec, const SourceCode& source)
	{
		ProgramExecutable* executable = new (NotNull, allocateCell<ProgramExecutable>(*exec->heap())) ProgramExecutable(exec, source);
		executable->finishCreation(exec->globalData());
		return executable;
	}

	static void destroy(JSCell*);

	ProgramCodeBlock* codeBlock()
	{
		return m_programCodeBlock.get();
	}

	ProgramCodeBlock* generatedBytecode()
	{
		ASSERT(m_programCodeBlock);
		return m_programCodeBlock.get();
	}

	JSObject* checkSyntax(ExecState*);

#if ENABLE(JIT)
	PassRefPtr<JITCode> generatedJITCode()
	{
		return generatedJITCodeForCall();
	}
#endif

	static Structure* createStructure(JSGlobalData& globalData, JSGlobalObject* globalObject, JSValue proto)
	{
		return Structure::create(globalData, globalObject, proto, TypeInfo(ProgramExecutableType, StructureFlags), &s_info);
	}

    DECLARE_INFO;

	void unlinkCalls();

	void clearCode();

private:
    friend class ScriptExecutable;

	ProgramExecutable(ExecState*, const SourceCode&);

	static void visitChildren(JSCell*, SlotVisitor&);

	RefPtr<ProgramCodeBlock> m_programCodeBlock;
};

class FunctionExecutable : public ScriptExecutable {
	friend class JIT;
	friend class LLIntOffsetsExtractor;
public:
	typedef ScriptExecutable Base;

	static FunctionExecutable* create(JSGlobalData& globalData, FunctionBodyNode* node, unsigned startColumn, bool isBuiltinFunction)
	{
		FunctionExecutable* executable = new (NotNull, allocateCell<FunctionExecutable>(globalData.heap)) FunctionExecutable(globalData, node, startColumn, isBuiltinFunction);
		executable->finishCreation(globalData);
		return executable;
	}

	static FunctionExecutable* fromGlobalCode(const Identifier& name, ExecState*, Debugger*, const SourceCode&, JSObject** exception);

	static void destroy(JSCell*);

	// Returns either call or construct bytecode. This can be appropriate
	// for answering questions that that don't vary between call and construct --
	// for example, argumentsRegister().
	FunctionCodeBlock& generatedBytecode()
	{
		if (m_codeBlockForCall)
			return *m_codeBlockForCall;
		ASSERT(m_codeBlockForConstruct);
		return *m_codeBlockForConstruct;
	}

	bool isGeneratedForCall() const
	{
		return m_codeBlockForCall;
	}

	FunctionCodeBlock* generatedBytecodeForCall()
	{
		ASSERT(m_codeBlockForCall);
		return m_codeBlockForCall.get();
	}

	bool isGeneratedForConstruct() const
	{
		return m_codeBlockForConstruct;
	}

	FunctionCodeBlock* generatedBytecodeForConstruct()
	{
		ASSERT(m_codeBlockForConstruct);
		return m_codeBlockForConstruct.get();
	}

	bool isGeneratedFor(CodeSpecializationKind kind)
	{
		if (kind == CodeForCall)
			return isGeneratedForCall();
		ASSERT(kind == CodeForConstruct);
		return isGeneratedForConstruct();
	}

	FunctionCodeBlock* generatedBytecodeFor(CodeSpecializationKind kind)
	{
		if (kind == CodeForCall)
			return generatedBytecodeForCall();
		ASSERT(kind == CodeForConstruct);
		return generatedBytecodeForConstruct();
	}

	FunctionCodeBlock* baselineCodeBlockFor(CodeSpecializationKind);

	FunctionCodeBlock* profiledCodeBlockFor(CodeSpecializationKind kind)
	{
		return baselineCodeBlockFor(kind);
	}

	JSParserStrictness toStrictness() const
	{
		if (m_isBuiltinFunction)
			return JSParseBuiltin;
		if (isStrictMode())
			return JSParseStrict;
		return JSParseNormal;
	}

	bool isBuiltinFunction() const { return m_isBuiltinFunction; }
	const Identifier& name() { return m_name; }
	const String& inferredName() { return m_inferredName; }
	JSString* nameValue() const { return m_nameValue.get(); }
	size_t parameterCount() const { return m_parameters->size(); } // Excluding 'this'!
	String paramString() const;

	static void visitChildren(JSCell*, SlotVisitor&);
	static Structure* createStructure(JSGlobalData& globalData, JSGlobalObject* globalObject, JSValue proto)
	{
		return Structure::create(globalData, globalObject, proto, TypeInfo(FunctionExecutableType, StructureFlags), &s_info);
	}

	DECLARE_INFO;

	void unlinkCalls();

	void clearCode();

protected:
	void finishCreation(JSGlobalData& globalData)
	{
		Base::finishCreation(globalData);
		m_nameValue.set(globalData, this, jsString(&globalData, name().ustring()));
	}

private:
	friend class BuiltinExecutables;
	FunctionExecutable(JSGlobalData&, FunctionBodyNode*, unsigned startColumn, bool isBuiltinFunction);

	RefPtr<FunctionCodeBlock> codeBlockFor(CodeSpecializationKind kind)
	{
		if (kind == CodeForCall)
			return m_codeBlockForCall;
		ASSERT(kind == CodeForConstruct);
		return m_codeBlockForConstruct;
	}

    friend class ScriptExecutable;

	bool m_forceUsesArguments;

	RefPtr<FunctionParameters> m_parameters;
	RefPtr<FunctionCodeBlock> m_codeBlockForCall;
	RefPtr<FunctionCodeBlock> m_codeBlockForConstruct;
    bool m_isBuiltinFunction : 1;
	Identifier m_name;
	String m_inferredName;
	FunctionNameIsInScopeToggle m_functionNameIsInScopeToggle;
	WriteBarrier<JSString> m_nameValue;
};

inline void ExecutableBase::clearCodeVirtual(ExecutableBase* executable)
{
	switch (executable->structure()->typeInfo().type()) {
	case EvalExecutableType:
		return jsCast<EvalExecutable*>(executable)->clearCode();
	case ProgramExecutableType:
		return jsCast<ProgramExecutable*>(executable)->clearCode();
	case FunctionExecutableType:
		return jsCast<FunctionExecutable*>(executable)->clearCode();
	default:
		return jsCast<NativeExecutable*>(executable)->clearCode();
	}
}

inline void ScriptExecutable::unlinkCalls()
{
	switch (structure()->typeInfo().type()) {
	case EvalExecutableType:
		return jsCast<EvalExecutable*>(this)->unlinkCalls();
	case ProgramExecutableType:
		return jsCast<ProgramExecutable*>(this)->unlinkCalls();
	case FunctionExecutableType:
		return jsCast<FunctionExecutable*>(this)->unlinkCalls();
	default:
		ASSERT_NOT_REACHED();
	}
}

template <typename ExecutableType>
JSObject* ScriptExecutable::prepareForExecution(ExecState* exec, JSScope* scope, CodeSpecializationKind kind, CodeBlock*& resultCodeBlock)
{
    if (hasJITCodeFor(kind)) {
        if (classInfo() == EvalExecutable::info())
            resultCodeBlock = jsCast<CodeBlock*>(jsCast<EvalExecutable*>(this)->codeBlock());
        else if (classInfo() == ProgramExecutable::info())
            resultCodeBlock = jsCast<CodeBlock*>(jsCast<ProgramExecutable*>(this)->codeBlock());
        //else if (std::is_same<ExecutableType, ModuleProgramExecutable>::value)
        //    resultCodeBlock = jsCast<CodeBlock*>(jsCast<ModuleProgramExecutable*>(this)->codeBlock());
        else if (classInfo() == FunctionExecutable::info())
            resultCodeBlock = jsCast<CodeBlock*>(jsCast<FunctionExecutable*>(this)->codeBlockFor(kind).get());
        else
            RELEASE_ASSERT_NOT_REACHED();
        return 0;
    }
    return prepareForExecutionImpl(exec, scope, kind, resultCodeBlock);
}

}

#include "JSFunctionInlines.h"

#endif
