/*
 *  Copyright (C) 2007 Eric Seidel <eric@webkit.org>
 *  Copyright (C) 2007, 2008, 2009 Apple Inc. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public License
 *  along with this library; see the file COPYING.LIB.  If not, write to
 *  the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 *  Boston, MA 02110-1301, USA.
 *
 */

#ifndef JSGlobalObject_h
#define JSGlobalObject_h

#include "ArrayAllocationProfile.h"
#include "ConstantMode.h"
#include "JSArray.h"
#include "JSArrayBufferPrototype.h"
#include "JSClassRef.h"
#include "JSProxy.h"
#include "JSSegmentedVariableObject.h"
#include "JSWeakObjectMapRefInternal.h"
#include "NumberPrototype.h"
#include "SpecialPointer.h"
#include "StringPrototype.h"
#include "StructureChain.h"
#include "StructureRareDataInlines.h"
#include "VM.h"
#include "Watchpoint.h"
#include <wtf/HashSet.h>
#include <wtf/OwnPtr.h>
#include <wtf/PassRefPtr.h>
#include <wtf/RandomNumber.h>

struct OpaqueJSClass;
struct OpaqueJSClassContextData;

namespace JSC {

class BuiltinExecutables;
class ArrayPrototype;
class BooleanPrototype;
class Debugger;
class ErrorConstructor;
class ErrorPrototype;
class FunctionPrototype;
class GetterSetter;
class GlobalCodeBlock;
class JSPromiseConstructor;
class JSPromisePrototype;
class LLIntOffsetsExtractor;
class Microtask;
class NativeErrorConstructor;
class ObjectConstructor;
class ProgramCodeBlock;
class RegExpConstructor;
class RegExpPrototype;
class JSStack;

struct ActivationStackNode;
struct HashTable;

#define FOR_EACH_SIMPLE_BUILTIN_TYPE(macro) \
    macro(Set, set, set, JSSet, Set) \
    macro(Map, map, map, JSMap, Map) \
    macro(Date, date, date, DateInstance, Date) \
    macro(String, string, stringObject, StringObject, String) \
    macro(Boolean, boolean, booleanObject, BooleanObject, Boolean) \
    macro(Number, number, numberObject, NumberObject, Number) \
    macro(Error, error, error, ErrorInstance, Error) \
    macro(JSArrayBuffer, arrayBuffer, arrayBuffer, JSArrayBuffer, ArrayBuffer) \
    macro(ArrayIterator, arrayIterator, arrayIterator, JSArrayIterator, ArrayIterator) \
	macro(ArgumentsIterator, argumentsIterator, argumentsIterator, JSArgumentsIterator, ArgumentsIterator) \

#define DECLARE_SIMPLE_BUILTIN_TYPE(capitalName, lowerName, properName, instanceType, jsName) \
    class JS ## capitalName; \
    class capitalName ## Prototype; \
    class capitalName ## Constructor;

FOR_EACH_SIMPLE_BUILTIN_TYPE(DECLARE_SIMPLE_BUILTIN_TYPE)

#undef DECLARE_SIMPLE_BUILTIN_TYPE

typedef Vector<ExecState*, 16> ExecStateStack;

struct GlobalObjectMethodTable {
	typedef bool (*AllowsAccessFromFunctionPtr)(const JSGlobalObject*, ExecState*);
	AllowsAccessFromFunctionPtr allowsAccessFrom;

	typedef bool (*SupportsProfilingFunctionPtr)(const JSGlobalObject*);
	SupportsProfilingFunctionPtr supportsProfiling;

	typedef bool (*SupportsRichSourceInfoFunctionPtr)(const JSGlobalObject*);
	SupportsRichSourceInfoFunctionPtr supportsRichSourceInfo;

	typedef bool (*ShouldInterruptScriptFunctionPtr)(const JSGlobalObject*);
	ShouldInterruptScriptFunctionPtr shouldInterruptScript;

	typedef bool (*JavaScriptExperimentsEnabledFunctionPtr)(const JSGlobalObject*);
	JavaScriptExperimentsEnabledFunctionPtr javaScriptExperimentsEnabled;

    typedef void (*QueueTaskToEventLoopFunctionPtr)(const JSGlobalObject*, PassRefPtr<Microtask>);
    QueueTaskToEventLoopFunctionPtr queueTaskToEventLoop;
};

class JSGlobalObject : public JSSegmentedVariableObject {
private:
	typedef HashSet<RefPtr<OpaqueJSWeakObjectMap> > WeakMapSet;
	typedef HashMap<OpaqueJSClass*, OwnPtr<OpaqueJSClassContextData> > OpaqueJSClassDataMap;

	struct JSGlobalObjectRareData {
		WTF_MAKE_FAST_ALLOCATED;
	public:
		JSGlobalObjectRareData()
			: profileGroup(0)
		{
		}

		WeakMapSet weakMaps;
		unsigned profileGroup;

        OpaqueJSClassDataMap opaqueJSClassData;
	};

protected:
    Register m_globalCallFrame[JSStack::CallFrameHeaderSize];

	WriteBarrier<JSObject> m_globalThis;

	WriteBarrier<RegExpConstructor> m_regExpConstructor;
	WriteBarrier<ErrorConstructor> m_errorConstructor;
	WriteBarrier<NativeErrorConstructor> m_evalErrorConstructor;
	WriteBarrier<NativeErrorConstructor> m_rangeErrorConstructor;
	WriteBarrier<NativeErrorConstructor> m_referenceErrorConstructor;
	WriteBarrier<NativeErrorConstructor> m_syntaxErrorConstructor;
	WriteBarrier<NativeErrorConstructor> m_typeErrorConstructor;
	WriteBarrier<NativeErrorConstructor> m_URIErrorConstructor;
    WriteBarrier<ObjectConstructor> m_objectConstructor;
    WriteBarrier<JSPromiseConstructor> m_promiseConstructor;

    WriteBarrier<JSFunction> m_parseIntFunction;

	WriteBarrier<JSFunction> m_evalFunction;
	WriteBarrier<JSFunction> m_callFunction;
	WriteBarrier<JSFunction> m_applyFunction;
    WriteBarrier<JSFunction> m_initializePromiseFunction;
    WriteBarrier<JSFunction> m_newPromiseDeferredFunction;
	WriteBarrier<GetterSetter> m_throwTypeErrorGetterSetter;

	WriteBarrier<ObjectPrototype> m_objectPrototype;
	WriteBarrier<FunctionPrototype> m_functionPrototype;
	WriteBarrier<ArrayPrototype> m_arrayPrototype;
	WriteBarrier<RegExpPrototype> m_regExpPrototype;
    WriteBarrier<JSPromisePrototype> m_promisePrototype;

	WriteBarrier<Structure> m_withScopeStructure;
	WriteBarrier<Structure> m_strictEvalActivationStructure;
	WriteBarrier<Structure> m_activationStructure;
	WriteBarrier<Structure> m_nameScopeStructure;
	WriteBarrier<Structure> m_argumentsStructure;

    WriteBarrier<Structure> m_objectStructureForObjectConstructor;

	// Lists the actual structures used for having these particular indexing shapes.
	WriteBarrier<Structure> m_originalArrayStructureForIndexingShape[NumberOfIndexingShapes];
	// Lists the structures we should use during allocation for these particular indexing shapes.
	WriteBarrier<Structure> m_arrayStructureForIndexingShapeDuringAllocation[NumberOfIndexingShapes];

	WriteBarrier<Structure> m_callbackConstructorStructure;
	WriteBarrier<Structure> m_callbackFunctionStructure;
	WriteBarrier<Structure> m_callbackObjectStructure;
	WriteBarrier<Structure> m_nullPrototypeObjectStructure;
    WriteBarrier<Structure> m_calleeStructure;
	WriteBarrier<Structure> m_functionStructure;
	WriteBarrier<Structure> m_boundFunctionStructure;
	WriteBarrier<Structure> m_namedFunctionStructure;
	PropertyOffset m_functionNameOffset;
	WriteBarrier<Structure> m_privateNameStructure;
	WriteBarrier<Structure> m_regExpStructure;
	WriteBarrier<Structure> m_internalFunctionStructure;

    WriteBarrier<Structure> m_iteratorResultStructure;

    WriteBarrier<Structure> m_regExpMatchesArrayStructure;
    WriteBarrier<Structure> m_regExpMatchesArraySlowPutStructure;

    WriteBarrier<Structure> m_promiseStructure;

    WriteBarrier<Structure> m_mapDataStructure;

    WriteBarrier<StructureChain> m_cachedArrayPrototypeChain;
    WriteBarrier<StructureChain> m_cachedNullPrototypeChain;
    WriteBarrier<StructureChain> m_cachedObjectPrototypeChain;
    WriteBarrier<StructureChain> m_cachedStringPrototypeChain;

#define DEFINE_STORAGE_FOR_SIMPLE_TYPE(capitalName, lowerName, properName, instanceType, jsName) \
    WriteBarrier<capitalName ## Prototype> m_ ## lowerName ## Prototype; \
    WriteBarrier<Structure> m_ ## properName ## Structure;

    FOR_EACH_SIMPLE_BUILTIN_TYPE(DEFINE_STORAGE_FOR_SIMPLE_TYPE)

#undef DEFINE_STORAGE_FOR_SIMPLE_TYPE

    struct TypedArrayData {
        WriteBarrier<JSObject> prototype;
        WriteBarrier<Structure> structure;
    };

    FixedArray<TypedArrayData, NUMBER_OF_TYPED_ARRAY_TYPES> m_typedArrays;

	void* m_specialPointers[Special::TableSize]; // Special pointers used by the LLInt and JIT.

	Debugger* m_debugger;

    VM& m_vm;

	RefPtr<WatchpointSet> m_masqueradesAsUndefinedWatchpoint;
	RefPtr<WatchpointSet> m_havingABadTimeWatchpoint;
    RefPtr<WatchpointSet> m_varInjectionWatchpoint;

	OwnPtr<JSGlobalObjectRareData> m_rareData;

    OwnPtr<BuiltinExecutables> m_builtinExecutables;

	WeakRandom m_weakRandom;

	bool m_evalEnabled;
	bool m_experimentsEnabled;

	static JS_EXPORTDATA const GlobalObjectMethodTable s_globalObjectMethodTable;
	const GlobalObjectMethodTable* m_globalObjectMethodTable;

	void createRareDataIfNeeded()
	{
		if (m_rareData)
			return;
		m_rareData = adoptPtr(new JSGlobalObjectRareData);
	}

public:
	typedef JSSegmentedVariableObject Base;

	static JSGlobalObject* create(JSGlobalData& globalData, Structure* structure)
	{
		JSGlobalObject* globalObject = new (NotNull, allocateCell<JSGlobalObject>(globalData.heap)) JSGlobalObject(globalData, structure);
		globalObject->finishCreation(globalData);
		globalData.heap.addFinalizer(globalObject, destroy);
		return globalObject;
	}

	static JS_EXPORTDATA const ClassInfo s_info;

    bool hasDebugger() const { return m_debugger; }
    bool hasProfiler() const { return globalObjectMethodTable()->supportsProfiling(this); }

protected:
	JS_EXPORT_PRIVATE explicit JSGlobalObject(JSGlobalData&, Structure*, const GlobalObjectMethodTable* = 0);

	void finishCreation(JSGlobalData& vm)
	{
		Base::finishCreation(vm);
		structure()->setGlobalObject(vm, this);
		m_experimentsEnabled = m_globalObjectMethodTable->javaScriptExperimentsEnabled(this);
		init(vm);
		setGlobalThis(vm, JSProxy::create(vm, JSProxy::createStructure(vm, this, prototype(), PureForwardingProxyType), this));
	}

	void finishCreation(JSGlobalData& vm, JSObject* thisValue)
	{
		Base::finishCreation(vm);
		structure()->setGlobalObject(vm, this);
		m_experimentsEnabled = m_globalObjectMethodTable->javaScriptExperimentsEnabled(this);
		init(vm);
        setGlobalThis(vm, thisValue);
	}

    struct NewGlobalVar {
        int registerNumber;
        VariableWatchpointSet* set;
    };
    NewGlobalVar addGlobalVar(const Identifier&, ConstantMode);

public:
	JS_EXPORT_PRIVATE ~JSGlobalObject();
	JS_EXPORT_PRIVATE static void destroy(JSCell*);
	// We don't need a destructor because we use a finalizer instead.
	static const bool needsDestruction = false;

	JS_EXPORT_PRIVATE static void visitChildren(JSCell*, SlotVisitor&);

	JS_EXPORT_PRIVATE static bool getOwnPropertySlot(JSObject*, ExecState*, PropertyName, PropertySlot&);
	bool hasOwnPropertyForWrite(ExecState*, PropertyName);
	JS_EXPORT_PRIVATE static void put(JSCell*, ExecState*, PropertyName, JSValue, PutPropertySlot&);

	JS_EXPORT_PRIVATE static void defineGetter(JSObject*, ExecState*, PropertyName, JSObject* getterFunc, unsigned attributes);
	JS_EXPORT_PRIVATE static void defineSetter(JSObject*, ExecState*, PropertyName, JSObject* setterFunc, unsigned attributes);
	JS_EXPORT_PRIVATE static bool defineOwnProperty(JSObject*, ExecState*, PropertyName, const PropertyDescriptor&, bool shouldThrow);

	// We use this in the code generator as we perform symbol table
	// lookups prior to initializing the properties
	bool symbolTableHasProperty(PropertyName);

    void addVar(ExecState* exec, const Identifier& propertyName)
    {
        if (!hasProperty(exec, propertyName))
            addGlobalVar(propertyName, IsVariable);
    }
    void addConst(ExecState* exec, const Identifier& propertyName)
    {
        if (!hasProperty(exec, propertyName))
            addGlobalVar(propertyName, IsConstant);
    }
    void addFunction(ExecState*, const Identifier&, JSValue);

	// The following accessors return pristine values, even if a script
	// replaces the global object's associated property.

	RegExpConstructor* regExpConstructor() const { return m_regExpConstructor.get(); }

	ErrorConstructor* errorConstructor() const { return m_errorConstructor.get(); }
	ObjectConstructor* objectConstructor() const { return m_objectConstructor.get(); }
	NativeErrorConstructor* evalErrorConstructor() const { return m_evalErrorConstructor.get(); }
	NativeErrorConstructor* rangeErrorConstructor() const { return m_rangeErrorConstructor.get(); }
	NativeErrorConstructor* referenceErrorConstructor() const { return m_referenceErrorConstructor.get(); }
	NativeErrorConstructor* syntaxErrorConstructor() const { return m_syntaxErrorConstructor.get(); }
	NativeErrorConstructor* typeErrorConstructor() const { return m_typeErrorConstructor.get(); }
	NativeErrorConstructor* URIErrorConstructor() const { return m_URIErrorConstructor.get(); }
	JSPromiseConstructor* promiseConstructor() const { return m_promiseConstructor.get(); }

    JSFunction* parseIntFunction() const { return m_parseIntFunction.get(); }

	JSFunction* evalFunction() const { return m_evalFunction.get(); }
	JSFunction* callFunction() const { return m_callFunction.get(); }
	JSFunction* applyFunction() const { return m_applyFunction.get(); }
	JSFunction* initializePromiseFunction() const { return m_initializePromiseFunction.get(); }
	JSFunction* newPromiseDeferredFunction() const { return m_newPromiseDeferredFunction.get(); }
	GetterSetter* throwTypeErrorGetterSetter(JSGlobalData& vm)
	{
		if (!m_throwTypeErrorGetterSetter)
			createThrowTypeError(vm);
		return m_throwTypeErrorGetterSetter.get();
	}

	ObjectPrototype* objectPrototype() const { return m_objectPrototype.get(); }
	FunctionPrototype* functionPrototype() const { return m_functionPrototype.get(); }
	ArrayPrototype* arrayPrototype() const { return m_arrayPrototype.get(); }
	BooleanPrototype* booleanPrototype() const { return m_booleanPrototype.get(); }
	StringPrototype* stringPrototype() const { return m_stringPrototype.get(); }
	NumberPrototype* numberPrototype() const { return m_numberPrototype.get(); }
	DatePrototype* datePrototype() const { return m_datePrototype.get(); }
	RegExpPrototype* regExpPrototype() const { return m_regExpPrototype.get(); }
	ErrorPrototype* errorPrototype() const { return m_errorPrototype.get(); }
	JSPromisePrototype* promisePrototype() const { return m_promisePrototype.get(); }

	StructureChain* cachedArrayPrototypeChain() const { return m_cachedArrayPrototypeChain.get(); }
	StructureChain* cachedNullPrototypeChain() const { return m_cachedNullPrototypeChain.get(); }
	StructureChain* cachedObjectPrototypeChain() const { return m_cachedObjectPrototypeChain.get(); }
	StructureChain* cachedStringPrototypeChain() const { return m_cachedStringPrototypeChain.get(); }

	Structure* withScopeStructure() const { return m_withScopeStructure.get(); }
	Structure* strictEvalActivationStructure() const { return m_strictEvalActivationStructure.get(); }
	Structure* activationStructure() const { return m_activationStructure.get(); }
	Structure* nameScopeStructure() const { return m_nameScopeStructure.get(); }
	Structure* argumentsStructure() const { return m_argumentsStructure.get(); }
	Structure* objectStructureForObjectConstructor() const { return m_objectStructureForObjectConstructor.get(); }
	Structure* originalArrayStructureForIndexingType(IndexingType indexingType) const
	{
		ASSERT(indexingType & IsArray);
		return m_originalArrayStructureForIndexingShape[(indexingType & IndexingShapeMask) >> IndexingShapeShift].get();
	}
	Structure* arrayStructureForIndexingTypeDuringAllocation(IndexingType indexingType) const
	{
		ASSERT(indexingType & IsArray);
		return m_arrayStructureForIndexingShapeDuringAllocation[(indexingType & IndexingShapeMask) >> IndexingShapeShift].get();
	}
	Structure* arrayStructureForProfileDuringAllocation(ArrayAllocationProfile* profile) const
	{
		return arrayStructureForIndexingTypeDuringAllocation(ArrayAllocationProfile::selectIndexingTypeFor(profile));
	}

	bool isOriginalArrayStructure(Structure* structure)
	{
		return originalArrayStructureForIndexingType(structure->indexingType() | IsArray) == structure;
	}

	Structure* booleanObjectStructure() const { return m_booleanObjectStructure.get(); }
	Structure* callbackConstructorStructure() const { return m_callbackConstructorStructure.get(); }
	Structure* callbackFunctionStructure() const { return m_callbackFunctionStructure.get(); }
	Structure* callbackObjectStructure() const { return m_callbackObjectStructure.get(); }
	Structure* dateStructure() const { return m_dateStructure.get(); }
	Structure* nullPrototypeObjectStructure() const { return m_nullPrototypeObjectStructure.get(); }
	Structure* errorStructure() const { return m_errorStructure.get(); }
    Structure* calleeStructure() const { return m_calleeStructure.get(); }
	Structure* functionStructure() const { return m_functionStructure.get(); }
	Structure* boundFunctionStructure() const { return m_boundFunctionStructure.get(); }
	Structure* namedFunctionStructure() const { return m_namedFunctionStructure.get(); }
	PropertyOffset functionNameOffset() const { return m_functionNameOffset; }
	Structure* numberObjectStructure() const { return m_numberObjectStructure.get(); }
	Structure* privateNameStructure() const { return m_privateNameStructure.get(); }
	Structure* internalFunctionStructure() const { return m_internalFunctionStructure.get(); }
	Structure* mapStructure() const { return m_mapStructure.get(); }
	Structure* mapDataStructure() const { return m_mapDataStructure.get(); }
	Structure* regExpStructure() const { return m_regExpStructure.get(); }
	Structure* setStructure() const { return m_setStructure.get(); }
	Structure* stringObjectStructure() const { return m_stringObjectStructure.get(); }
	Structure* iteratorResultStructure() const { return m_iteratorResultStructure.get(); }
	static ptrdiff_t iteratorResultStructureOffset() { return OBJECT_OFFSETOF(JSGlobalObject, m_iteratorResultStructure); }
	Structure* regExpMatchesArrayStructure() const { return m_regExpMatchesArrayStructure.get(); }

	Structure* promiseStructure() const { return m_promiseStructure.get(); }

    JSArrayBufferPrototype* arrayBufferPrototype() const { return m_arrayBufferPrototype.get(); }

#define DEFINE_ACCESSORS_FOR_SIMPLE_TYPE(capitalName, lowerName, properName, instanceType, jsName) \
    Structure* properName ## Structure() { return m_ ## properName ## Structure.get(); }

    FOR_EACH_SIMPLE_BUILTIN_TYPE(DEFINE_ACCESSORS_FOR_SIMPLE_TYPE)

#undef DEFINE_ACCESSORS_FOR_SIMPLE_TYPE

    Structure* typedArrayStructure(TypedArrayType type) const
    {
        return m_typedArrays[toIndex(type)].structure.get();
    }

    bool isOriginalTypedArrayStructure(Structure* structure)
    {
        TypedArrayType type = structure->classInfo()->typedArrayStorageType;
        if (type == NotTypedArray)
            return false;
        return typedArrayStructure(type) == structure;
    }

	void* actualPointerFor(Special::Pointer pointer)
	{
		ASSERT(pointer < Special::TableSize);
		return m_specialPointers[pointer];
	}

	WatchpointSet* masqueradesAsUndefinedWatchpoint() { return m_masqueradesAsUndefinedWatchpoint.get(); }
	WatchpointSet* havingABadTimeWatchpoint() { return m_havingABadTimeWatchpoint.get(); }
	WatchpointSet* varInjectionWatchpoint() { return m_varInjectionWatchpoint.get(); }

	bool isHavingABadTime() const
	{
		return m_havingABadTimeWatchpoint->hasBeenInvalidated();
	}

	void haveABadTime(JSGlobalData&);

    bool objectPrototypeIsSane();
    bool arrayPrototypeChainIsSane();
    bool stringPrototypeChainIsSane();

	void setProfileGroup(unsigned value) { createRareDataIfNeeded(); m_rareData->profileGroup = value; }
	unsigned profileGroup() const
	{
		if (!m_rareData)
			return 0;
		return m_rareData->profileGroup;
	}

	Debugger* debugger() const { return m_debugger; }
	void setDebugger(Debugger* debugger) { m_debugger = debugger; }

	const GlobalObjectMethodTable* globalObjectMethodTable() const { return m_globalObjectMethodTable; }

	static bool allowsAccessFrom(const JSGlobalObject*, ExecState*) { return true; }
	static bool supportsProfiling(const JSGlobalObject*) { return false; }
	static bool supportsRichSourceInfo(const JSGlobalObject*) { return true; }

	JS_EXPORT_PRIVATE ExecState* globalExec();

	static bool shouldInterruptScript(const JSGlobalObject*) { return true; }
	static bool javaScriptExperimentsEnabled(const JSGlobalObject*) { return false; }

    void queueMicrotask(PassRefPtr<Microtask>);

	void setEvalEnabled(bool enabled) { m_evalEnabled = enabled; }
	bool evalEnabled() { return m_evalEnabled; }

	void resetPrototype(JSGlobalData&, JSValue prototype);

	JSGlobalData& globalData() const { return m_vm; }
	JSGlobalData& vm() const { return globalData(); }
	JSObject* globalThis() const;

	static Structure* createStructure(JSGlobalData& globalData, JSValue prototype)
	{
		Structure* result = Structure::create(globalData, 0, prototype, TypeInfo(GlobalObjectType, StructureFlags), &s_info);
		result->setTransitionWatchpointIsLikelyToBeFired(true);
		return result;
	}

	void registerWeakMap(OpaqueJSWeakObjectMap* map)
	{
		createRareDataIfNeeded();
		m_rareData->weakMaps.add(map);
	}

	void unregisterWeakMap(OpaqueJSWeakObjectMap* map)
	{
		if (m_rareData)
			m_rareData->weakMaps.remove(map);
	}

    OpaqueJSClassDataMap& opaqueJSClassData()
    {
        createRareDataIfNeeded();
        return m_rareData->opaqueJSClassData;
    }

    static ptrdiff_t weakRandomOffset() { return OBJECT_OFFSETOF(JSGlobalObject, m_weakRandom); }
	double weakRandomNumber() { return m_weakRandom.get(); }
	unsigned weakRandomInteger() { return m_weakRandom.getUint32(); }

    BuiltinExecutables* builtinExecutables() { return m_builtinExecutables.get(); }

protected:

	static const unsigned StructureFlags = OverridesGetOwnPropertySlot | OverridesGetPropertyNames | Base::StructureFlags;

	struct GlobalPropertyInfo {
		GlobalPropertyInfo(const Identifier& i, JSValue v, unsigned a)
			: identifier(i)
			, value(v)
			, attributes(a)
		{
		}

		const Identifier identifier;
		JSValue value;
		unsigned attributes;
	};
	JS_EXPORT_PRIVATE void addStaticGlobals(GlobalPropertyInfo*, int count);

    JS_EXPORT_PRIVATE static JSC::JSValue toThis(JSC::JSCell*, JSC::ExecState*, ECMAMode);

private:
	friend class LLIntOffsetsExtractor;

    JS_EXPORT_PRIVATE void setGlobalThis(VM&, JSObject* globalThis);

	// FIXME: Fold reset into init.
	JS_EXPORT_PRIVATE void init(JSGlobalData&);
	void reset(JSValue prototype);

	void createThrowTypeError(VM&);

	JS_EXPORT_PRIVATE static void clearRareData(JSCell*);
};

JSGlobalObject* asGlobalObject(JSValue);

ALWAYS_INLINE JSGlobalObject* asGlobalObject(JSValue value)
{
	ASSERT(asObject(value)->isGlobalObject());
	return jsCast<JSGlobalObject*>(asObject(value));
}

inline bool JSGlobalObject::hasOwnPropertyForWrite(ExecState* exec, PropertyName propertyName)
{
	PropertySlot slot(this);
	if (Base::getOwnPropertySlot(this, exec, propertyName, slot))
		return true;
	bool slotIsWriteable;
	return symbolTableGet(this, propertyName, slot, slotIsWriteable);
}

inline bool JSGlobalObject::symbolTableHasProperty(PropertyName propertyName)
{
	SymbolTableEntry entry = symbolTable()->inlineGet(propertyName.uid());
	return !entry.isNull();
}

inline JSGlobalObject* ExecState::dynamicGlobalObject()
{
	if (this == lexicalGlobalObject()->globalExec())
		return lexicalGlobalObject();

	// For any ExecState that's not a globalExec, the
	// dynamic global object must be set since code is running
	ASSERT(globalData().dynamicGlobalObject);
	return globalData().dynamicGlobalObject;
}

inline JSArray* constructEmptyArray(ExecState* exec, ArrayAllocationProfile* profile, JSGlobalObject* globalObject, unsigned initialLength = 0)
{
	return ArrayAllocationProfile::updateLastAllocationFor(profile, JSArray::create(exec->globalData(), initialLength >= MIN_ARRAY_STORAGE_CONSTRUCTION_LENGTH ? globalObject->arrayStructureForIndexingTypeDuringAllocation(ArrayWithArrayStorage) : globalObject->arrayStructureForProfileDuringAllocation(profile), initialLength));
}

inline JSArray* constructEmptyArray(ExecState* exec, ArrayAllocationProfile* profile, unsigned initialLength = 0)
{
	return constructEmptyArray(exec, profile, exec->lexicalGlobalObject(), initialLength);
}

inline JSArray* constructArray(ExecState* exec, ArrayAllocationProfile* profile, JSGlobalObject* globalObject, const ArgList& values)
{
	return ArrayAllocationProfile::updateLastAllocationFor(profile, constructArray(exec, globalObject->arrayStructureForProfileDuringAllocation(profile), values));
}

inline JSArray* constructArray(ExecState* exec, ArrayAllocationProfile* profile, const ArgList& values)
{
	return constructArray(exec, profile, exec->lexicalGlobalObject(), values);
}

inline JSArray* constructArray(ExecState* exec, ArrayAllocationProfile* profile, JSGlobalObject* globalObject, const JSValue* values, unsigned length)
{
	return ArrayAllocationProfile::updateLastAllocationFor(profile, constructArray(exec, globalObject->arrayStructureForProfileDuringAllocation(profile), values, length));
}

inline JSArray* constructArray(ExecState* exec, ArrayAllocationProfile* profile, const JSValue* values, unsigned length)
{
	return constructArray(exec, profile, exec->lexicalGlobalObject(), values, length);
}

inline JSArray* constructArrayNegativeIndexed(ExecState* exec, ArrayAllocationProfile* profile, JSGlobalObject* globalObject, const JSValue* values, unsigned length)
{
    return ArrayAllocationProfile::updateLastAllocationFor(profile, constructArrayNegativeIndexed(exec, globalObject->arrayStructureForProfileDuringAllocation(profile), values, length));
}

inline JSArray* constructArrayNegativeIndexed(ExecState* exec, ArrayAllocationProfile* profile, const JSValue* values, unsigned length)
{
    return constructArrayNegativeIndexed(exec, profile, exec->lexicalGlobalObject(), values, length);
}

class DynamicGlobalObjectScope {
	WTF_MAKE_NONCOPYABLE(DynamicGlobalObjectScope);
public:
	JS_EXPORT_PRIVATE DynamicGlobalObjectScope(JSGlobalData&, JSGlobalObject*);

	~DynamicGlobalObjectScope()
	{
		m_dynamicGlobalObjectSlot = m_savedDynamicGlobalObject;
	}

private:
	JSGlobalObject*& m_dynamicGlobalObjectSlot;
	JSGlobalObject* m_savedDynamicGlobalObject;
};

ALWAYS_INLINE JSObject* JSScope::globalThis()
{
	return globalObject()->globalThis();
}

ALWAYS_INLINE JSObject* JSGlobalObject::globalThis() const
{
	return m_globalThis.get();
}

} // namespace JSC

#endif // JSGlobalObject_h
