/*
 *  Copyright (C) 1999-2001 Harri Porten (porten@kde.org)
 *  Copyright (C) 2001 Peter Kelly (pmk@post.com)
 *  Copyright (C) 2003, 2004, 2005, 2006, 2007, 2008 Apple Inc. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public License
 *  along with this library; see the file COPYING.LIB.  If not, write to
 *  the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 *  Boston, MA 02110-1301, USA.
 *
 */

#ifndef JSString_h
#define JSString_h
#include "CallFrame.h"
#include "CommonIdentifiers.h"
#include "Identifier.h"
#include "PropertyDescriptor.h"
#include "PropertySlot.h"
#include "Structure.h"
#include <wtf/text/StringView.h>

namespace JSC {

class JSString;
class JSRopeString;
class LLIntOffsetsExtractor;

JSString* jsEmptyString(JSGlobalData*);
JSString* jsEmptyString(ExecState*);
JSString* jsString(JSGlobalData*, const String&); // returns empty string if passed null string
JSString* jsString(ExecState*, const String&); // returns empty string if passed null string

JSString* jsSingleCharacterString(JSGlobalData*, UChar);
JSString* jsSingleCharacterString(ExecState*, UChar);
JSString* jsSubstring(JSGlobalData*, const String&, unsigned offset, unsigned length);
JSString* jsSubstring(ExecState*, const String&, unsigned offset, unsigned length);
JSString* jsSubstring8(JSGlobalData*, const String&, unsigned offset, unsigned length);
JSString* jsSubstring8(ExecState*, const String&, unsigned offset, unsigned length);

// Non-trivial strings are two or more characters long.
// These functions are faster than just calling jsString.
JSString* jsNontrivialString(JSGlobalData*, const String&);
JSString* jsNontrivialString(ExecState*, const String&);
JSString* jsNontrivialString(JSGlobalData*, const char*);
JSString* jsNontrivialString(ExecState*, const char*);

// Should be used for strings that are owned by an object that will
// likely outlive the JSValue this makes, such as the parse tree or a
// DOM object that contains a String
JSString* jsOwnedString(JSGlobalData*, const String&);
JSString* jsOwnedString(ExecState*, const String&);

JSRopeString* jsStringBuilder(JSGlobalData*);

bool isJSString(JSValue);
JSString* asString(JSValue);

struct StringViewWithUnderlyingString {
	StringViewWithUnderlyingString(StringView v, String& u)
	: view(v)
	, underlyingString(u)
	{
	}

    StringView view;
    String underlyingString;
};

class JSString : public JSCell {
public:
	friend class JIT;
	friend class JSGlobalData;
	friend class SpecializedThunkJIT;
	friend class JSRopeString;
	friend class MarkStack;
	friend class SlotVisitor;
	friend struct ThunkHelpers;

	typedef JSCell Base;

	static const bool needsDestruction = true;
	static const bool hasImmortalStructure = true;
	static void destroy(JSCell*);

private:
	ALWAYS_INLINE JSString(JSGlobalData& globalData, PassRefPtr<StringImpl> value)
		: JSCell(globalData, globalData.stringStructure.get())
		, m_flags(0)
		, m_value(value)
	{
	}

	ALWAYS_INLINE JSString(JSGlobalData& globalData)
		: JSCell(globalData, globalData.stringStructure.get())
		, m_flags(0)
	{
	}

	ALWAYS_INLINE void finishCreation(JSGlobalData& globalData, size_t length)
	{
		ASSERT(!m_value.isNull());
		Base::finishCreation(globalData);
		m_length = length;
		setIs8Bit(m_value.impl()->is8Bit());
	}

	ALWAYS_INLINE void finishCreation(JSGlobalData& globalData, size_t length, size_t cost)
	{
		ASSERT(!m_value.isNull());
		Base::finishCreation(globalData);
		m_length = length;
		setIs8Bit(m_value.impl()->is8Bit());
		Heap::heap(this)->reportExtraMemoryAllocated(cost);
	}

protected:
	ALWAYS_INLINE void finishCreation(JSGlobalData& globalData)
	{
		Base::finishCreation(globalData);
		m_length = 0;
		setIs8Bit(true);
	}

public:
	static ALWAYS_INLINE JSString* create(JSGlobalData& globalData, PassRefPtr<StringImpl> value)
	{
		ASSERT(value);
		unsigned length = value->length();
        RELEASE_ASSERT(length >= 0);
		size_t cost = value->cost();
		JSString* newString = new (NotNull, allocateCell<JSString>(globalData.heap)) JSString(globalData, value);
		newString->finishCreation(globalData, length, cost);
		return newString;
	}
	static ALWAYS_INLINE JSString* createHasOtherOwner(JSGlobalData& globalData, PassRefPtr<StringImpl> value)
	{
		ASSERT(value);
		size_t length = value->length();
		JSString* newString = new (NotNull, allocateCell<JSString>(globalData.heap)) JSString(globalData, value);
		newString->finishCreation(globalData, length);
		return newString;
	}

    Identifier toIdentifier(ExecState*) const;
    AtomicString toAtomicString(ExecState*) const;
    RefPtr<AtomicStringImpl> toExistingAtomicString(ExecState*) const;

    StringViewWithUnderlyingString viewWithUnderlyingString(ExecState&) const;

    inline bool equal(ExecState*, JSString* other) const;
	const String& value(ExecState*) const;
	const String& value(ExecState&) const;
	const String& tryGetValue() const;
    const StringImpl* tryGetValueImpl() const;
	unsigned length() { return m_length; }

	JSValue toPrimitive(ExecState*, PreferredPrimitiveType) const;
	JS_EXPORT_PRIVATE bool toBoolean() const;
	bool getPrimitiveNumber(ExecState*, double& number, JSValue&) const;
	JSObject* toObject(ExecState*, JSGlobalObject*) const;
	double toNumber(ExecState*) const;

	bool getStringPropertySlot(ExecState*, PropertyName, PropertySlot&);
	bool getStringPropertySlot(ExecState*, unsigned propertyName, PropertySlot&);
	bool getStringPropertyDescriptor(ExecState*, PropertyName, PropertyDescriptor&);

	ALWAYS_INLINE bool canGetIndex(unsigned i) { return i < m_length; }
	JSString* getIndex(ExecState*, unsigned);

	static Structure* createStructure(JSGlobalData& globalData, JSGlobalObject* globalObject, JSValue proto)
	{
		return Structure::create(globalData, globalObject, proto, TypeInfo(StringType, OverridesGetOwnPropertySlot | InterceptsGetOwnPropertySlotByIndexEvenWhenLengthIsNotZero), &s_info);
	}

	static size_t offsetOfLength() { return OBJECT_OFFSETOF(JSString, m_length); }
	static size_t offsetOfFlags() { return OBJECT_OFFSETOF(JSString, m_flags); }
	static size_t offsetOfValue() { return OBJECT_OFFSETOF(JSString, m_value); }

	static JS_EXPORTDATA const ClassInfo s_info;

	static void visitChildren(JSCell*, SlotVisitor&);

    enum {
        Is8Bit = 1u
    };

protected:
    friend class JSValue;

    JS_EXPORT_PRIVATE bool equalSlowCase(ExecState*, JSString* other) const;
	ALWAYS_INLINE bool isRope() const { return m_value.isNull(); }
    bool isSubstring() const;
	ALWAYS_INLINE bool is8Bit() const { return m_flags & Is8Bit; }
	ALWAYS_INLINE void setIs8Bit(bool flag) const
	{
		if (flag)
			m_flags |= Is8Bit;
		else
			m_flags &= ~Is8Bit;
	}

    mutable unsigned m_flags;

	// A string is represented either by a String or a rope of fibers.
	unsigned m_length;
	mutable String m_value;

private:
	friend class LLIntOffsetsExtractor;

    static JSValue toThis(JSCell*, ExecState*, ECMAMode);

	ALWAYS_INLINE String& string() { ASSERT(!isRope()); return m_value; }
    StringView unsafeView(ExecState&) const;

	friend JSValue jsString(ExecState*, JSString*, JSString*);
	friend JSString* jsSubstring(ExecState*, JSString*, unsigned offset, unsigned length);
};

class JSRopeString : public JSString {
	friend class JSString;

	friend JSRopeString* jsStringBuilder(JSGlobalData*);

	class RopeBuilder {
	public:
		ALWAYS_INLINE RopeBuilder(JSGlobalData& globalData)
		: m_globalData(globalData)
		, m_jsString(jsStringBuilder(&globalData))
		, m_index(0)
		{
		}

		ALWAYS_INLINE bool append(JSString* jsString)
		{
			if (m_index == JSRopeString::s_maxInternalRopeLength)
				expand();
            if (static_cast<int32_t>(m_jsString->length() + jsString->length()) < 0) {
                m_jsString = 0;
                return false;
            }
			m_jsString->append(m_globalData, m_index++, jsString);
            return true;
		}

		ALWAYS_INLINE JSRopeString* release()
		{
            RELEASE_ASSERT(m_jsString);
			JSRopeString* tmp = m_jsString;
			m_jsString = 0;
			return tmp;
		}

		ALWAYS_INLINE unsigned length() { return m_jsString->m_length; }

	private:
		void expand();

		JSGlobalData& m_globalData;
		JSRopeString* m_jsString;
		size_t m_index;
	};

private:
	ALWAYS_INLINE JSRopeString(JSGlobalData& globalData)
		: JSString(globalData)
	{
	}

	ALWAYS_INLINE void finishCreation(JSGlobalData& globalData, JSString* s1, JSString* s2)
	{
		Base::finishCreation(globalData);
		m_length = s1->length() + s2->length();
		setIs8Bit(s1->is8Bit() && s2->is8Bit());
		setIsSubstring(false);
        fiber(0).set(globalData, this, s1);
        fiber(1).set(globalData, this, s2);
        fiber(2).clear();
	}

	ALWAYS_INLINE void finishCreation(JSGlobalData& globalData, JSString* s1, JSString* s2, JSString* s3)
	{
		Base::finishCreation(globalData);
		m_length = s1->length() + s2->length() + s3->length();
		setIs8Bit(s1->is8Bit() && s2->is8Bit() &&  s3->is8Bit());
		setIsSubstring(false);
		fiber(0).set(globalData, this, s1);
		fiber(1).set(globalData, this, s2);
		fiber(2).set(globalData, this, s3);
	}

	ALWAYS_INLINE void finishCreation(JSGlobalData& vm, ExecState* exec, JSString* base, unsigned offset, unsigned length)
	{
        Base::finishCreation(vm);
        ASSERT(!sumOverflows<int32_t>(offset, length));
        ASSERT(offset + length <= base->length());
        m_length = length;
        setIs8Bit(base->is8Bit());
        setIsSubstring(true);
        if (base->isSubstring()) {
            JSRopeString* baseRope = jsCast<JSRopeString*>(base);
            substringBase().set(vm, this, baseRope->substringBase().get());
            substringOffset() = baseRope->substringOffset() + offset;
        } else {
            substringBase().set(vm, this, base);
            substringOffset() = offset;

            // For now, let's not allow substrings with a rope base.
            // Resolve non-substring rope bases so we don't have to deal with it.
            // FIXME: Evaluate if this would be worth adding more branches.
            if (base->isRope())
            	jsCast<JSRopeString*>(base)->resolveRope(exec);
        }
	}

    ALWAYS_INLINE void finishCreationSubstringOfResolved(VM& vm, JSString* base, unsigned offset, unsigned length)
    {
        Base::finishCreation(vm);
        ASSERT(!sumOverflows<int32_t>(offset, length));
        ASSERT(offset + length <= base->length());
        m_length = length;
        setIs8Bit(base->is8Bit());
        setIsSubstring(true);
        substringBase().set(vm, this, base);
        substringOffset() = offset;
    }

	ALWAYS_INLINE void finishCreation(JSGlobalData& globalData)
	{
		JSString::finishCreation(globalData);
		setIsSubstring(false);
		fiber(0).clear();
		fiber(1).clear();
		fiber(2).clear();
	}

	void append(JSGlobalData& globalData, size_t index, JSString* jsString)
	{
		fiber(index).set(globalData, this, jsString);
		m_length += jsString->m_length;
        RELEASE_ASSERT(static_cast<int32_t>(m_length) >= 0);
		setIs8Bit(is8Bit() && jsString->is8Bit());
	}

	static ALWAYS_INLINE JSRopeString* createNull(JSGlobalData& globalData)
	{
		JSRopeString* newString = new (NotNull, allocateCell<JSRopeString>(globalData.heap)) JSRopeString(globalData);
		newString->finishCreation(globalData);
		return newString;
	}

public:
	static ALWAYS_INLINE JSString* create(JSGlobalData& globalData, JSString* s1, JSString* s2)
	{
		JSRopeString* newString = new (NotNull, allocateCell<JSRopeString>(globalData.heap)) JSRopeString(globalData);
		newString->finishCreation(globalData, s1, s2);
		return newString;
	}
	static ALWAYS_INLINE JSString* create(JSGlobalData& globalData, JSString* s1, JSString* s2, JSString* s3)
	{
		JSRopeString* newString = new (NotNull, allocateCell<JSRopeString>(globalData.heap)) JSRopeString(globalData);
		newString->finishCreation(globalData, s1, s2, s3);
		return newString;
	}

	ALWAYS_INLINE static JSString* create(JSGlobalData& vm, ExecState* exec, JSString* base, unsigned offset, unsigned length)
    {
        JSRopeString* newString = new (NotNull, allocateCell<JSRopeString>(vm.heap)) JSRopeString(vm);
        newString->finishCreation(vm, exec, base, offset, length);
        return newString;
    }

    ALWAYS_INLINE static JSString* createSubstringOfResolved(VM& vm, JSString* base, unsigned offset, unsigned length)
    {
        JSRopeString* newString = new (NotNull, allocateCell<JSRopeString>(vm.heap)) JSRopeString(vm);
        newString->finishCreationSubstringOfResolved(vm, base, offset, length);
        return newString;
    }

	void visitFibers(SlotVisitor&);

    static ptrdiff_t offsetOfFibers() { return OBJECT_OFFSETOF(JSRopeString, u); }

    static const unsigned s_maxInternalRopeLength = 3;

private:
	friend JSValue jsStringFromRegisterArray(ExecState*, Register*, unsigned);
	friend JSValue jsStringFromArguments(ExecState*, JSValue);

	JS_EXPORT_PRIVATE void resolveRope(ExecState*) const;
	JS_EXPORT_PRIVATE void resolveRopeToAtomicString(ExecState*) const;
	JS_EXPORT_PRIVATE RefPtr<AtomicStringImpl> resolveRopeToExistingAtomicString(ExecState*) const;
	void resolveRopeSlowCase8(LChar*) const;
	void resolveRopeSlowCase(UChar*) const;
	void outOfMemory(ExecState*) const;
    void resolveRopeInternal8(LChar*) const;
    void resolveRopeInternal8NoSubstring(LChar*) const;
    void resolveRopeInternal16(UChar*) const;
    void resolveRopeInternal16NoSubstring(UChar*) const;
    void clearFibers() const;
    StringView unsafeView(ExecState&) const;
    StringViewWithUnderlyingString viewWithUnderlyingString(ExecState&) const;

    JS_EXPORT_PRIVATE JSString* getIndexSlowCase(ExecState*, unsigned);

    ALWAYS_INLINE WriteBarrierBase<JSString>& fiber(unsigned i) const
    {
        ASSERT(!isSubstring());
        ASSERT(i < s_maxInternalRopeLength);
        return u[i].string;
    }

    ALWAYS_INLINE WriteBarrierBase<JSString>& substringBase() const
    {
        return u[1].string;
    }

    ALWAYS_INLINE uintptr_t& substringOffset() const
    {
        return u[2].number;
    }

    static uintptr_t notSubstringSentinel()
    {
        return 0;
    }

    static uintptr_t substringSentinel()
    {
        return 1;
    }

    ALWAYS_INLINE bool isSubstring() const
    {
        return u[0].number == substringSentinel();
    }

    ALWAYS_INLINE void setIsSubstring(bool isSubstring)
    {
        u[0].number = isSubstring ? substringSentinel() : notSubstringSentinel();
    }

    mutable union {
        uintptr_t number;
        WriteBarrierBase<JSString> string;
    } u[s_maxInternalRopeLength];
};

JS_EXPORT_PRIVATE JSString* jsStringWithCacheSlowCase(VM&, StringImpl&);

ALWAYS_INLINE const StringImpl* JSString::tryGetValueImpl() const
{
    return m_value.impl();
}

ALWAYS_INLINE JSString* asString(JSValue value)
{
	ASSERT(value.asCell()->isString());
	return jsCast<JSString*>(value.asCell());
}

ALWAYS_INLINE JSString* jsEmptyString(JSGlobalData& globalData)
{
	return globalData.smallStrings.emptyString();
}

ALWAYS_INLINE JSString* jsEmptyString(JSGlobalData* globalData)
{
	return jsEmptyString(*globalData);
}

ALWAYS_INLINE JSString* jsSingleCharacterString(JSGlobalData* globalData, UChar c)
{
	if (c <= maxSingleCharacterString)
		return globalData->smallStrings.singleCharacterString(c);
	return JSString::create(*globalData, String(&c, 1).impl());
}

ALWAYS_INLINE JSString* jsNontrivialString(JSGlobalData* globalData, const char* s)
{
	ASSERT(s);
	ASSERT(s[0]);
	ASSERT(s[1]);
	return JSString::create(*globalData, String(s).impl());
}

ALWAYS_INLINE JSString* jsNontrivialString(JSGlobalData* globalData, const String& s)
{
	ASSERT(s.length() > 1);
	return JSString::create(*globalData, s.impl());
}

ALWAYS_INLINE Identifier JSString::toIdentifier(ExecState* exec) const
{
    return Identifier::fromString(exec, toAtomicString(exec));
}

ALWAYS_INLINE AtomicString JSString::toAtomicString(ExecState* exec) const
{
    if (isRope())
        static_cast<const JSRopeString*>(this)->resolveRopeToAtomicString(exec);
    return AtomicString(m_value);
}

ALWAYS_INLINE RefPtr<AtomicStringImpl> JSString::toExistingAtomicString(ExecState* exec) const
{
    if (isRope())
        return static_cast<const JSRopeString*>(this)->resolveRopeToExistingAtomicString(exec);
    if (m_value.impl()->isAtomic())
        return static_cast<AtomicStringImpl*>(m_value.impl());
    return AtomicString::find(m_value.impl());
}

ALWAYS_INLINE const String& JSString::value(ExecState* exec) const
{
	if (isRope())
		static_cast<const JSRopeString*>(this)->resolveRope(exec);
	return m_value;
}

ALWAYS_INLINE const String& JSString::value(ExecState& exec) const
{
	return value(&exec);
}

ALWAYS_INLINE const String& JSString::tryGetValue() const
{
	if (isRope())
		static_cast<const JSRopeString*>(this)->resolveRope(0);
	return m_value;
}

ALWAYS_INLINE JSString* JSString::getIndex(ExecState* exec, unsigned i)
{
	ASSERT(canGetIndex(i));
	if (isRope())
		return static_cast<JSRopeString*>(this)->getIndexSlowCase(exec, i);
	ASSERT(i < m_value.length());
    return jsSingleCharacterString(exec, m_value[i]);
}

ALWAYS_INLINE JSString* jsString(JSGlobalData& vm, const String& s)
{
	int size = s.length();
	if (!size)
		return vm.smallStrings.emptyString();
	if (size == 1) {
        UChar c = s.uncheckedCharacterAt(0);
		if (c <= maxSingleCharacterString)
			return vm.smallStrings.singleCharacterString(c);
	}
	return JSString::create(vm, s.impl());
}

ALWAYS_INLINE JSString* jsString(JSGlobalData* globalData, const String& s)
{
	return jsString(*globalData, s);
}

ALWAYS_INLINE JSString* jsSubstring(JSGlobalData& vm, ExecState* exec, JSString* s, unsigned offset, unsigned length)
{
	ASSERT(offset <= static_cast<unsigned>(s->length()));
	ASSERT(length <= static_cast<unsigned>(s->length()));
	ASSERT(offset + length <= static_cast<unsigned>(s->length()));
	if (!length)
		return vm.smallStrings.emptyString();
    if (!offset && length == s->length())
        return s;
    return JSRopeString::create(vm, exec, s, offset, length);
}

inline JSString* jsSubstringOfResolved(VM& vm, JSString* s, unsigned offset, unsigned length)
{
    ASSERT(offset <= static_cast<unsigned>(s->length()));
    ASSERT(length <= static_cast<unsigned>(s->length()));
    ASSERT(offset + length <= static_cast<unsigned>(s->length()));
    if (!length)
        return vm.smallStrings.emptyString();
    if (!offset && length == s->length())
        return s;
    return JSRopeString::createSubstringOfResolved(vm, s, offset, length);
}

inline JSString* jsSubstring(ExecState* exec, JSString* s, unsigned offset, unsigned length)
{
    return jsSubstring(exec->vm(), exec, s, offset, length);
}

ALWAYS_INLINE JSString* jsSubstring8(JSGlobalData* globalData, const String& s, unsigned offset, unsigned length)
{
	ASSERT(offset <= static_cast<unsigned>(s.length()));
	ASSERT(length <= static_cast<unsigned>(s.length()));
	ASSERT(offset + length <= static_cast<unsigned>(s.length()));
	if (!length)
		return globalData->smallStrings.emptyString();
	if (length == 1) {
        UChar c = s.characterAt(offset);
		if (c <= maxSingleCharacterString)
			return globalData->smallStrings.singleCharacterString(c);
	}
	return JSString::createHasOtherOwner(*globalData, StringImpl::create8(s.impl(), offset, length));
}

ALWAYS_INLINE JSString* jsSubstring(JSGlobalData* globalData, const String& s, unsigned offset, unsigned length)
{
	ASSERT(offset <= static_cast<unsigned>(s.length()));
	ASSERT(length <= static_cast<unsigned>(s.length()));
	ASSERT(offset + length <= static_cast<unsigned>(s.length()));
	if (!length)
		return globalData->smallStrings.emptyString();
	if (length == 1) {
        UChar c = s.characterAt(offset);
		if (c <= maxSingleCharacterString)
			return globalData->smallStrings.singleCharacterString(c);
	}
	return JSString::createHasOtherOwner(*globalData, StringImpl::create(s.impl(), offset, length));
}

ALWAYS_INLINE JSString* jsOwnedString(JSGlobalData* globalData, const String& s)
{
	int size = s.length();
	if (!size)
		return globalData->smallStrings.emptyString();
	if (size == 1) {
        UChar c = s.uncheckedCharacterAt(0);
		if (c <= maxSingleCharacterString)
			return globalData->smallStrings.singleCharacterString(c);
	}
	return JSString::createHasOtherOwner(*globalData, s.impl());
}

ALWAYS_INLINE JSRopeString* jsStringBuilder(JSGlobalData* globalData)
{
	return JSRopeString::createNull(*globalData);
}

ALWAYS_INLINE JSString* jsEmptyString(ExecState* exec) { return jsEmptyString(&exec->globalData()); }
ALWAYS_INLINE JSString* jsString(ExecState* exec, const String& s) { return jsString(&exec->globalData(), s); }
ALWAYS_INLINE JSString* jsSingleCharacterString(ExecState* exec, UChar c) { return jsSingleCharacterString(&exec->globalData(), c); }
ALWAYS_INLINE JSString* jsSubstring8(ExecState* exec, const String& s, unsigned offset, unsigned length) { return jsSubstring8(&exec->globalData(), s, offset, length); }
ALWAYS_INLINE JSString* jsSubstring(ExecState* exec, const String& s, unsigned offset, unsigned length) { return jsSubstring(&exec->globalData(), s, offset, length); }
ALWAYS_INLINE JSString* jsNontrivialString(ExecState* exec, const String& s) { return jsNontrivialString(&exec->globalData(), s); }
ALWAYS_INLINE JSString* jsNontrivialString(ExecState* exec, const char* s) { return jsNontrivialString(&exec->globalData(), s); }
ALWAYS_INLINE JSString* jsOwnedString(ExecState* exec, const String& s) { return jsOwnedString(&exec->globalData(), s); }

ALWAYS_INLINE bool JSString::getStringPropertySlot(ExecState* exec, PropertyName propertyName, PropertySlot& slot)
{
	if (propertyName == exec->propertyNames().length) {
		slot.setValue(this, DontEnum | DontDelete | ReadOnly, jsNumber(m_length));
		return true;
	}

    Optional<uint32_t> index = parseIndex(propertyName);
    if (index && index.value() < m_length) {
		slot.setValue(this, DontDelete | ReadOnly, getIndex(exec, index.value()));
		return true;
	}

	return false;
}

ALWAYS_INLINE bool JSString::getStringPropertySlot(ExecState* exec, unsigned propertyName, PropertySlot& slot)
{
	if (propertyName < m_length) {
		slot.setValue(this, DontDelete | ReadOnly, getIndex(exec, propertyName));
		return true;
	}

	return false;
}

ALWAYS_INLINE bool isJSString(const JSCell* cell)
{
	return cell->structure()->typeInfo().type() == StringType;
}

ALWAYS_INLINE bool isJSString(JSValue v) { return v.isCell() && isJSString(v.asCell()); }

ALWAYS_INLINE bool JSString::toBoolean() const
{
	return m_length;
}

ALWAYS_INLINE JSValue JSString::toPrimitive(ExecState*, PreferredPrimitiveType) const
{
	return const_cast<JSString*>(this);
}

ALWAYS_INLINE StringView JSRopeString::unsafeView(ExecState& state) const
{
    if (isSubstring()) {
        if (is8Bit())
            return StringView(substringBase()->m_value.characters8() + substringOffset(), m_length);
        return StringView(substringBase()->m_value.characters16() + substringOffset(), m_length);
    }
    resolveRope(&state);
    return m_value;
}

ALWAYS_INLINE StringViewWithUnderlyingString JSRopeString::viewWithUnderlyingString(ExecState& state) const
{
    if (isSubstring()) {
        String& base = substringBase()->m_value;
        if (is8Bit())
            return StringViewWithUnderlyingString(StringView(base.characters8() + substringOffset(), m_length), base);
        return StringViewWithUnderlyingString(StringView(base.characters16() + substringOffset(), m_length), base);
    }
    resolveRope(&state);
    return StringViewWithUnderlyingString(m_value, m_value);
}

ALWAYS_INLINE StringView JSString::unsafeView(ExecState& state) const
{
    if (isRope())
        return static_cast<const JSRopeString*>(this)->unsafeView(state);
    return m_value;
}

ALWAYS_INLINE StringViewWithUnderlyingString JSString::viewWithUnderlyingString(ExecState& state) const
{
    if (isRope())
        return static_cast<const JSRopeString&>(*this).viewWithUnderlyingString(state);
    return StringViewWithUnderlyingString(m_value, m_value);
}

ALWAYS_INLINE bool JSString::isSubstring() const
{
    return isRope() && static_cast<const JSRopeString*>(this)->isSubstring();
}

// --- JSValue inlines ----------------------------

inline bool JSValue::toBoolean(ExecState* exec) const
{
	if (isInt32())
		return asInt32();
	if (isDouble())
		return asDouble() > 0.0 || asDouble() < 0.0; // false for NaN
	if (isCell())
		return asCell()->toBoolean(exec);
	return isTrue(); // false, null, and undefined all convert to false.
}

inline JSString* JSValue::toString(ExecState* exec) const
{
	if (isString())
		return jsCast<JSString*>(asCell());
    bool returnEmptyStringOnError = true;
    return toStringSlowCase(exec, returnEmptyStringOnError);
}

inline JSString* JSValue::toStringOrNull(ExecState* exec) const
{
    if (isString())
        return jsCast<JSString*>(asCell());
    bool returnEmptyStringOnError = false;
    return toStringSlowCase(exec, returnEmptyStringOnError);
}

inline const String& JSValue::toWTFString(ExecState* exec) const
{
	if (isString())
		return static_cast<JSString*>(asCell())->value(exec);
	return toWTFStringSlowCase(exec);
}

} // namespace JSC

#endif // JSString_h
