/*
 * Copyright (C) 2008, 2009, 2012 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE COMPUTER, INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE COMPUTER, INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#ifndef Structure_h
#define Structure_h

#include "ClassInfo.h"
#include "DeferGC.h"
#include "IndexingType.h"
#include "JSCell.h"
#include "JSType.h"
#include "JSValue.h"
#include "PropertyName.h"
#include "PropertyNameArray.h"
#include "PropertyOffset.h"
#include "Protect.h"
#include "StructureRareData.h"
#include "PutPropertySlot.h"
#include "StructureTransitionTable.h"
#include "JSTypeInfo.h"
#include "Watchpoint.h"
#include "Weak.h"
#include <wtf/CompilationThread.h>
#include <wtf/PassRefPtr.h>
#include <wtf/RefCounted.h>
#include <wtf/text/StringImpl.h>

namespace JSC {

class LLIntOffsetsExtractor;
class PropertyNameArray;
class PropertyNameArrayData;
class PropertyTable;
class StructureChain;
class SlotVisitor;
class JSString;
struct DumpContext;

// The out-of-line property storage capacity to use when first allocating out-of-line
// storage. Note that all objects start out without having any out-of-line storage;
// this comes into play only on the first property store that exhausts inline storage.
static const unsigned initialOutOfLineCapacity = 4;

// The factor by which to grow out-of-line storage when it is exhausted, after the
// initial allocation.
static const unsigned outOfLineGrowthFactor = 2;

class StructureFireDetail : public FireDetail {
public:
    StructureFireDetail(const Structure* structure)
        : m_structure(structure)
    {
    }

    virtual void dump(PrintStream& out) const;

private:
    const Structure* m_structure;
};

class DeferredStructureTransitionWatchpointFire {
    WTF_MAKE_NONCOPYABLE(DeferredStructureTransitionWatchpointFire);
public:
    JS_EXPORT_PRIVATE DeferredStructureTransitionWatchpointFire();
    JS_EXPORT_PRIVATE ~DeferredStructureTransitionWatchpointFire();

    void add(const Structure*);

private:
    const Structure* m_structure;
};

class Structure : public JSCell {
public:
	friend class StructureTransitionTable;

	typedef JSCell Base;

	static Structure* create(JSGlobalData&, JSGlobalObject*, JSValue prototype, const TypeInfo&, const ClassInfo*, IndexingType = NonArray, unsigned inlineCapacity = 0);

protected:
	void finishCreation(JSGlobalData& globalData)
	{
		Base::finishCreation(globalData);
		ASSERT(m_prototype);
		ASSERT(m_prototype.isObject() || m_prototype.isNull());
	}

	void finishCreation(JSGlobalData& globalData, CreatingEarlyCellTag)
	{
		Base::finishCreation(globalData, this, CreatingEarlyCell);
		ASSERT(m_prototype);
		ASSERT(m_prototype.isNull());
		ASSERT(!globalData.structureStructure);
	}

public:
    bool isProxy() const
    {
        JSType type = typeInfo().type();
        return type == ImpureProxyType || type == PureForwardingProxyType;
    }

	static void dumpStatistics();

	JS_EXPORT_PRIVATE static Structure* addPropertyTransition(JSGlobalData&, Structure*, PropertyName, unsigned attributes, PropertyOffset&, PutPropertySlot::Context = PutPropertySlot::UnknownContext, DeferredStructureTransitionWatchpointFire* = 0);
	JS_EXPORT_PRIVATE static Structure* addNewPropertyTransition(JSGlobalData&, Structure*, PropertyName, unsigned attributes, PropertyOffset&, PutPropertySlot::Context = PutPropertySlot::UnknownContext, DeferredStructureTransitionWatchpointFire* = 0);
	JS_EXPORT_PRIVATE static Structure* addPropertyTransitionToExistingStructure(Structure*, StringImpl* uid, unsigned attributes, PropertyOffset&);
	static Structure* removePropertyTransition(JSGlobalData&, Structure*, PropertyName, PropertyOffset&);
	JS_EXPORT_PRIVATE static Structure* changePrototypeTransition(JSGlobalData&, Structure*, JSValue prototype);
	JS_EXPORT_PRIVATE static Structure* attributeChangeTransition(JSGlobalData&, Structure*, PropertyName, unsigned attributes);
	static Structure* toCacheableDictionaryTransition(JSGlobalData&, Structure*, DeferredStructureTransitionWatchpointFire* = 0);
	static Structure* toUncacheableDictionaryTransition(JSGlobalData&, Structure*);
	static Structure* sealTransition(JSGlobalData&, Structure*);
	static Structure* freezeTransition(JSGlobalData&, Structure*);
	static Structure* preventExtensionsTransition(JSGlobalData&, Structure*);
	static Structure* nonPropertyTransition(JSGlobalData&, Structure*, NonPropertyTransition);

	bool isSealed(JSGlobalData&);
	bool isFrozen(JSGlobalData&);
	ALWAYS_INLINE bool isExtensible() const { return !m_preventExtensions; }
	ALWAYS_INLINE bool didTransition() const { return m_didTransition; }
	bool putWillGrowOutOfLineStorage();
	JS_EXPORT_PRIVATE size_t suggestedNewOutOfLineStorageCapacity();

	Structure* flattenDictionaryStructure(JSGlobalData&, JSObject*);

	static const bool needsDestruction = true;
	static const bool hasImmortalStructure = true;
	static void destroy(JSCell*);

	// These should be used with caution.
	JS_EXPORT_PRIVATE PropertyOffset addPropertyWithoutTransition(JSGlobalData&, PropertyName, unsigned attributes);
	PropertyOffset removePropertyWithoutTransition(JSGlobalData&, PropertyName);
	void setPrototypeWithoutTransition(JSGlobalData& globalData, JSValue prototype) { m_prototype.set(globalData, this, prototype); }

	ALWAYS_INLINE bool isDictionary() const { return m_dictionaryKind != NoneDictionaryKind; }
	ALWAYS_INLINE bool isUncacheableDictionary() const { return m_dictionaryKind == UncachedDictionaryKind; }

    bool hasBeenFlattenedBefore() const { return m_hasBeenFlattenedBefore; }

	ALWAYS_INLINE bool propertyAccessesAreCacheable() { return m_dictionaryKind != UncachedDictionaryKind && !typeInfo().prohibitsPropertyCaching(); }

    // We use SlowPath in GetByIdStatus for structures that may get new impure properties later to prevent
    // DFG from inlining property accesses since structures don't transition when a new impure property appears.
    bool takesSlowPathInDFGForImpureProperty()
    {
        return typeInfo().hasImpureGetOwnPropertySlot();
    }

	// Type accessors.
	const TypeInfo& typeInfo() const { ASSERT(structure()->classInfo() == &s_info); return m_typeInfo; }
	bool isObject() const { return typeInfo().isObject(); }

	ALWAYS_INLINE IndexingType indexingType() const { return m_indexingType & AllArrayTypes; }
	IndexingType indexingTypeIncludingHistory() const { return m_indexingType; }

	bool mayInterceptIndexedAccesses() const
	{
		return !!(indexingTypeIncludingHistory() & MayHaveIndexedAccessors);
	}

	bool anyObjectInChainMayInterceptIndexedAccesses() const;
    bool holesMustForwardToPrototype(VM&) const;

	bool needsSlowPutIndexing() const;
	NonPropertyTransition suggestedArrayStorageTransition() const;

	ALWAYS_INLINE JSGlobalObject* globalObject() const { return m_globalObject.get(); }

    // NOTE: This method should only be called during the creation of structures, since the global
    // object of a structure is presumed to be immutable in a bunch of places.
	void setGlobalObject(JSGlobalData& globalData, JSGlobalObject* globalObject) { m_globalObject.set(globalData, this, globalObject); }

	ALWAYS_INLINE JSValue storedPrototype() const { return m_prototype.get(); }
    JSObject* storedPrototypeObject() const;
    Structure* storedPrototypeStructure() const;
	JSValue prototypeForLookup(ExecState*) const;
	JSValue prototypeForLookup(JSGlobalObject*) const;
	JSValue prototypeForLookup(CodeBlock*) const;
    StructureChain* prototypeChain(JSGlobalData&, JSGlobalObject*) const;
	StructureChain* prototypeChain(ExecState*) const;
	static void visitChildren(JSCell*, SlotVisitor&);

    // A Structure is cheap to mark during GC if doing so would only add a small and bounded amount
    // to our heap footprint. For example, if the structure refers to a global object that is not
    // yet marked, then as far as we know, the decision to mark this Structure would lead to a large
    // increase in footprint because no other object refers to that global object. This method
    // returns true if all user-controlled (and hence unbounded in size) objects referenced from the
    // Structure are already marked.
    bool isCheapDuringGC();

    // Returns true if this structure is now marked.
    bool markIfCheap(SlotVisitor&);

	// Will just the prototype chain intercept this property access?
    bool prototypeChainMayInterceptStoreTo(VM&, StringImpl* uid);
	bool prototypeChainMayInterceptStoreTo(JSGlobalData&, PropertyName);

	Structure* previousID() const
	{
		ASSERT(structure()->classInfo() == &s_info);
        if (hasRareData())
            return rareData()->previousID();
        return previous();
	}
	bool transitivelyTransitionedFrom(Structure* structureToFind);

    unsigned outOfLineCapacity() const
    {
        ASSERT(checkOffsetConsistency());

        unsigned outOfLineSize = this->outOfLineSize();

        if (!outOfLineSize)
            return 0;

        if (outOfLineSize <= initialOutOfLineCapacity)
            return initialOutOfLineCapacity;

        ASSERT(outOfLineSize > initialOutOfLineCapacity);
        COMPILE_ASSERT(outOfLineGrowthFactor == 2, outOfLineGrowthFactor_is_two);
        return WTF::roundUpToPowerOfTwo(outOfLineSize);
    }
	unsigned outOfLineSize() const
	{
        ASSERT(checkOffsetConsistency());
        ASSERT(structure()->classInfo() == &s_info);

		return numberOfOutOfLineSlotsForLastOffset(m_offset);
	}
	bool hasInlineStorage() const
	{
		return !!m_inlineCapacity;
	}
	unsigned inlineCapacity() const
	{
		return m_inlineCapacity;
	}
	unsigned inlineSize() const
	{
        return std::min<unsigned>(m_offset + 1, m_inlineCapacity);
	}
	unsigned totalStorageSize() const
	{
		return numberOfSlotsForLastOffset(m_offset, m_inlineCapacity);
	}
	unsigned totalStorageCapacity() const
	{
		ASSERT(structure()->classInfo() == &s_info);
		return outOfLineCapacity() + inlineCapacity();
	}

	bool isValidOffset(PropertyOffset offset) const
	{
        return JSC::isValidOffset(offset)
            && offset <= m_offset
            && (offset < m_inlineCapacity || offset >= firstOutOfLineOffset);
    }

    bool hijacksIndexingHeader() const
    {
        return isTypedView(m_classInfo->typedArrayStorageType);
    }

    bool couldHaveIndexingHeader() const
    {
        return hasIndexedProperties(indexingType())
            || hijacksIndexingHeader();
    }

    bool hasIndexingHeader(const JSCell*) const;

	bool masqueradesAsUndefined(JSGlobalObject* lexicalGlobalObject);

	PropertyOffset get(JSGlobalData&, PropertyName);
	PropertyOffset get(JSGlobalData&, StringImpl* uid);
	JS_EXPORT_PRIVATE PropertyOffset get(JSGlobalData&, StringImpl*, unsigned& attributes);
	JS_EXPORT_PRIVATE PropertyOffset get(JSGlobalData&, PropertyName, unsigned& attributes);

    PropertyOffset getConcurrently(VM&, StringImpl* uid);
    PropertyOffset getConcurrently(VM&, StringImpl* uid, unsigned& attributes);

    PropertyOffset getWithoutMaterializing(VM&, PropertyName);
    PropertyOffset getWithoutMaterializing(VM&, PropertyName, unsigned& attributes);

    PropertyOffset getWithoutMaterializing(VM&, StringImpl* uid);
    PropertyOffset getWithoutMaterializing(VM&, StringImpl* uid, unsigned& attributes);

	bool hasGetterSetterProperties() const { return m_hasGetterSetterProperties; }
	bool hasReadOnlyOrGetterSetterPropertiesExcludingProto() const { return m_hasReadOnlyOrGetterSetterPropertiesExcludingProto; }
	void setHasGetterSetterProperties(bool is__proto__)
	{
		m_hasGetterSetterProperties = true;
		if (!is__proto__)
			m_hasReadOnlyOrGetterSetterPropertiesExcludingProto = true;
	}

    bool hasCustomGetterSetterProperties() const { return m_hasCustomGetterSetterProperties; }
    void setHasCustomGetterSetterProperties() { m_hasCustomGetterSetterProperties = true; }

	void setContainsReadOnlyProperties()
	{
		m_hasReadOnlyOrGetterSetterPropertiesExcludingProto = true;
	}

	bool hasNonEnumerableProperties() const { return m_hasNonEnumerableProperties; }

	bool isEmpty() const
	{
        ASSERT(checkOffsetConsistency());
		return !JSC::isValidOffset(m_offset);
	}

	void setEnumerationCache(JSGlobalData&, JSPropertyNameIterator* enumerationCache); // Defined in JSPropertyNameIterator.h.
	JSPropertyNameIterator* enumerationCache(); // Defined in JSPropertyNameIterator.h.
	void getPropertyNamesFromStructure(JSGlobalData&, PropertyNameArray&, EnumerationMode);

    JSString* objectToStringValue()
    {
        if (!hasRareData())
            return 0;
        return rareData()->objectToStringValue();
    }

    void setObjectToStringValue(JSGlobalData& globalData, const JSCell* owner, JSString* value)
    {
        if (!hasRareData())
            allocateRareData(globalData);
        rareData()->setObjectToStringValue(globalData, owner, value);
    }

	bool staticFunctionsReified()
	{
		return m_staticFunctionReified;
	}

	void setStaticFunctionsReified()
	{
		m_staticFunctionReified = true;
	}

	const ClassInfo* classInfo() const { return m_classInfo; }

	static ptrdiff_t prototypeOffset()
	{
		return OBJECT_OFFSETOF(Structure, m_prototype);
	}

	static ptrdiff_t globalObjectOffset()
	{
		return OBJECT_OFFSETOF(Structure, m_globalObject);
	}

	static ptrdiff_t typeInfoFlagsOffset()
	{
		return OBJECT_OFFSETOF(Structure, m_typeInfo) + TypeInfo::flagsOffset();
	}

	static ptrdiff_t typeInfoTypeOffset()
	{
		return OBJECT_OFFSETOF(Structure, m_typeInfo) + TypeInfo::typeOffset();
	}

	static ptrdiff_t classInfoOffset()
	{
		return OBJECT_OFFSETOF(Structure, m_classInfo);
	}

	static ptrdiff_t indexingTypeOffset()
	{
		return OBJECT_OFFSETOF(Structure, m_indexingType);
	}

	static Structure* createStructure(JSGlobalData&);

	bool transitionWatchpointSetHasBeenInvalidated() const
	{
		return m_transitionWatchpointSet.hasBeenInvalidated();
	}

	bool transitionWatchpointSetIsStillValid() const
	{
		return m_transitionWatchpointSet.isStillValid();
	}

    bool dfgShouldWatchIfPossible() const
    {
        // FIXME: We would like to not watch things that are unprofitable to watch, like
        // dictionaries. Unfortunately, we can't do such things: a dictionary could get flattened,
        // in which case it will start to appear watchable and so the DFG will think that it is
        // watching it. We should come up with a comprehensive story for not watching things that
        // aren't profitable to watch.
        // https://bugs.webkit.org/show_bug.cgi?id=133625

    	// - We don't watch Structures that either decided not to be watched, or whose predecessors
    	//   decided not to be watched. This happens either when a transition is fired while being
    	//   watched, or if a dictionary transition occurs.
    	if (transitionWatchpointIsLikelyToBeFired())
    		return false;

        return true;
    }

    bool dfgShouldWatch() const
    {
        return dfgShouldWatchIfPossible() && transitionWatchpointSetIsStillValid();
    }

	void addTransitionWatchpoint(Watchpoint* watchpoint) const
	{
		ASSERT(transitionWatchpointSetIsStillValid());
		m_transitionWatchpointSet.add(watchpoint);
	}

	void didTransitionFromThisStructure(DeferredStructureTransitionWatchpointFire* = 0) const;

    InlineWatchpointSet& transitionWatchpointSet() const
    {
        return m_transitionWatchpointSet;
    }

    bool transitionWatchpointIsLikelyToBeFired() const { return m_transitionWatchpointIsLikelyToBeFired; }
    void setTransitionWatchpointIsLikelyToBeFired(bool value) { m_transitionWatchpointIsLikelyToBeFired = value; }

    bool didWatchInternalProperties() const { return m_didWatchInternalProperties; }
    void setDidWatchInternalProperties(bool value) { m_didWatchInternalProperties = value; }

    WatchpointSet* ensurePropertyReplacementWatchpointSet(VM&, PropertyOffset);
    void startWatchingPropertyForReplacements(VM& vm, PropertyOffset offset)
    {
        ensurePropertyReplacementWatchpointSet(vm, offset);
    }
    void startWatchingPropertyForReplacements(VM&, PropertyName);
    RefPtr<WatchpointSet> propertyReplacementWatchpointSet(PropertyOffset);
    void didReplaceProperty(VM&, PropertyOffset);
    void didCachePropertyReplacement(VM&, PropertyOffset);

    void startWatchingInternalPropertiesIfNecessary(VM& vm)
    {
        if (LIKELY(didWatchInternalProperties()))
            return;
        startWatchingInternalProperties(vm);
    }

    void startWatchingInternalPropertiesIfNecessaryForEntireChain(VM& vm)
    {
        for (Structure* structure = this; structure; structure = structure->storedPrototypeStructure())
            structure->startWatchingInternalPropertiesIfNecessary(vm);
    }

    void dump(PrintStream&) const;
    void dumpInContext(PrintStream&, DumpContext*) const;
    void dumpBrief(PrintStream&, const CString&) const;

    static void dumpContextHeader(PrintStream&);

    DECLARE_EXPORT_INFO;

private:
	friend class LLIntOffsetsExtractor;

	JS_EXPORT_PRIVATE Structure(JSGlobalData&, JSGlobalObject*, JSValue prototype, const TypeInfo&, const ClassInfo*, IndexingType, unsigned inlineCapacity);
	Structure(JSGlobalData&);
	Structure(JSGlobalData&, Structure*, DeferredStructureTransitionWatchpointFire*);

	static Structure* create(JSGlobalData&, Structure*, DeferredStructureTransitionWatchpointFire* = 0);

    void findStructuresAndMapForMaterialization(Vector<Structure*, 8>& structures, PropertyTable*&);

	typedef enum {
		NoneDictionaryKind = 0,
		CachedDictionaryKind = 1,
		UncachedDictionaryKind = 2
	} DictionaryKind;
	static Structure* toDictionaryTransition(JSGlobalData&, Structure*, DictionaryKind, DeferredStructureTransitionWatchpointFire* = 0);

	PropertyOffset add(JSGlobalData&, PropertyName, unsigned attributes);
	PropertyOffset remove(PropertyName);

	void createPropertyMap(JSGlobalData&, unsigned keyCount = 0);
	void checkConsistency();

	WriteBarrier<PropertyTable>& propertyTable();
	PropertyTable* takePropertyTableOrCloneIfPinned(JSGlobalData&);
	PropertyTable* copyPropertyTable(JSGlobalData&);
	PropertyTable* copyPropertyTableForPinning(JSGlobalData&);
	JS_EXPORT_PRIVATE void materializePropertyMap(JSGlobalData&);
	ALWAYS_INLINE void materializePropertyMapIfNecessary(JSGlobalData& globalData, DeferGC&)
	{
        ASSERT(!isCompilationThread());
		ASSERT(structure()->classInfo() == &s_info);
        ASSERT(checkOffsetConsistency());
		if (!propertyTable() && previousID())
			materializePropertyMap(globalData);
	}
	ALWAYS_INLINE void materializePropertyMapIfNecessary(JSGlobalData& vm, PropertyTable*& table)
    {
        ASSERT(!isCompilationThread());
		ASSERT(structure()->classInfo() == &s_info);
        ASSERT(checkOffsetConsistency());
        table = propertyTable().get();
        if (!table && previousID()) {
            DeferGC deferGC(vm.heap);
            materializePropertyMap(vm);
            table = propertyTable().get();
        }
    }
	ALWAYS_INLINE void materializePropertyMapIfNecessaryForPinning(JSGlobalData& globalData, DeferGC&)
	{
		ASSERT(structure()->classInfo() == &s_info);
        checkOffsetConsistency();
		if (!propertyTable())
			materializePropertyMap(globalData);
	}

    void setPreviousID(JSGlobalData& globalData, Structure* transition, Structure* structure)
    {
        if (hasRareData())
            rareData()->setPreviousID(globalData, transition, structure);
        else
            m_previousOrRareData.set(globalData, transition, structure);
    }

    void clearPreviousID()
    {
        if (hasRareData())
            rareData()->clearPreviousID();
        else
            m_previousOrRareData.clear();
    }

	int transitionCount() const
	{
		// Since the number of transitions is always the same as m_offset, we keep the size of Structure down by not storing both.
		return numberOfSlotsForLastOffset(m_offset, m_inlineCapacity);
	}

	bool isValid(JSGlobalObject*, StructureChain* cachedPrototypeChain) const;
	bool isValid(ExecState*, StructureChain* cachedPrototypeChain) const;

	void pin();

	bool hasRareData() const { return typeInfo().structureHasRareData(); }

    Structure* previous() const
    {
        ASSERT(!hasRareData());
        return static_cast<Structure*>(m_previousOrRareData.get());
    }

    StructureRareData* rareData() const
    {
        ASSERT(hasRareData());
        return static_cast<StructureRareData*>(m_previousOrRareData.get());
    }

    bool checkOffsetConsistency() const;

    void allocateRareData(JSGlobalData&);
    void cloneRareDataFrom(JSGlobalData&, const Structure*);

    void startWatchingInternalProperties(VM&);

	static const int s_maxTransitionLength = 64;
	static const int s_maxTransitionLengthForNonEvalPutById = 512;

	WriteBarrier<JSGlobalObject> m_globalObject;
	WriteBarrier<Unknown> m_prototype;
	mutable WriteBarrier<StructureChain> m_cachedPrototypeChain;

    WriteBarrier<JSCell> m_previousOrRareData;

	RefPtr<StringImpl> m_nameInPrevious;

	const ClassInfo* m_classInfo;

	StructureTransitionTable m_transitionTable;

    // Should be accessed through propertyTable(). During GC, it may be set to 0 by another thread.
    WriteBarrier<PropertyTable> m_propertyTableUnsafe;

	mutable InlineWatchpointSet m_transitionWatchpointSet;

	COMPILE_ASSERT(firstOutOfLineOffset < 256, firstOutOfLineOffset_fits);

	// m_offset does not account for anonymous slots
	PropertyOffset m_offset;

    TypeInfo m_typeInfo;
    IndexingType m_indexingType;

    uint8_t m_inlineCapacity;
	unsigned m_dictionaryKind : 2;
	unsigned m_hasBeenFlattenedBefore : 1;
	unsigned m_isPinnedPropertyTable : 1;
	unsigned m_hasGetterSetterProperties : 1;
	unsigned m_hasCustomGetterSetterProperties : 1;
	unsigned m_hasReadOnlyOrGetterSetterPropertiesExcludingProto : 1;
	unsigned m_hasNonEnumerableProperties : 1;
	unsigned m_attributesInPrevious : 14;
	unsigned m_preventExtensions : 1;
	unsigned m_didTransition : 1;
	unsigned m_staticFunctionReified : 1;
	unsigned m_transitionWatchpointIsLikelyToBeFired : 1;
	unsigned m_didWatchInternalProperties : 1;
};

} // namespace JSC

#endif // Structure_h
