/*
 * Copyright (C) 2012 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 * 3.  Neither the name of Apple Computer, Inc. ("Apple") nor the names of
 *     its contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE AND ITS CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "JSGlobalObject.h"
#include "JSScope.h"
#include "Operations.h"
#include "SymbolTable.h"
#include "VariableWatchpointSetInlines.h"

namespace JSC {

const ClassInfo SharedSymbolTable::s_info = { "SharedSymbolTable", 0, 0, CREATE_METHOD_TABLE(SharedSymbolTable) };

SymbolTableEntry& SymbolTableEntry::copySlow(const SymbolTableEntry& other)
{
    ASSERT(other.isFat());
    FatEntry* newFatEntry = new FatEntry(*other.fatEntry());
    freeFatEntry();
    m_bits = bitwise_cast<intptr_t>(newFatEntry);
    return *this;
}

void SharedSymbolTable::destroy(JSCell* cell)
{
    SharedSymbolTable* thisObject = jsCast<SharedSymbolTable*>(cell);
    thisObject->SharedSymbolTable::~SharedSymbolTable();
}

void SymbolTableEntry::freeFatEntrySlow()
{
    ASSERT(isFat());
    delete fatEntry();
}

JSValue SymbolTableEntry::inferredValue()
{
    if (!isFat())
        return JSValue();
    return fatEntry()->m_watchpoints->inferredValue();
}

void SymbolTableEntry::prepareToWatch(SharedSymbolTable* symbolTable)
{
    FatEntry* entry = inflate();
    if (entry->m_watchpoints)
        return;
    entry->m_watchpoints = adoptRef(new VariableWatchpointSet(*symbolTable));
}

void SymbolTableEntry::addWatchpoint(Watchpoint* watchpoint)
{
    fatEntry()->m_watchpoints->add(watchpoint);
}

void SymbolTableEntry::notifyWriteSlow(JSGlobalData& vm, JSValue value)
{
    VariableWatchpointSet* watchpoints = fatEntry()->m_watchpoints.get();
    if (!watchpoints)
        return;

    watchpoints->notifyWrite(vm, value);
}

SymbolTableEntry::FatEntry* SymbolTableEntry::inflateSlow()
{
    FatEntry* entry = new FatEntry(m_bits);
    m_bits = bitwise_cast<intptr_t>(entry);
    return entry;
}

void SharedSymbolTable::visitChildren(JSCell* thisCell, SlotVisitor& visitor)
{
	SharedSymbolTable* thisSymbolTable = jsCast<SharedSymbolTable*>(thisCell);
    if (!thisSymbolTable->m_watchpointCleanup) {
        thisSymbolTable->m_watchpointCleanup =
            adoptPtr(new WatchpointCleanup(thisSymbolTable));
    }

    Base::visitChildren(thisCell, visitor);
    visitor.addUnconditionalFinalizer(thisSymbolTable->m_watchpointCleanup.get());

    // Save some memory. This is O(n) to rebuild and we do so on the fly.
    //ConcurrentJITLocker locker(thisSymbolTable->m_lock);
    thisSymbolTable->m_localToEntry = nullptr;
}

SharedSymbolTable::WatchpointCleanup::WatchpointCleanup(SharedSymbolTable* symbolTable)
    : m_symbolTable(symbolTable)
{
}

SharedSymbolTable::WatchpointCleanup::~WatchpointCleanup() { }

void SharedSymbolTable::WatchpointCleanup::finalizeUnconditionally(JSGlobalData& vm)
{
    iterator iter = m_symbolTable->begin();
    iterator end = m_symbolTable->end();
    for (; iter != end; ++iter) {
        if (VariableWatchpointSet* set = iter->second.watchpointSet())
            set->finalizeUnconditionally(vm);
    }
}

const SharedSymbolTable::LocalToEntryVec& SharedSymbolTable::localToEntry()
{
    if (UNLIKELY(!m_localToEntry)) {
        unsigned size = 0;
        for (iterator iter = begin(), end = this->end(); iter != end; ++iter) {
            VirtualRegister_ reg(iter->second.getIndex());
            if (reg.isLocal())
                size = std::max(size, static_cast<unsigned>(reg.toLocal()) + 1);
        }

        m_localToEntry = adoptPtr(new LocalToEntryVec(size, 0));
        for (iterator iter = begin(), end = this->end(); iter != end; ++iter) {
            VirtualRegister_ reg(iter->second.getIndex());
            if (reg.isLocal())
                m_localToEntry->at(reg.toLocal()) = &iter->second;
        }
    }

    return *m_localToEntry;
}

SymbolTableEntry* SharedSymbolTable::entryFor(VirtualRegister_ reg)
{
    if (!reg.isLocal())
        return 0;
    return localToEntry()[reg.toLocal()];
}

} // namespace JSC

