/*
 * Copyright (C) 2008 Apple Inc. All Rights Reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE COMPUTER, INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE COMPUTER, INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 *
 */

#include "config.h"
#include "ActiveDOMObject.h"

#include "ScriptExecutionContext.h"
#include "WorkerContext.h"
#include "WorkerThread.h"
#include <wtf/HashSet.h>
#include <wtf/Threading.h>
#include <wtf/Vector.h>

namespace WebCore {

static Mutex& mutex()
{
	static Mutex mutex;
	return mutex;
}

static HashSet<ActiveDOMObject*>& activeDOMObjects()
{
	static HashSet<ActiveDOMObject*> objects;
	return objects;
}

DLL_PUBLIC void dumpActiveDOMObjects(Vector<pair<const char*, int> >& dumpObjects)
{
	HashSet<ActiveDOMObject*>& objects = activeDOMObjects();

	for(HashSet<WebCore::ActiveDOMObject*>::iterator it = objects.begin(); it != objects.end(); ++it) {
		const char* name = (*it)->activeDOMObjectName();
		bool found = false;

		for(int j = 0; j < dumpObjects.size(); ++j) {
			if(strcmp(name, dumpObjects[j].first) == 0) {
				++dumpObjects[j].second;
				found = true;
				break;
			}
		}

		if(!found)
			dumpObjects.append(pair<const char*, int>(name, 1));
	}
}

ActiveDOMObject::ActiveDOMObject(ScriptExecutionContext* scriptExecutionContext, void* upcastPointer)
    : ContextDestructionObserver(scriptExecutionContext)
    , m_pendingActivityCount(0)
#if !ASSERT_DISABLED
    , m_suspendIfNeededCalled(false)
#endif
{
	{
		MutexLocker locker(mutex());
		activeDOMObjects().add(this);
	}

    if (!m_scriptExecutionContext)
        return;

    ASSERT(m_scriptExecutionContext->isContextThread());
    m_scriptExecutionContext->didCreateActiveDOMObject(this, upcastPointer);
}

ActiveDOMObject::~ActiveDOMObject()
{
	{
		MutexLocker locker(mutex());
		activeDOMObjects().remove(this);
	}

    if (!m_scriptExecutionContext)
        return;

    ASSERT(m_suspendIfNeededCalled);
    ASSERT(m_scriptExecutionContext->isContextThread());
    m_scriptExecutionContext->willDestroyActiveDOMObject(this);
}

void ActiveDOMObject::suspendIfNeeded()
{
#if !ASSERT_DISABLED
    ASSERT(!m_suspendIfNeededCalled);
    m_suspendIfNeededCalled = true;
#endif
    if (!m_scriptExecutionContext)
        return;

    m_scriptExecutionContext->suspendActiveDOMObjectIfNeeded(this);
}

} // namespace WebCore
