/*
 * Copyright (C) 1999 Lars Knoll (knoll@kde.org)
 *           (C) 1999 Antti Koivisto (koivisto@kde.org)
 *           (C) 2001 Peter Kelly (pmk@post.com)
 *           (C) 2001 Dirk Mueller (mueller@kde.org)
 * Copyright (C) 2003, 2004, 2005, 2006, 2008, 2010 Apple Inc. All rights reserved.
 * Copyright (C) 2012 Nokia Corporation and/or its subsidiary(-ies).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 *
 */

#ifndef ElementAttributeData_h
#define ElementAttributeData_h

#include "Attribute.h"
#include "SpaceSplitString.h"
#include "StylePropertySet.h"
#include <wtf/NotFound.h>

namespace WebCore {

class Attr;
class Element;
class ShareableElementData;
class UniqueElementData;

typedef Vector<RefPtr<Attr> > AttrList;

class ElementData : public RefCounted<ElementData> {
    WTF_MAKE_FAST_ALLOCATED;
public:
    // Override RefCounted's deref() to ensure operator delete is called on
    // the appropriate subclass type.
    void deref();

    ALWAYS_INLINE void clearClass() const { m_classNames.clear(); }
    ALWAYS_INLINE void setClass(const AtomicString& className, bool shouldFoldCase) const { m_classNames.set(className, shouldFoldCase); }
    ALWAYS_INLINE const SpaceSplitString& classNames() const { return m_classNames; }

    ALWAYS_INLINE const AtomicString& idForStyleResolution() const { return m_idForStyleResolution; }
    ALWAYS_INLINE void setIdForStyleResolution(const AtomicString& newId) const { m_idForStyleResolution = newId; }

    ALWAYS_INLINE const StylePropertySet* inlineStyle() const { return m_inlineStyle.get(); }
    ALWAYS_INLINE StylePropertySet* inlineStyle() { return m_inlineStyle.get(); }

    const StylePropertySet* presentationAttributeStyle() const;

    size_t length() const;
    ALWAYS_INLINE bool isEmpty() const { return !length(); }

    PassRefPtr<Attr> getAttributeNode(const String&, bool shouldIgnoreAttributeCase, Element*) const;
    PassRefPtr<Attr> getAttributeNode(const QualifiedName&, Element*) const;

    AttrList* attrList(Element*) const;

    // Internal interface.
    const Attribute* attributeItem(unsigned index) const;
    const Attribute* getAttributeItem(const QualifiedName&) const;
    size_t findAttributeIndexByName(const QualifiedName&) const;
    size_t findAttributeIndexByName(const AtomicString& name, bool shouldIgnoreAttributeCase) const;

    ALWAYS_INLINE bool hasID() const { return !m_idForStyleResolution.isNull(); }
    ALWAYS_INLINE bool hasClass() const { return !m_classNames.isEmpty(); }

    bool isEquivalent(const ElementData* other) const;

    ALWAYS_INLINE bool isUnique() const { return m_isUnique; }

    void setAttr(Element*, const QualifiedName&, Attr*) const;
    void removeAttr(Element*, const QualifiedName&) const;
    PassRefPtr<Attr> attrIfExists(Element*, const QualifiedName&) const;
    PassRefPtr<Attr> ensureAttr(Element*, const QualifiedName&) const;
    void detachAttrObjectsFromElement(Element*) const;

protected:
    ElementData()
        : m_isUnique(true)
        , m_arraySize(0)
		, m_presentationAttributeStyleIsDirty(false)
		, m_styleAttributeIsDirty(false)
#if ENABLE(SVG)
        , m_animatedSVGAttributesAreDirty(false)
#endif
    { }

    ElementData(unsigned arraySize)
        : m_isUnique(false)
        , m_arraySize(arraySize)
    	, m_presentationAttributeStyleIsDirty(false)
    	, m_styleAttributeIsDirty(false)
#if ENABLE(SVG)
        , m_animatedSVGAttributesAreDirty(false)
#endif
    { }

    ElementData(const ElementData&, bool isUnique);

    unsigned m_isUnique : 1;
    unsigned m_arraySize : 28;
    mutable unsigned m_presentationAttributeStyleIsDirty : 1;
    mutable unsigned m_styleAttributeIsDirty : 1;
#if ENABLE(SVG)
    mutable unsigned m_animatedSVGAttributesAreDirty : 1;
#endif

    mutable RefPtr<StylePropertySet> m_inlineStyle;
    mutable SpaceSplitString m_classNames;
    mutable AtomicString m_idForStyleResolution;

private:
    friend class Element;
    friend class StyledElement;
    friend class ShareableElementData;
    friend class UniqueElementData;
#if ENABLE(SVG)
    friend class SVGElement;
#endif

    const Attribute* getAttributeItem(const AtomicString& name, bool shouldIgnoreAttributeCase) const;
    size_t findAttributeIndexByNameSlowCase(const AtomicString& name, bool shouldIgnoreAttributeCase) const;

    PassRefPtr<UniqueElementData> makeUniqueCopy() const;
};

#if COMPILER(MSVC)
#pragma warning(push)
#pragma warning(disable: 4200) // Disable "zero-sized array in struct/union" warning
#endif

class ShareableElementData : public ElementData {
public:
    static PassRefPtr<ShareableElementData> createWithAttributes(const Vector<Attribute>&);

    explicit ShareableElementData(const Vector<Attribute>&);
    explicit ShareableElementData(const UniqueElementData&);
    ~ShareableElementData();

    Attribute m_attributeArray[0];
};

#if COMPILER(MSVC)
#pragma warning(pop)
#endif

class UniqueElementData : public ElementData {
public:
    static PassRefPtr<UniqueElementData> create();
    PassRefPtr<ShareableElementData> makeShareableCopy() const;

    // These functions do no error/duplicate checking.
    void addAttribute(const QualifiedName&, const AtomicString&);
    void removeAttribute(size_t index);

    Attribute* attributeItem(unsigned index);
    Attribute* getAttributeItem(const QualifiedName&);

    UniqueElementData() { }
    explicit UniqueElementData(const ShareableElementData&);
    explicit UniqueElementData(const UniqueElementData&);

    mutable RefPtr<StylePropertySet> m_presentationAttributeStyle;
    Vector<Attribute, 4> m_attributeVector;
};

ALWAYS_INLINE size_t ElementData::length() const
{
    if (isUnique())
    	return static_cast<const UniqueElementData*>(this)->m_attributeVector.size();
    return m_arraySize;
}

inline const StylePropertySet* ElementData::presentationAttributeStyle() const
{
    if (!m_isUnique)
        return 0;
    return static_cast<const UniqueElementData*>(this)->m_presentationAttributeStyle.get();
}

inline const Attribute* ElementData::getAttributeItem(const AtomicString& name, bool shouldIgnoreAttributeCase) const
{
    size_t index = findAttributeIndexByName(name, shouldIgnoreAttributeCase);
    if (index != notFound)
        return attributeItem(index);
    return 0;
}

inline size_t ElementData::findAttributeIndexByName(const QualifiedName& name) const
{
    for (unsigned i = 0; i < length(); ++i) {
        if (attributeItem(i)->name().matches(name))
            return i;
    }
    return notFound;
}

// We use a boolean parameter instead of calling shouldIgnoreAttributeCase so that the caller
// can tune the behavior (hasAttribute is case sensitive whereas getAttribute is not).
inline size_t ElementData::findAttributeIndexByName(const AtomicString& name, bool shouldIgnoreAttributeCase) const
{
    unsigned len = length();
    bool doSlowCheck = shouldIgnoreAttributeCase;

    const AtomicString caseAdjustedName = shouldIgnoreAttributeCase ? name.convertToASCIILowercase() : name;
    // Optimize for the case where the attribute exists and its name exactly matches.
    for (unsigned i = 0; i < len; ++i) {
        const Attribute* attribute = attributeItem(i);
        if (!attribute->name().hasPrefix()) {
            if (caseAdjustedName == attribute->localName())
                return i;
        } else
            doSlowCheck = true;
    }

    if (doSlowCheck)
        return findAttributeIndexByNameSlowCase(name, shouldIgnoreAttributeCase);
    return notFound;
}

inline const Attribute* ElementData::getAttributeItem(const QualifiedName& name) const
{
    for (unsigned i = 0; i < length(); ++i) {
        if (attributeItem(i)->name().matches(name))
            return attributeItem(i);
    }
    return 0;
}

ALWAYS_INLINE const Attribute* ElementData::attributeItem(unsigned index) const
{
    ASSERT(index < length());
    if (m_isUnique)
		return &static_cast<const UniqueElementData*>(this)->m_attributeVector.at(index);
	return &static_cast<const ShareableElementData*>(this)->m_attributeArray[index];
}

inline void ElementData::deref()
{
    if (!derefBase())
        return;

    if (m_isUnique)
        delete static_cast<UniqueElementData*>(this);
    else
        delete static_cast<ShareableElementData*>(this);
}

}

#endif // ElementAttributeData_h
