/*
 * Copyright (C) 2008, 2009, 2010 Apple Inc. All rights reserved.
 * Copyright (C) 2008 David Smith <catfish.man@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 *
 */

#ifndef ElementRareData_h
#define ElementRareData_h

#include "ClassList.h"
#include "DatasetDOMStringMap.h"
#include "Element.h"
#include "ElementShadow.h"
#include "HTMLCollection.h"
#include "NamedNodeMap.h"
#include "NodeRareData.h"
#include "PseudoElement.h"
#include <wtf/OwnPtr.h>

namespace WebCore {

class HTMLCollection;

class ElementRareData : public NodeRareData {
	WTF_MAKE_FAST_ALLOCATED;
public:
    ElementRareData(Document*);
    ~ElementRareData();

    void setPseudoElement(PseudoId, PassRefPtr<PseudoElement>);
    PseudoElement* pseudoElement(PseudoId) const;
    bool hasPseudoElements() const { return m_generatedBefore || m_generatedAfter; }

    void resetComputedStyle();
    void resetDynamicRestyleObservations();

    using NodeRareData::needsFocusAppearanceUpdateSoonAfterAttach;
    using NodeRareData::setNeedsFocusAppearanceUpdateSoonAfterAttach;
    using NodeRareData::styleAffectedByEmpty;
    using NodeRareData::setStyleAffectedByEmpty;
    using NodeRareData::isInCanvasSubtree;
    using NodeRareData::setIsInCanvasSubtree;
#if ENABLE(FULLSCREEN_API)
    using NodeRareData::containsFullScreenElement;
    using NodeRareData::setContainsFullScreenElement;
#endif
#if ENABLE(DIALOG_ELEMENT)
    using NodeRareData::isInTopLayer;
    using NodeRareData::setIsInTopLayer;
#endif
#if ENABLE(SVG)
    using NodeRareData::hasPendingResources;
    using NodeRareData::setHasPendingResources;
#endif
    using NodeRareData::childrenAffectedByHover;
    using NodeRareData::setChildrenAffectedByHover;
    using NodeRareData::childrenAffectedByActive;
    using NodeRareData::setChildrenAffectedByActive;
    using NodeRareData::childrenAffectedByDrag;
    using NodeRareData::setChildrenAffectedByDrag;
    using NodeRareData::childrenAffectedByFirstChildRules;
    using NodeRareData::setChildrenAffectedByFirstChildRules;
    using NodeRareData::childrenAffectedByLastChildRules;
    using NodeRareData::setChildrenAffectedByLastChildRules;
    using NodeRareData::childrenAffectedByDirectAdjacentRules;
    using NodeRareData::setChildrenAffectedByDirectAdjacentRules;
    using NodeRareData::childrenAffectedByForwardPositionalRules;
    using NodeRareData::setChildrenAffectedByForwardPositionalRules;
    using NodeRareData::childrenAffectedByBackwardPositionalRules;
    using NodeRareData::setChildrenAffectedByBackwardPositionalRules;
    using NodeRareData::childIndex;
    using NodeRareData::setChildIndex;

    ElementShadow* shadow() const { return m_shadow.get(); }
    void setShadow(PassOwnPtr<ElementShadow> shadow) { m_shadow = shadow; }

    NamedNodeMap* attributeMap() const { return m_attributeMap.get(); }
    void setAttributeMap(PassOwnPtr<NamedNodeMap> attributeMap) { m_attributeMap = attributeMap; }

    RenderStyle* computedStyle() const { return m_computedStyle.get(); }
    void setComputedStyle(PassRefPtr<RenderStyle> computedStyle) { m_computedStyle = computedStyle; }

    ClassList* classList() const { return m_classList.get(); }
    void setClassList(PassOwnPtr<ClassList> classList) { m_classList = classList; }
    void clearClassListValueForQuirksMode()
    {
        if (!m_classList)
            return;
        m_classList->clearValueForQuirksMode();
    }

    PassRefPtr<HTMLCollection> ensureCachedHTMLCollection(Element*, CollectionType);
    HTMLCollection* cachedHTMLCollection(CollectionType type)
    {
        return nodeLists() ? nodeLists()->cacheWithAtomicName<HTMLCollection>(type) : 0;
    }

    void removeCachedHTMLCollection(HTMLCollection* collection, CollectionType type)
    {
        ASSERT(nodeLists());
        nodeLists()->removeCacheWithAtomicName(collection, type);
    }

    LayoutSize m_minimumSizeForResizing;
    IntSize m_savedLayerScrollOffset;
    RefPtr<RenderStyle> m_computedStyle;

    OwnPtr<DatasetDOMStringMap> m_datasetDOMStringMap;
    OwnPtr<ClassList> m_classList;
    OwnPtr<ElementShadow> m_shadow;
    OwnPtr<NamedNodeMap> m_attributeMap;

    RefPtr<PseudoElement> m_generatedBefore;
    RefPtr<PseudoElement> m_generatedAfter;

private:
    void releasePseudoElement(PseudoElement*);
};

ALWAYS_INLINE IntSize defaultMinimumSizeForResizing()
{
    return IntSize(MAX_LAYOUT_UNIT, MAX_LAYOUT_UNIT);
}

ALWAYS_INLINE ElementRareData::ElementRareData(Document* document)
: NodeRareData(document)
, m_minimumSizeForResizing(defaultMinimumSizeForResizing())
, m_generatedBefore(0)
, m_generatedAfter(0)
{
}

ALWAYS_INLINE ElementRareData::~ElementRareData()
{
    ASSERT(!m_shadow);
    ASSERT(!m_generatedBefore);
    ASSERT(!m_generatedAfter);
}

inline void ElementRareData::setPseudoElement(PseudoId pseudoId, PassRefPtr<PseudoElement> element)
{
    switch (pseudoId) {
    case BEFORE:
        releasePseudoElement(m_generatedBefore.get());
        m_generatedBefore = element;
        break;
    case AFTER:
        releasePseudoElement(m_generatedAfter.get());
        m_generatedAfter = element;
        break;
    default:
        ASSERT_NOT_REACHED();
    }
}

inline PseudoElement* ElementRareData::pseudoElement(PseudoId pseudoId) const
{
    switch (pseudoId) {
    case BEFORE:
        return m_generatedBefore.get();
    case AFTER:
        return m_generatedAfter.get();
    default:
        return 0;
    }
}

inline void ElementRareData::releasePseudoElement(PseudoElement* element)
{
    if (!element)
        return;

    if (element->attached())
        element->detach();

    ASSERT(!element->nextSibling());
    ASSERT(!element->previousSibling());

    element->setParentOrHostNode(0);
}

ALWAYS_INLINE void ElementRareData::resetComputedStyle()
{
    m_computedStyle.clear();
    setStyleAffectedByEmpty(false);
    setChildIndex(0);
}

inline void ElementRareData::resetDynamicRestyleObservations()
{
    setChildrenAffectedByHover(false);
    setChildrenAffectedByActive(false);
    setChildrenAffectedByDrag(false);
    setChildrenAffectedByFirstChildRules(false);
    setChildrenAffectedByLastChildRules(false);
    setChildrenAffectedByDirectAdjacentRules(false);
    setChildrenAffectedByForwardPositionalRules(false);
    setChildrenAffectedByBackwardPositionalRules(false);
}

}
#endif // ElementRareData_h
