/*
 * Copyright (C) 2008, 2010 Apple Inc. All rights reserved.
 * Copyright (C) 2008 David Smith <catfish.man@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 *
 */

#ifndef NodeRareData_h
#define NodeRareData_h

#include "ChildNodeList.h"
#include "DOMSettableTokenList.h"
#include "DynamicNodeList.h"
#include "MutationObserver.h"
#include "MutationObserverRegistration.h"
#include "Page.h"
#include "QualifiedName.h"
#include "TagNodeList.h"
#if ENABLE(VIDEO_TRACK)
#include "TextTrack.h"
#endif
#include <wtf/HashSet.h>
#include <wtf/OwnPtr.h>
#include <wtf/PassOwnPtr.h>
#include <wtf/text/AtomicString.h>
#include <wtf/text/StringHash.h>

#if ENABLE(MICRODATA)
#include "HTMLPropertiesCollection.h"
#include "MicroDataItemList.h"
#endif

namespace WebCore {

class LabelsNodeList;
class RadioNodeList;
class TreeScope;

class NodeListsNodeData {
    WTF_MAKE_NONCOPYABLE(NodeListsNodeData); WTF_MAKE_FAST_ALLOCATED;
public:
    void clearChildNodeListCache()
    {
        if (m_childNodeList)
            m_childNodeList->invalidateCache();
    }

    PassRefPtr<ChildNodeList> ensureChildNodeList(Node* node)
    {
        if (m_childNodeList)
            return m_childNodeList;
        RefPtr<ChildNodeList> list = ChildNodeList::create(node);
        m_childNodeList = list.get();
        return list.release();
    }

    void removeChildNodeList(ChildNodeList* list)
    {
        ASSERT(m_childNodeList = list);
        if (deleteThisAndUpdateNodeRareDataIfAboutToRemoveLastList(list->ownerNode()))
            return;
        m_childNodeList = 0;
    }

    template <typename StringType>
    struct NodeListCacheMapEntryHash {
        static unsigned hash(const std::pair<unsigned char, StringType>& entry)
        {
            return DefaultHash<StringType>::Hash::hash(entry.second) + entry.first;
        }
        static bool equal(const std::pair<unsigned char, StringType>& a, const std::pair<unsigned char, StringType>& b) { return a.first == b.first && DefaultHash<StringType>::Hash::equal(a.second, b.second); }
		static const bool safeToCompareToEmptyOrDeleted = DefaultHash<StringType>::Hash::safeToCompareToEmptyOrDeleted;
    };

    typedef HashMap<std::pair<unsigned char, AtomicString>, DynamicNodeListCacheBase*, NodeListCacheMapEntryHash<AtomicString> > NodeListAtomicNameCacheMap;
    typedef HashMap<std::pair<unsigned char, String>, DynamicSubtreeNodeList*, NodeListCacheMapEntryHash<String> > NodeListNameCacheMap;
    typedef HashMap<QualifiedName, TagNodeList*> TagNodeListCacheNS;

    // FIXME: This function should be renamed since it doesn't have an atomic name.
    template<typename T>
    PassRefPtr<T> addCacheWithAtomicName(Node* node, DynamicNodeList::NodeListType listType, const AtomicString& name)
    {
        NodeListAtomicNameCacheMap::AddResult result = m_atomicNameCaches.add(namedNodeListKey(listType, name), 0);
        if (!result.isNewEntry)
            return static_cast<T*>(result.iterator->second);

        RefPtr<T> list = T::create(node, name);
        result.iterator->second = list.get();
        return list.release();
    }

    template<typename T>
    PassRefPtr<T> addCacheWithAtomicName(Element* node, CollectionType collectionType)
    {
        unsigned char type = DynamicNodeList::PastLastNodeListType + collectionType - FirstNodeCollectionType;
        NodeListAtomicNameCacheMap::AddResult result = m_atomicNameCaches.add(namedNodeListKey(type, starAtom), 0);
        if (!result.isNewEntry)
            return static_cast<T*>(result.iterator->second);

        RefPtr<T> list = T::create(node, collectionType);
        result.iterator->second = list.get();
        return list.release();
    }

    template<typename T>
    T* cacheWithAtomicName(CollectionType collectionType)
    {
        unsigned char type = DynamicNodeList::PastLastNodeListType + collectionType - FirstNodeCollectionType;
        return static_cast<T*>(m_atomicNameCaches.get(namedNodeListKey(type, starAtom)));
    }

    template<typename T>
    PassRefPtr<T> addCacheWithName(Node* node, DynamicNodeList::NodeListType listType, const String& name)
    {
        NodeListNameCacheMap::AddResult result = m_nameCaches.add(namedNodeListKey(listType, name), 0);
        if (!result.isNewEntry)
            return static_cast<T*>(result.iterator->second);

        RefPtr<T> list = T::create(node, name);
        result.iterator->second = list.get();
        return list.release();
    }

    PassRefPtr<TagNodeList> addCacheWithQualifiedName(Node* node, const AtomicString& namespaceURI, const AtomicString& localName)
    {
        QualifiedName name(nullAtom, localName, namespaceURI);
        TagNodeListCacheNS::AddResult result = m_tagNodeListCacheNS.add(name, 0);
        if (!result.isNewEntry)
            return result.iterator->second;

        RefPtr<TagNodeList> list = TagNodeList::create(node, namespaceURI, localName);
        result.iterator->second = list.get();
        return list.release();
    }

    void removeCacheWithAtomicName(DynamicNodeListCacheBase* list, DynamicNodeList::NodeListType listType, const AtomicString& name)
    {
        ASSERT(list == m_atomicNameCaches.get(namedNodeListKey(collectionType, name)));
        if (deleteThisAndUpdateNodeRareDataIfAboutToRemoveLastList(list->ownerNode()))
            return;
        m_atomicNameCaches.remove(namedNodeListKey(listType, name));
    }

    void removeCacheWithAtomicName(DynamicNodeListCacheBase* list, CollectionType collectionType)
    {
        unsigned char type = DynamicNodeList::PastLastNodeListType + collectionType - FirstNodeCollectionType;
        ASSERT_UNUSED(list, list == m_atomicNameCaches.get(namedNodeListKey(type, starAtom)));
        m_atomicNameCaches.remove(namedNodeListKey(type, starAtom));
    }

    void removeCacheWithName(DynamicSubtreeNodeList* list, DynamicNodeList::NodeListType listType, const String& name)
    {
        ASSERT(list == m_nameCaches.get(namedNodeListKey(collectionType, name)));
        if (deleteThisAndUpdateNodeRareDataIfAboutToRemoveLastList(list->ownerNode()))
            return;
        m_nameCaches.remove(namedNodeListKey(listType, name));
    }

    void removeCacheWithQualifiedName(DynamicSubtreeNodeList* list, const AtomicString& namespaceURI, const AtomicString& localName)
    {
        QualifiedName name(nullAtom, localName, namespaceURI);
        ASSERT(list == m_tagNodeListCacheNS.get(name));
        if (deleteThisAndUpdateNodeRareDataIfAboutToRemoveLastList(list->ownerNode()))
            return;
        m_tagNodeListCacheNS.remove(name);
    }

    static ALWAYS_INLINE PassOwnPtr<NodeListsNodeData> create()
    {
        return adoptPtr(new NodeListsNodeData);
    }

    void invalidateCaches(const QualifiedName* attrName = 0);
    bool isEmpty() const
    {
        return m_atomicNameCaches.isEmpty() && m_nameCaches.isEmpty() && m_tagNodeListCacheNS.isEmpty();
    }

    void adoptTreeScope()
    {
        invalidateCaches();
    }

    void adoptDocument(Document* oldDocument, Document* newDocument)
    {
        invalidateCaches();

        if (oldDocument != newDocument) {
            NodeListAtomicNameCacheMap::const_iterator atomicNameCacheEnd = m_atomicNameCaches.end();
            for (NodeListAtomicNameCacheMap::const_iterator it = m_atomicNameCaches.begin(); it != atomicNameCacheEnd; ++it) {
            	DynamicNodeListCacheBase* list = it->second;
                oldDocument->unregisterNodeListCache(list); // r124339
                newDocument->registerNodeListCache(list);
            }

            NodeListNameCacheMap::const_iterator nameCacheEnd = m_nameCaches.end();
            for (NodeListNameCacheMap::const_iterator it = m_nameCaches.begin(); it != nameCacheEnd; ++it) {
            	DynamicNodeListCacheBase* list = it->second;
                oldDocument->unregisterNodeListCache(list);
                newDocument->registerNodeListCache(list);
            }

            TagNodeListCacheNS::const_iterator tagEnd = m_tagNodeListCacheNS.end();
            for (TagNodeListCacheNS::const_iterator it = m_tagNodeListCacheNS.begin(); it != tagEnd; ++it) {
            	DynamicNodeListCacheBase* list = it->second;
                ASSERT(!list->isRootedAtDocument());
                oldDocument->unregisterNodeListCache(list);
                newDocument->registerNodeListCache(list);
            }
        }
    }

private:
    NodeListsNodeData()
    : m_childNodeList(0)
    { }

    ALWAYS_INLINE std::pair<unsigned char, AtomicString> namedNodeListKey(unsigned char listType, const AtomicString& name)
    {
        return std::pair<unsigned char, AtomicString>(listType, name);
    }

    ALWAYS_INLINE std::pair<unsigned char, String> namedNodeListKey(unsigned char listType, const String& name)
    {
        return std::pair<unsigned char, String>(listType, name);
    }

    bool deleteThisAndUpdateNodeRareDataIfAboutToRemoveLastList(Node*);

    // FIXME: m_childNodeList should be merged into m_atomicNameCaches or at least be shared with HTMLCollection returned by Element::children
    // but it's tricky because invalidateCaches shouldn't invalidate this cache and adoptTreeScope shouldn't call registerNodeList or unregisterNodeList.
    ChildNodeList* m_childNodeList;
    NodeListAtomicNameCacheMap m_atomicNameCaches;
    NodeListNameCacheMap m_nameCaches;
    TagNodeListCacheNS m_tagNodeListCacheNS;
};

struct NavPropertyData
{
	String values[5];
};

class NodeRareData  : public NodeRareDataBase {
    WTF_MAKE_NONCOPYABLE(NodeRareData); WTF_MAKE_FAST_ALLOCATED;

public:
#if ENABLE(MUTATION_OBSERVERS)
    struct NodeMutationObserverData {
        Vector<OwnPtr<MutationObserverRegistration> > m_registry;
        HashSet<MutationObserverRegistration*> m_transientRegistry;

        static PassOwnPtr<NodeMutationObserverData> create() { return adoptPtr(new NodeMutationObserverData); }
    };
#endif

#if ENABLE(MICRODATA)
    struct NodeMicroDataTokenLists {
        RefPtr<DOMSettableTokenList> m_itemProp;
        RefPtr<DOMSettableTokenList> m_itemRef;
        RefPtr<DOMSettableTokenList> m_itemType;

        static PassOwnPtr<NodeMicroDataTokenLists> create() { return adoptPtr(new NodeMicroDataTokenLists); }
    };
#endif

    unsigned connectedSubframeCount() const { return m_connectedFrameCount; }
    void incrementConnectedSubframeCount(unsigned amount)
    {
        m_connectedFrameCount += amount;
    }
    void decrementConnectedSubframeCount(unsigned amount)
    {
        ASSERT(m_connectedFrameCount);
        ASSERT(amount <= m_connectedFrameCount);
        m_connectedFrameCount -= amount;
    }

    NodeRareData(Document* document)
	: NodeRareDataBase(document)
	, m_tabIndex(0)
    , m_childIndex(0)
	, m_tabIndexWasSetExplicitly(false)
	, m_needsFocusAppearanceUpdateSoonAfterAttach(false)
    , m_styleAffectedByEmpty(false)
    , m_isInCanvasSubtree(false)
#if ENABLE(FULLSCREEN_API)
    , m_containsFullScreenElement(false)
#endif
#if ENABLE(DIALOG_ELEMENT)
	, m_isInTopLayer(false)
#endif
#if ENABLE(SVG)
	, m_hasPendingResources(false)
#endif
    , m_childrenAffectedByHover(false)
    , m_childrenAffectedByActive(false)
    , m_childrenAffectedByDrag(false)
    , m_childrenAffectedByFirstChildRules(false)
    , m_childrenAffectedByLastChildRules(false)
    , m_childrenAffectedByDirectAdjacentRules(false)
    , m_childrenAffectedByForwardPositionalRules(false)
    , m_childrenAffectedByBackwardPositionalRules(false)
	, m_connectedFrameCount(0)
    {
    }

    ALWAYS_INLINE void clearNodeLists() { m_nodeLists.clear(); }
    ALWAYS_INLINE NodeListsNodeData* nodeLists() const { return m_nodeLists.get(); }
    NodeListsNodeData* ensureNodeLists()
    {
        if (!m_nodeLists)
        	m_nodeLists = NodeListsNodeData::create();
        return m_nodeLists.get();
    }

    ALWAYS_INLINE short tabIndex() const { return m_tabIndex; }
    ALWAYS_INLINE void setTabIndexExplicitly(short index) { m_tabIndex = index; m_tabIndexWasSetExplicitly = true; }
    ALWAYS_INLINE bool tabIndexSetExplicitly() const { return m_tabIndexWasSetExplicitly; }
    ALWAYS_INLINE void clearTabIndexExplicitly() { m_tabIndex = 0; m_tabIndexWasSetExplicitly = false; }

#if ENABLE(MUTATION_OBSERVERS)
    Vector<OwnPtr<MutationObserverRegistration> >* mutationObserverRegistry() { return m_mutationObserverData ? &m_mutationObserverData->m_registry : 0; }
    Vector<OwnPtr<MutationObserverRegistration> >* ensureMutationObserverRegistry()
    {
        if (!m_mutationObserverData)
            m_mutationObserverData = NodeMutationObserverData::create();
        return &m_mutationObserverData->m_registry;
    }

    HashSet<MutationObserverRegistration*>* transientMutationObserverRegistry() { return m_mutationObserverData ? &m_mutationObserverData->m_transientRegistry : 0; }
    HashSet<MutationObserverRegistration*>* ensureTransientMutationObserverRegistry()
    {
        if (!m_mutationObserverData)
            m_mutationObserverData = NodeMutationObserverData::create();
        return &m_mutationObserverData->m_transientRegistry;
    }
#endif

#if ENABLE(MICRODATA)
    NodeMicroDataTokenLists* ensureMicroDataTokenLists() const
    {
        if (!m_microDataTokenLists)
            m_microDataTokenLists = NodeMicroDataTokenLists::create();
        return m_microDataTokenLists.get();
    }

    DOMSettableTokenList* itemProp() const
    {
        if (!ensureMicroDataTokenLists()->m_itemProp)
            m_microDataTokenLists->m_itemProp = DOMSettableTokenList::create();

        return m_microDataTokenLists->m_itemProp.get();
    }

    void setItemProp(const String& value)
    {
        if (!ensureMicroDataTokenLists()->m_itemProp)
            m_microDataTokenLists->m_itemProp = DOMSettableTokenList::create();

        m_microDataTokenLists->m_itemProp->setValue(value);
    }

    DOMSettableTokenList* itemRef() const
    {
        if (!ensureMicroDataTokenLists()->m_itemRef)
            m_microDataTokenLists->m_itemRef = DOMSettableTokenList::create();

        return m_microDataTokenLists->m_itemRef.get();
    }

    void setItemRef(const String& value)
    {
        if (!ensureMicroDataTokenLists()->m_itemRef)
            m_microDataTokenLists->m_itemRef = DOMSettableTokenList::create();

        m_microDataTokenLists->m_itemRef->setValue(value);
    }

    DOMSettableTokenList* itemType() const
    {
        if (!ensureMicroDataTokenLists()->m_itemType)
            m_microDataTokenLists->m_itemType = DOMSettableTokenList::create();

        return m_microDataTokenLists->m_itemType.get();
    }

    void setItemType(const String& value)
    {
        if (!ensureMicroDataTokenLists()->m_itemType)
            m_microDataTokenLists->m_itemType = DOMSettableTokenList::create();

        m_microDataTokenLists->m_itemType->setValue(value);
    }
#endif

    ALWAYS_INLINE NavPropertyData* navPropertyData() { return m_navPropertyData.get(); }
    NavPropertyData* ensureNavPropertyData()
    {
        if (!m_navPropertyData)
        	m_navPropertyData = adoptPtr(new NavPropertyData);
        return m_navPropertyData.get();
    }

protected:
    // for ElementRareData
    ALWAYS_INLINE bool needsFocusAppearanceUpdateSoonAfterAttach() const { return m_needsFocusAppearanceUpdateSoonAfterAttach; }
    ALWAYS_INLINE void setNeedsFocusAppearanceUpdateSoonAfterAttach(bool needs) { m_needsFocusAppearanceUpdateSoonAfterAttach = needs; }
    ALWAYS_INLINE bool styleAffectedByEmpty() const { return m_styleAffectedByEmpty; }
    ALWAYS_INLINE void setStyleAffectedByEmpty(bool value) { m_styleAffectedByEmpty = value; }
    ALWAYS_INLINE bool isInCanvasSubtree() const { return m_isInCanvasSubtree; }
    ALWAYS_INLINE void setIsInCanvasSubtree(bool value) { m_isInCanvasSubtree = value; }
#if ENABLE(FULLSCREEN_API)
    ALWAYS_INLINE bool containsFullScreenElement() { return m_containsFullScreenElement; }
    ALWAYS_INLINE void setContainsFullScreenElement(bool value) { m_containsFullScreenElement = value; }
#endif
#if ENABLE(DIALOG_ELEMENT)
    ALWAYS_INLINE bool isInTopLayer() const { return m_isInTopLayer; }
    ALWAYS_INLINE void setIsInTopLayer(bool value) { m_isInTopLayer = value; }
#endif

#if ENABLE(SVG)
    ALWAYS_INLINE bool hasPendingResources() const { return m_hasPendingResources; }
    ALWAYS_INLINE void setHasPendingResources(bool has) { m_hasPendingResources = has; }
#endif

    ALWAYS_INLINE bool childrenAffectedByHover() const { return m_childrenAffectedByHover; }
    ALWAYS_INLINE void setChildrenAffectedByHover(bool value) { m_childrenAffectedByHover = value; }
    ALWAYS_INLINE bool childrenAffectedByActive() const { return m_childrenAffectedByActive; }
    ALWAYS_INLINE void setChildrenAffectedByActive(bool value) { m_childrenAffectedByActive = value; }
    ALWAYS_INLINE bool childrenAffectedByDrag() const { return m_childrenAffectedByDrag; }
    ALWAYS_INLINE void setChildrenAffectedByDrag(bool value) { m_childrenAffectedByDrag = value; }

    ALWAYS_INLINE bool childrenAffectedByFirstChildRules() const { return m_childrenAffectedByFirstChildRules; }
    ALWAYS_INLINE void setChildrenAffectedByFirstChildRules(bool value) { m_childrenAffectedByFirstChildRules = value; }
    ALWAYS_INLINE bool childrenAffectedByLastChildRules() const { return m_childrenAffectedByLastChildRules; }
    ALWAYS_INLINE void setChildrenAffectedByLastChildRules(bool value) { m_childrenAffectedByLastChildRules = value; }
    ALWAYS_INLINE bool childrenAffectedByDirectAdjacentRules() const { return m_childrenAffectedByDirectAdjacentRules; }
    ALWAYS_INLINE void setChildrenAffectedByDirectAdjacentRules(bool value) { m_childrenAffectedByDirectAdjacentRules = value; }
    ALWAYS_INLINE bool childrenAffectedByForwardPositionalRules() const { return m_childrenAffectedByForwardPositionalRules; }
    ALWAYS_INLINE void setChildrenAffectedByForwardPositionalRules(bool value) { m_childrenAffectedByForwardPositionalRules = value; }
    ALWAYS_INLINE bool childrenAffectedByBackwardPositionalRules() const { return m_childrenAffectedByBackwardPositionalRules; }
    ALWAYS_INLINE void setChildrenAffectedByBackwardPositionalRules(bool value) { m_childrenAffectedByBackwardPositionalRules = value; }
    ALWAYS_INLINE unsigned childIndex() const { return m_childIndex; }
    ALWAYS_INLINE void setChildIndex(unsigned index) { m_childIndex = index; }

private:
    OwnPtr<NavPropertyData> m_navPropertyData;

    short m_tabIndex;
    unsigned short m_childIndex;
    unsigned m_tabIndexWasSetExplicitly : 1;
    unsigned m_needsFocusAppearanceUpdateSoonAfterAttach : 1;
    unsigned m_styleAffectedByEmpty : 1;
    unsigned m_isInCanvasSubtree : 1;
#if ENABLE(FULLSCREEN_API)
    unsigned m_containsFullScreenElement : 1;
#endif
#if ENABLE(DIALOG_ELEMENT)
    unsigned m_isInTopLayer : 1;
#endif
#if ENABLE(SVG)
    unsigned m_hasPendingResources : 1;
#endif
    unsigned m_childrenAffectedByHover : 1;
    unsigned m_childrenAffectedByActive : 1;
    unsigned m_childrenAffectedByDrag : 1;
    // Bits for dynamic child matching.
    // We optimize for :first-child and :last-child. The other positional child selectors like nth-child or
    // *-child-of-type, we will just give up and re-evaluate whenever children change at all.
    unsigned m_childrenAffectedByFirstChildRules : 1;
    unsigned m_childrenAffectedByLastChildRules : 1;
    unsigned m_childrenAffectedByDirectAdjacentRules : 1;
    unsigned m_childrenAffectedByForwardPositionalRules : 1;
    unsigned m_childrenAffectedByBackwardPositionalRules : 1;
    unsigned m_connectedFrameCount : 10; // Must fit Page::maxNumberOfFrames.

private:
    OwnPtr<NodeListsNodeData> m_nodeLists;

#if ENABLE(MUTATION_OBSERVERS)
    OwnPtr<NodeMutationObserverData> m_mutationObserverData;
#endif

#if ENABLE(MICRODATA)
    mutable OwnPtr<NodeMicroDataTokenLists> m_microDataTokenLists;
#endif
};

inline bool NodeListsNodeData::deleteThisAndUpdateNodeRareDataIfAboutToRemoveLastList(Node* ownerNode)
{
    ASSERT(ownerNode);
    ASSERT(ownerNode->nodeLists() == this);
    if ((m_childNodeList ? 1 : 0) + m_atomicNameCaches.size() + m_nameCaches.size() + m_tagNodeListCacheNS.size() != 1)
        return false;
    ownerNode->clearNodeLists();
    return true;
}

// Ensure the 10 bits reserved for the m_connectedFrameCount cannot overflow
COMPILE_ASSERT(Page::maxNumberOfFrames < 1024, Frame_limit_should_fit_in_rare_data_count);

} // namespace WebCore

#endif // NodeRareData_h
