/*
 * Copyright (C) 2000 Lars Knoll (knoll@kde.org)
 *           (C) 2000 Antti Koivisto (koivisto@kde.org)
 *           (C) 2000 Dirk Mueller (mueller@kde.org)
 * Copyright (C) 2003, 2004, 2005, 2006, 2007, 2008 Apple Inc. All rights reserved.
 * Copyright (C) 2007 Nicholas Shanks <webkit@nickshanks.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIother.m_  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USm_
 *
 */

#ifndef FontDescription_h
#define FontDescription_h

#include "FontFamily.h"
#include "FontFeatureSettings.h"
#include "FontOrientation.h"
#include "FontRenderingMode.h"
#include "FontSmoothingMode.h"
#include "FontTraitsMask.h"
#include "FontWidthVariant.h"
#include "TextOrientation.h"
#include "TextRenderingMode.h"
#include "WebKitFontFamilyNames.h"
#include <wtf/MathExtras.h>

#include <wtf/RefPtr.h>

namespace WebCore {

using namespace WebKitFontFamilyNames;

enum FontWeight {
    FontWeight100,
    FontWeight200,
    FontWeight300,
    FontWeight400,
    FontWeight500,
    FontWeight600,
    FontWeight700,
    FontWeight800,
    FontWeight900,
    FontWeightNormal = FontWeight400,
    FontWeightBold = FontWeight700
};

enum FontItalic {
    FontItalicOff = 0,
    FontItalicOn = 1
};

enum FontSmallCaps {
    FontSmallCapsOff = 0,
    FontSmallCapsOn = 1
};

class FontDescription {
public:
    enum GenericFamilyType { NoFamily, StandardFamily, SerifFamily, SansSerifFamily, 
                             MonospaceFamily, CursiveFamily, FantasyFamily, PictographFamily };

    enum Kerning { AutoKerning, NormalKerning, NoneKerning };

    enum LigaturesState { NormalLigaturesState, DisabledLigaturesState, EnabledLigaturesState };

    FontDescription()
		: m_families(1)
		, m_specifiedSize(0)
        , m_computedSize(0)
        , m_orientation(Horizontal)
        , m_textOrientation(TextOrientationVerticalRight)
        , m_widthVariant(RegularWidth)
        , m_italic(FontItalicOff)
        , m_smallCaps(FontSmallCapsOff)
        , m_isAbsoluteSize(false)
        , m_weight(FontWeightNormal)
        , m_genericFamily(NoFamily)
        , m_usePrinterFont(false)
        , m_renderingMode(NormalRenderingMode)
        , m_kerning(AutoKerning)
        , m_commonLigaturesState(NormalLigaturesState)
        , m_discretionaryLigaturesState(NormalLigaturesState)
        , m_historicalLigaturesState(NormalLigaturesState)
        , m_keywordSize(0)
        , m_fontSmoothing(AutoSmoothing)
        , m_textRendering(AutoTextRendering)
        , m_isSpecifiedFont(false)
        , m_script(USCRIPT_COMMON)
    {
    }

    bool operator==(const FontDescription&) const;
    ALWAYS_INLINE bool operator!=(const FontDescription& other) const { return !(*this == other); }

    unsigned familyCount() const { return m_families.size(); }
    const AtomicString& firstFamily() const { return familyAt(0); }
    const AtomicString& familyAt(unsigned i) const { return m_families[i]; }
    const Vector<AtomicString, 1>& families() const { return m_families; }

    ALWAYS_INLINE float specifiedSize() const { return m_specifiedSize; }
    ALWAYS_INLINE float computedSize() const { return m_computedSize; }
    ALWAYS_INLINE FontItalic italic() const { return static_cast<FontItalic>(m_italic); }
    ALWAYS_INLINE int computedPixelSize() const { return int(m_computedSize + 0.5f); }
    ALWAYS_INLINE FontSmallCaps smallCaps() const { return static_cast<FontSmallCaps>(m_smallCaps); }
    ALWAYS_INLINE bool isAbsoluteSize() const { return m_isAbsoluteSize; }
    ALWAYS_INLINE FontWeight weight() const { return static_cast<FontWeight>(m_weight); }
    FontWeight lighterWeight() const;
    FontWeight bolderWeight() const;
    ALWAYS_INLINE GenericFamilyType genericFamily() const { return static_cast<GenericFamilyType>(m_genericFamily); }
    ALWAYS_INLINE bool usePrinterFont() const { return m_usePrinterFont; }
    // only use fixed default size when there is only one font family, and that family is "monospace"
    ALWAYS_INLINE bool useFixedDefaultSize() const { return genericFamily() == MonospaceFamily && familyCount() == 1 && firstFamily() == monospaceFamily; }
    ALWAYS_INLINE FontRenderingMode renderingMode() const { return static_cast<FontRenderingMode>(m_renderingMode); }
    ALWAYS_INLINE Kerning kerning() const { return static_cast<Kerning>(m_kerning); }
    ALWAYS_INLINE LigaturesState commonLigaturesState() const { return static_cast<LigaturesState>(m_commonLigaturesState); }
    ALWAYS_INLINE LigaturesState discretionaryLigaturesState() const { return static_cast<LigaturesState>(m_discretionaryLigaturesState); }
    ALWAYS_INLINE LigaturesState historicalLigaturesState() const { return static_cast<LigaturesState>(m_historicalLigaturesState); }
    ALWAYS_INLINE unsigned keywordSize() const { return m_keywordSize; }
    ALWAYS_INLINE FontSmoothingMode fontSmoothing() const { return static_cast<FontSmoothingMode>(m_fontSmoothing); }
    ALWAYS_INLINE TextRenderingMode textRenderingMode() const { return static_cast<TextRenderingMode>(m_textRendering); }
    ALWAYS_INLINE UScriptCode script() const { return static_cast<UScriptCode>(m_script); }

    FontTraitsMask traitsMask() const;
    ALWAYS_INLINE bool isSpecifiedFont() const { return m_isSpecifiedFont; }
    ALWAYS_INLINE FontOrientation orientation() const { return static_cast<FontOrientation>(m_orientation); }
    ALWAYS_INLINE TextOrientation textOrientation() const { return static_cast<TextOrientation>(m_textOrientation); }
    ALWAYS_INLINE FontWidthVariant widthVariant() const { return static_cast<FontWidthVariant>(m_widthVariant); }
    ALWAYS_INLINE FontFeatureSettings* featureSettings() const { return m_featureSettings.get(); }
    FontDescription makeNormalFeatureSettings() const;

    ALWAYS_INLINE void setOneFamily(const AtomicString& family) { ASSERT(m_families.size() == 1); m_families[0] = family; }
    ALWAYS_INLINE void setFamilies(const Vector<AtomicString, 1>& families) { m_families = families; }
    ALWAYS_INLINE void adoptFamilies(Vector<AtomicString, 1>& families) { m_families.swap(families); }
    ALWAYS_INLINE void setComputedSize(float s) { m_computedSize = clampToFloat(s); }
    ALWAYS_INLINE void setSpecifiedSize(float s) { m_specifiedSize = clampToFloat(s); }
    ALWAYS_INLINE void setItalic(FontItalic i) { m_italic = i; }
    ALWAYS_INLINE void setItalic(bool i) { setItalic(i ? FontItalicOn : FontItalicOff); }
    ALWAYS_INLINE void setSmallCaps(FontSmallCaps c) { m_smallCaps = c; }
    ALWAYS_INLINE void setSmallCaps(bool c) { setSmallCaps(c ? FontSmallCapsOn : FontSmallCapsOff); }
    ALWAYS_INLINE void setIsAbsoluteSize(bool s) { m_isAbsoluteSize = s; }
    ALWAYS_INLINE void setWeight(FontWeight w) { m_weight = w; }
    ALWAYS_INLINE void setGenericFamily(GenericFamilyType genericFamily) { m_genericFamily = genericFamily; }
#if PLATFORM(CHROMIUM) && OS(DARWIN)
    ALWAYS_INLINE void setUsePrinterFont(bool) { }
#else
    ALWAYS_INLINE void setUsePrinterFont(bool p) { m_usePrinterFont = p; }
#endif
    ALWAYS_INLINE void setRenderingMode(FontRenderingMode mode) { m_renderingMode = mode; }
    ALWAYS_INLINE void setKerning(Kerning kerning) { m_kerning = kerning; }
    ALWAYS_INLINE void setCommonLigaturesState(LigaturesState commonLigaturesState) { m_commonLigaturesState = commonLigaturesState; }
    ALWAYS_INLINE void setDiscretionaryLigaturesState(LigaturesState discretionaryLigaturesState) { m_discretionaryLigaturesState = discretionaryLigaturesState; }
    ALWAYS_INLINE void setHistoricalLigaturesState(LigaturesState historicalLigaturesState) { m_historicalLigaturesState = historicalLigaturesState; }
    ALWAYS_INLINE void setKeywordSize(unsigned s) { m_keywordSize = s; }
    ALWAYS_INLINE void setFontSmoothing(FontSmoothingMode smoothing) { m_fontSmoothing = smoothing; }
    ALWAYS_INLINE void setTextRenderingMode(TextRenderingMode rendering) { m_textRendering = rendering; }
    ALWAYS_INLINE void setIsSpecifiedFont(bool isSpecifiedFont) { m_isSpecifiedFont = isSpecifiedFont; }
    ALWAYS_INLINE void setOrientation(FontOrientation orientation) { m_orientation = orientation; }
    ALWAYS_INLINE void setTextOrientation(TextOrientation textOrientation) { m_textOrientation = textOrientation; }
    ALWAYS_INLINE void setWidthVariant(FontWidthVariant widthVariant) { m_widthVariant = widthVariant; }
    ALWAYS_INLINE void setScript(UScriptCode s) { m_script = s; }
    ALWAYS_INLINE void setFeatureSettings(PassRefPtr<FontFeatureSettings> settings) { m_featureSettings = settings; }

private:
    Vector<AtomicString, 1> m_families;
    RefPtr<FontFeatureSettings> m_featureSettings;

    float m_specifiedSize;   // Specified CSS value. Independent of rendering issues such as integer
                             // rounding, minimum font sizes, and zooming.
    float m_computedSize;    // Computed size adjusted for the minimum font size and the zoom factor.  

    unsigned m_orientation : 1; // FontOrientation - Whether the font is rendering on a horizontal line or a vertical line.
    unsigned m_textOrientation : 1; // TextOrientation - Only used by vertical text. Determines the default orientation for non-ideograph glyphs.

    unsigned m_widthVariant : 2; // FontWidthVariant

    unsigned m_italic : 1; // FontItalic
    unsigned m_smallCaps : 1; // FontSmallCaps
    unsigned m_isAbsoluteSize : 1; // Whether or not CSS specified an explicit size
                                  // (logical sizes like "medium" don't count).
    unsigned m_weight : 8; // FontWeight
    unsigned m_genericFamily : 3; // GenericFamilyType
    unsigned m_usePrinterFont : 1;

    unsigned m_renderingMode : 1;  // Used to switch between CG and GDI text on Windows.
    unsigned m_kerning : 2; // Kerning

    unsigned m_commonLigaturesState : 2;
    unsigned m_discretionaryLigaturesState : 2;
    unsigned m_historicalLigaturesState : 2;

    unsigned m_keywordSize : 4; // We cache whether or not a font is currently represented by a CSS keyword (e.g., medium).  If so,
                           // then we can accurately translate across different generic families to adjust for different preference settings
                           // (e.g., 13px monospace vs. 16px everything else).  Sizes are 1-8 (like the HTML size values for <font>).

    unsigned m_fontSmoothing : 2; // FontSmoothingMode
    unsigned m_textRendering : 2; // TextRenderingMode
    unsigned m_isSpecifiedFont : 1; // True if a web page specifies a non-generic font family as the first font family.
    unsigned m_script : 7; // Used to help choose an appropriate font for generic font families.
};

inline bool FontDescription::operator==(const FontDescription& other) const
{
    return m_families == other.m_families
        && m_specifiedSize == other.m_specifiedSize
        && m_computedSize == other.m_computedSize
        && m_italic == other.m_italic
        && m_smallCaps == other.m_smallCaps
        && m_isAbsoluteSize == other.m_isAbsoluteSize
        && m_weight == other.m_weight
        && m_genericFamily == other.m_genericFamily
        && m_usePrinterFont == other.m_usePrinterFont
        && m_renderingMode == other.m_renderingMode
        && m_kerning == other.m_kerning
        && m_commonLigaturesState == other.m_commonLigaturesState
        && m_discretionaryLigaturesState == other.m_discretionaryLigaturesState
        && m_historicalLigaturesState == other.m_historicalLigaturesState
        && m_keywordSize == other.m_keywordSize
        && m_fontSmoothing == other.m_fontSmoothing
        && m_textRendering == other.m_textRendering
        && m_isSpecifiedFont == other.m_isSpecifiedFont
        && m_orientation == other.m_orientation
        && m_textOrientation == other.m_textOrientation
        && m_widthVariant == other.m_widthVariant
        && m_script == other.m_script
        && m_featureSettings == other.m_featureSettings;
}

}

#endif
