/*
 * Copyright (C) 2006 Apple Computer, Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer. 
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution. 
 * 3.  Neither the name of Apple Computer, Inc. ("Apple") nor the names of
 *     its contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE AND ITS CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "FontFallbackList.h"

#include "Font.h"
#include "FontCache.h"
#include "SegmentedFontData.h"
#include <wtf/unicode/Unicode.h>

namespace WebCore {

FontGlyphs::FontGlyphs()
: m_pageZero(0)
, m_cachedPrimarySimpleFontData(0)
, m_fontSelector(0)
, m_fontSelectorVersion(0)
, m_familyIndex(0)
, m_generation(fontCache()->generation())
, m_pitch(UnknownPitch)
, m_loadingCustomFonts(false)
{
}

FontGlyphs::FontGlyphs(const FontPlatformData& platformData)
    : m_pageZero(0)
    , m_cachedPrimarySimpleFontData(0)
    , m_fontSelector(0)
    , m_fontSelectorVersion(0)
    , m_familyIndex(cAllFamiliesScanned)
    , m_generation(fontCache()->generation())
    , m_pitch(UnknownPitch)
    , m_loadingCustomFonts(false)
{
    RefPtr<FontData> fontData = fontCache()->getCachedFontData(&platformData);
    m_realizedFontData.append(fontData.release());
}

void FontGlyphs::invalidate(PassRefPtr<FontSelector> fontSelector)
{
    releaseFontData();
    m_realizedFontData.clear();
    m_pageZero = 0;
    m_pages.clear();
    m_cachedPrimarySimpleFontData = 0;
    m_familyIndex = 0;    
    m_pitch = UnknownPitch;
    m_loadingCustomFonts = false;
    m_fontSelector = fontSelector;
    m_fontSelectorVersion = m_fontSelector ? m_fontSelector->version() : 0;
    m_generation = fontCache()->generation();
    m_widthCache.clear();
}

void FontGlyphs::releaseFontData()
{
    unsigned numFonts = m_realizedFontData.size();
    for (unsigned i = 0; i < numFonts; ++i) {
        if (m_realizedFontData[i]->isCustomFont())
        	continue;
		ASSERT(!m_realizedFontData[i]->isSegmented());
		fontCache()->releaseFontData(static_cast<const SimpleFontData*>(m_realizedFontData[i].get()));
    }
}

void FontGlyphs::determinePitch(const FontDescription& description) const
{
    const FontData* fontData = primaryFontData(description);
    if (!fontData->isSegmented())
        m_pitch = static_cast<const SimpleFontData*>(fontData)->pitch();
    else {
        const SegmentedFontData* segmentedFontData = static_cast<const SegmentedFontData*>(fontData);
        unsigned numRanges = segmentedFontData->numRanges();
        if (numRanges == 1)
            m_pitch = segmentedFontData->rangeAt(0).fontData()->pitch();
        else
            m_pitch = VariablePitch;
    }
}

DLL_PUBLIC const FontData* FontGlyphs::realizeFontDataAt(const FontDescription& description, unsigned realizedFontIndex) const
{
    if (realizedFontIndex < m_realizedFontData.size())
        return m_realizedFontData[realizedFontIndex].get(); // This fallback font is already in our list.

    // Make sure we're not passing in some crazy value here.
    ASSERT(realizedFontIndex == m_realizedFontData.size());
    
    if(m_familyIndex == cAllFamiliesScanned)
    	return 0;

    RefPtr<FontData> result = fontCache()->getFontData(description, m_familyIndex, m_fontSelector.get());
    if (result) {
        m_realizedFontData.append(result);
        if (result->isLoading())
            m_loadingCustomFonts = true;
    }
    return result.get();
}

void FontGlyphs::addFontData(RefPtr<FontData> data)
{
	if(m_realizedFontData.contains(data))
		return;
    m_realizedFontData.append(data);
}

/**
 * Finds a visually similar character to use in place of the given character.
 * Returns 0 if no character could be found.
 * The tryAgain flag will be set to true if we should search again for
 * another fallback character if the given fallback is not found.
 */
static UChar32 getFallbackCharacter(UChar32 c, bool& tryAgain) {
    tryAgain = false;

    // map full-width punctuation and latin characters to their
    // ascii counterparts
    if (c >= 0xff01 && c <= 0xff5e)
        return c - 0xfee0;

    // if we lack support for a combining character, better to leave it
    // off than to show a box
    if (c >= 0x0300 && c <= 0x036f)
        return zeroWidthSpace;
    if (c >= 0x1dc0 && c <= 0x1dff)
        return zeroWidthSpace;
    if (c >= 0x20d0 && c <= 0x20ff)
        return zeroWidthSpace;

    switch (c) {

        // note: many of these fallbacks come from the unicode standard
        // version 6.1 "general punctuation" range 0x2000-0x206f

        // all non-rendering control characters should map to zero width space
        case 0:
        case leftToRightEmbed:
        case leftToRightOverride:
        case rightToLeftEmbed:
        case rightToLeftOverride:
        case zeroWidthNonJoiner:
        case zeroWidthJoiner:
        case popDirectionalFormatting:
        case leftToRightMark:
        case rightToLeftMark:
        case 0xfeff: // zero width no-break space
        case 0xfffe: // byte-order marker, should not occur in text!
        case 0xfffc: // object replacement character
        case 0x2028: // line separator
        case 0x2029: // paragraph separator
        case 0x2060: // word joiner
        case 0x2061: // function application
        case 0x2062: // invisible times
        case 0x2063: // invisible separator
        case 0x2064: // invisible plus
        case 0x206a: // inhibit symmetric swapping
        case 0x206b: // activate symmetric swapping
        case 0x206c: // inhibit arabic form shaping
        case 0x206d: // activate arabic form shaping
        case 0x206e: // national digit shapes
        case 0x206f: // nominal digit shapes
            return zeroWidthSpace;

        case 0x30fb: // katakana middle dot
        case 0x180e: // mongolian vowel separator
        case 0x2022: // bullet
        case 0x2219: // bullet operator
        case 0x2027: // hyphenation point
            return 0x00b7; // middle dot

        case 0x00ad: // soft hyphen
        case 0x2010: // hyphen
        case 0x2011: // non-breaking hyphen
        case 0x2012: // figure dash
        case 0x2013: // en dash
        case 0x2014: // em dash
        case 0x2015: // horizontal bar
        case 0x30fc: // katakana-hiragana prolonged sound mark
        case 0x1680: // ogham space mark
        case 0xfe58: // small em dash
        case 0xfe63: // small hyphen-minus
            return '-'; // hyphen-minus

        case 0x30a0: // katakana-hiragana double hyphen
        case 0x1400: // canadian syllabics hyphen
            return '='; // equal sign

        case 0x00a0: // non-break space
        case 0x2000: // en quad
        case 0x2001: // em quad
        case 0x2002: // en space
        case 0x2003: // em space
        case 0x2004: // three-per-em space
        case 0x2005: // four-per-em space
        case 0x2006: // six-per-em space
        case 0x2007: // figure space
        case 0x2008: // punctuation space
        case 0x2009: // thin space
        case 0x200a: // hair space
        case 0x202f: // narrow no-break space
        case 0x205f: // medium mathematical space
        case 0x3000: // ideographic space
        case 0xffa0: // halfwidth hangul filler
        case 0xfff9: // interlinear annotation anchor
        case 0xfffa: // interlinear annotation separator
        case 0xfffb: // interlinear annotation terminator
            return ' '; // space

        case 0x00b4: // acute accent
        case 0x02b9: // modifier letter prime
        case 0x02bd: // modifier letter reversed comma
        case 0x2032: // prime
        case 0x2035: // reversed prime
        case 0x2018: // left single quotation mark
        case 0x2019: // right single quotation mark
        case 0x201b: // single high-reversed-9 quotation mark
            return '\''; // apostrophe

        case 0x201a: // single low-9 quotation mark
            return ','; // comma

        case 0x201c: // left double quotation mark
        case 0x201d: // right double quotation mark
        case 0x201e: // double low-9 quotation mark
        case 0x201f: // double high-reversed-9 quotation mark
        case 0x2033: // double prime
        case 0x2036: // reversed double prime
        case 0x275d: // heavy double turned comma quotation mark ornament
        case 0x275e: // heavy double comma quotation mark ornament
        case 0x301e: // double prime quotation mark
            return '"'; // quotation mark

        case 0x301c: // WAVE DASH
        	return 0xFF5E; // FULLWIDTH TILDE

        case 0x2024: // one dot leader
        case 0x3002: // ideographic full stop
        case 0xff61: // halfwidth ideographic full stop
            return '.'; // full stop

        case 0x2038: // caret
        case 0x2303: // up arrowhead
            return '^'; // circumflex accent

        case 0x2053: // swung dash
            return '~'; // tilde

        case 0x2039: // single left-pointing angle quotation mark
        case 0x2329: // left-pointing angle bracket
        case 0x3008: // left angle bracket
            return '<'; // less-than sign

        case 0x203a: // single right-pointing angle quotation mark
        case 0x232a: // right-pointing angle bracket
        case 0x3009: // right angle bracket
            return '>'; // greater-than sign

        case 0x2044: // fraction slash
        case 0x2215: // division slash
        case 0x29f8: // big solidus
            return '/'; // solidus

        case 0x2017: // double low line
            return '_'; // low line

        case 0x204f: // reversed semicolon
            return ';'; // semicolon

        case 0x02d0: // modifier letter triangular colon
        case 0x05c3: // hebrew punctuation sof pasuq
        case 0x2236: // ratio
            return ':'; // colon

        case 0x2e18: // inverted interrobang
            tryAgain = true;
            return 0x00bf; // inverted question mark

        case 0x203d: // interrobang
        case 0x00bf: // inverted question mark
            return '?'; // question mark

        case 0x2052: // commercial minus sign
        case 0x066a: // arabic percent sign
            return '%'; // percent sign

        case 0x204a: // tironian sign et
            return '&'; // ampersand

        case 0x204b: // reversed pilcrow sign
            return 0x00b6; // pilcrow sign

        case 0x204e: // low asterisk
        case 0x2055: // flower punctuation mark
        case 0x274b: // heavy eight teardrop-spoked propeller asterisk
            return '*'; // asterisk

        case 0x01c3: // latin letter retroflex click
            return '!'; // exclamation mark

        case 0xffe0: // fullwidth cent sign
            return 0x00a2; // cent sign

        case 0xffe1: // fullwidth pound sign
            return 0x00a3; // pound sign

        case 0xffe5: // fullwidth yen sign
            return 0x00a5; // yen sign

        case 0xffe6: // fullwidth won sign
            return 0x20a9; // won sign

        default:
            break;
    }

    // no fallback character found
    return 0;
}

std::pair<GlyphData, GlyphPage*> FontGlyphs::glyphDataAndPageForCharacter(const Font& font, UChar32 c, bool mirror, FontDataVariant variant, bool useSimilarChar) const
{
    ASSERT(isMainThread());

    const FontDescription& description = font.fontDescription();
    if (variant == AutoVariant) {
        if (description.smallCaps() && !primarySimpleFontData(description)->isSVGFont()) {
            UChar32 upperC = u_toupper(c);
            if (upperC != c) {
                c = upperC;
                variant = SmallCapsVariant;
            } else
                variant = NormalVariant;
        } else
            variant = NormalVariant;
    }

    if (mirror)
        c = u_charMirror(c);

    unsigned pageNumber = (c / GlyphPage::size);

    GlyphPageTreeNode* node = pageNumber ? m_pages.get(pageNumber) : m_pageZero;
    if (!node) {
        node = GlyphPageTreeNode::getRootChild(realizeFontDataAt(description, 0), pageNumber);
        if (pageNumber)
            m_pages.set(pageNumber, node);
        else
            m_pageZero = node;
    }

    GlyphPage* page = 0;
    if (variant == NormalVariant) {
        // Fastest loop, for the common case (normal variant).
        while (true) {
            page = node->page();
            if (page) {
                GlyphData data = page->glyphDataForCharacter(c);
                if (data.fontData && (data.fontData->platformData().orientation() == Horizontal || data.fontData->isTextOrientationFallback()))
                    return std::make_pair(data, page);

                if (data.fontData) {
                    if (font.isCJKIdeographOrSymbol(c)) {
                        if (!data.fontData->hasVerticalGlyphs()) {
                            // Use the broken ideograph font data. The broken ideograph font will use the horizontal width of glyphs
                            // to make sure you get a square (even for broken glyphs like symbols used for punctuation).
                            variant = BrokenIdeographVariant;
                            break;
                        }
                    } else {
                        if (description.textOrientation() == TextOrientationVerticalRight) {
                        	RefPtr<SimpleFontData> verticalRightFontData = data.fontData->verticalRightOrientationFontData();
                            GlyphPageTreeNode* verticalRightNode = GlyphPageTreeNode::getRootChild(verticalRightFontData.get(), pageNumber);
                            GlyphPage* verticalRightPage = verticalRightNode->page();
                            if (verticalRightPage) {
                                GlyphData verticalRightData = verticalRightPage->glyphDataForCharacter(c);
                                // If the glyphs are distinct, we will make the assumption that the font has a vertical-right glyph baked
                                // into it.
                                if (data.glyph != verticalRightData.glyph)
                                    return std::make_pair(data, page);
                                // The glyphs are identical, meaning that we should just use the horizontal glyph.
                                if (verticalRightData.fontData)
                                    return std::make_pair(verticalRightData, verticalRightPage);
                            }
                        } else if (description.textOrientation() == TextOrientationUpright) {
                        	RefPtr<SimpleFontData> uprightFontData = data.fontData->uprightOrientationFontData();
                            GlyphPageTreeNode* uprightNode = GlyphPageTreeNode::getRootChild(uprightFontData.get(), pageNumber);
                            GlyphPage* uprightPage = uprightNode->page();
                            if (uprightPage) {
                                GlyphData uprightData = uprightPage->glyphDataForCharacter(c);
                                // If the glyphs are the same, then we know we can just use the horizontal glyph rotated vertically to be upright.
                                if (data.glyph == uprightData.glyph)
                                    return std::make_pair(data, page);
                                // The glyphs are distinct, meaning that the font has a vertical-right glyph baked into it. We can't use that
                                // glyph, so we fall back to the upright data and use the horizontal glyph.
                                if (uprightData.fontData)
                                    return std::make_pair(uprightData, uprightPage);
                            }
                        }

                        // Shouldn't be possible to even reach this point.
                        ASSERT_NOT_REACHED();
                    }
                    return std::make_pair(data, page);
                }

                if (node->isSystemFallback())
                    break;
            }

            node = node->getChild(realizeFontDataAt(description, node->level()), pageNumber);
            if (pageNumber)
                m_pages.set(pageNumber, node);
            else
                m_pageZero = node;
        }
    }
    if (variant != NormalVariant) {
        while (true) {
            page = node->page();
            if (page) {
                GlyphData data = page->glyphDataForCharacter(c);
                if (data.fontData) {
                    // The variantFontData function should not normally return 0.
                    // But if it does, we will just render the capital letter big.
                	RefPtr<SimpleFontData> variantFontData = data.fontData->variantFontData(description, variant);
                    if (!variantFontData)
                        return std::make_pair(data, page);

                    GlyphPageTreeNode* variantNode = GlyphPageTreeNode::getRootChild(variantFontData.get(), pageNumber);
                    GlyphPage* variantPage = variantNode->page();
                    if (variantPage) {
                        GlyphData data = variantPage->glyphDataForCharacter(c);
                        if (data.fontData)
                            return std::make_pair(data, variantPage);
                    }

                    // Do not attempt system fallback off the variantFontData. This is the very unlikely case that
                    // a font has the lowercase character but the small caps font does not have its uppercase version.
                    return std::make_pair(variantFontData->missingGlyphData(), page);
                }

                if (node->isSystemFallback())
                    break;
            }

            node = node->getChild(realizeFontDataAt(description, node->level()), pageNumber);
            if (pageNumber)
                m_pages.set(pageNumber, node);
            else
                m_pageZero = node;
        }
    }

    ASSERT(page);
    ASSERT(node->isSystemFallback());

    // System fallback is character-dependent. When we get here, we
    // know that the character in question isn't in the system fallback
    // font's glyph page. Try to lazily create it here.
    UChar codeUnits[2];
    int codeUnitsLength;
    if (c <= 0xFFFF) {
        if (font.treatAsZeroWidthSpace(c))
            return std::make_pair(GlyphData(), page); // non-rendering

        codeUnits[0] = Font::normalizeSpaces(c);
        codeUnitsLength = 1;
    } else {
        codeUnits[0] = U16_LEAD(c);
        codeUnits[1] = U16_TRAIL(c);
        codeUnitsLength = 2;
    }
    RefPtr<SimpleFontData> characterFontData = fontCache()->getFontDataForCharacters(font, codeUnits, codeUnitsLength);
    if (characterFontData) {
        if (characterFontData->platformData().orientation() == Vertical && !characterFontData->hasVerticalGlyphs() && font.isCJKIdeographOrSymbol(c))
            variant = BrokenIdeographVariant;
        if (variant != NormalVariant)
            characterFontData = characterFontData->variantFontData(description, variant);
    }
    if (characterFontData) {
        // Got the fallback glyph and font.
        GlyphPage* fallbackPage = GlyphPageTreeNode::getRootChild(characterFontData.get(), pageNumber)->page();
        if(fallbackPage && fallbackPage->fontDataForCharacter(c)) {
			GlyphData data = fallbackPage->glyphDataForCharacter(c);
			// Cache it so we don't have to do system fallback again next time.
			if (variant == NormalVariant) {
				page->setGlyphDataForCharacter(c, data.glyph, data.fontData);
				data.fontData->setMaxGlyphPageTreeLevel(std::max(data.fontData->maxGlyphPageTreeLevel(), node->level()));
			}
        	const_cast<FontGlyphs*>(this)->addFontData(const_cast<SimpleFontData*>(data.fontData));
        	return std::make_pair(data, page);
        }
    }

    // Glyph was not found in any font. Try to replace it with a
    // visually similar character.
    if(c == zeroWidthSpace)
        return std::make_pair(GlyphData(), page); // non-rendering

    if(useSimilarChar){
        bool tryAgain = false;
        UChar32 replacement = getFallbackCharacter(c, tryAgain);
        if (replacement == zeroWidthSpace)
            return std::make_pair(GlyphData(), page); // non-rendering

        if (replacement)
            return glyphDataAndPageForCharacter(font, replacement, mirror, variant, tryAgain);

        FontCharReplaceMap::const_iterator it = fontCache()->similarCharacters().find(c);
		if(it != fontCache()->similarCharacters().end())
			return glyphDataAndPageForCharacter(font, it->second, mirror, variant, false);
    }

    UChar32 missingCharacter = fontCache()->missingCharacter(c);
    if(missingCharacter > 0)
    	return glyphDataAndPageForCharacter(font, missingCharacter, mirror, variant, false);

    // Even system fallback can fail; use the missing glyph in that case.
    // FIXME: It would be nicer to use the missing glyph from the last resort font instead.
    GlyphData data = primarySimpleFontData(description)->missingGlyphData();
    if (variant == NormalVariant) {
        page->setGlyphDataForCharacter(c, data.glyph, data.fontData);
        data.fontData->setMaxGlyphPageTreeLevel(std::max(data.fontData->maxGlyphPageTreeLevel(), node->level()));
    }
    return std::make_pair(data, page);
}

}
