/*
 * Copyright (C) 2008 Alp Toker <alp@atoker.com>
 * Copyright (C) 2010 Igalia S.L.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 *
 */

#include "config.h"
#include "FontCustomPlatformData.h"

#include "FontPlatformData.h"
#include "MemoryCache.h"
#include "SharedBuffer.h"
#include "WOFFFileFormat.h"
#include "WOFF2FileFormat.h"
#include <cairo-ft.h>
#include <cairo.h>

namespace WebCore {

DLL_PUBLIC int gFontCustomPlatformDataCount;

static void releaseCustomFontData(void* data)
{
    static_cast<SharedBuffer*>(data)->deref();
}

FontCustomPlatformData::FontCustomPlatformData(FT_Face freeTypeFace, SharedBuffer* buffer)
    : m_freeTypeFace(freeTypeFace)
    , m_fontFace(0)
	, m_simpleFontDataPruneTimer(this, &FontCustomPlatformData::simpleFontDataPruneTimerCallback)
{
    // FT_LOAD_FORCE_AUTOHINT prohibits use of the font's native hinting. This
    // is a safe option for custom fonts because (a) some such fonts may have
    // broken hinting, which site admins may not notice if other browsers do not
    // use the native hints, and (b) allowing native hints exposes the FreeType
    // bytecode interpreter to potentially-malicious input. Treating web fonts
    // differently than system fonts is non-ideal, but the result of autohinting
    // is always decent, whereas native hints sometimes look terrible, and
    // unlike system fonts where Fontconfig may change the hinting settings on a
    // per-font basis, the same settings are used for all web fonts. Note that
    // Chrome is considering switching from autohinting to native hinting in
    // https://code.google.com/p/chromium/issues/detail?id=173207 but this is
    // more risk than we want to assume for now. See
    // https://bugs.webkit.org/show_bug.cgi?id=140994 before changing this.

	// Disable bitmap strikes in all downloaded fonts if outlines are available.
	// This matches the behavior of Firefox and Chrome.
	// Bitmap strikes are 1-bit black-and-white images, so properly hinted
	// anti-aliased glyphs are usually better even at very small sizes.
	int ftflags = FT_LOAD_NO_BITMAP | FT_LOAD_FORCE_AUTOHINT;
	m_fontFace = cairo_ft_font_face_create_for_ft_face(freeTypeFace, ftflags);

    buffer->ref(); // This is balanced by the buffer->deref() in releaseCustomFontData.

    // Cairo doesn't do FreeType reference counting, so we need to ensure that when
    // this cairo_font_face_t is destroyed, it cleans up the FreeType face as well.
    static cairo_user_data_key_t freeTypeFaceKey;
    cairo_font_face_set_user_data(m_fontFace, &freeTypeFaceKey, freeTypeFace,
         reinterpret_cast<cairo_destroy_func_t>(FT_Done_Face));

    static cairo_user_data_key_t bufferKey;
    cairo_font_face_set_user_data(m_fontFace, &bufferKey, buffer,
         static_cast<cairo_destroy_func_t>(releaseCustomFontData));

    ++gFontCustomPlatformDataCount;
}

FontCustomPlatformData::~FontCustomPlatformData()
{
	m_cache.clear();

    // m_freeTypeFace will be destroyed along with m_fontFace. See the constructor.
    cairo_font_face_destroy(m_fontFace);

    --gFontCustomPlatformDataCount;
}

PassRefPtr<SimpleFontData> FontCustomPlatformData::fontPlatformData(const FontDescription& fontDescription, bool syntheticBold, bool syntheticItalic, FontOrientation orientation)
{
	int size = fontDescription.computedPixelSize();

	double now = WTF::monotonicallyIncreasingTime();

	for(int i = 0; i < m_cache.size(); ++i) {
		const FontPlatformData& platformData = m_cache[i].first->platformData();

		if(platformData.size() == size && platformData.syntheticBold() == syntheticBold && platformData.syntheticOblique() == syntheticItalic) {
			m_cache[i].second = now;

			return m_cache[i].first;
		}
	}

	RefPtr<SimpleFontData> simpleFontData = SimpleFontData::create(FontPlatformData(m_fontFace, size, syntheticBold, syntheticItalic, orientation), true, false);

	m_cache.append(std::pair<RefPtr<SimpleFontData>, double>(simpleFontData, now));

	if(double interval = memoryCache().simpleFontDataDeletionInterval())
		if(!m_simpleFontDataPruneTimer.isActive())
			m_simpleFontDataPruneTimer.startOneShot(0.25); // every 250ms

	return simpleFontData.release();
}

void FontCustomPlatformData::simpleFontDataPruneTimerCallback(Timer<FontCustomPlatformData>*)
{
	double interval = memoryCache().simpleFontDataDeletionInterval();

	if(!interval)
		return;

	double now = WTF::monotonicallyIncreasingTime();

	for(int i = m_cache.size() - 1; i >= 0; --i) {
		if(m_cache[i].first->refCount() > 1)
			continue;

		if((now - m_cache[i].second) <= interval)
			continue;

		m_cache.remove(i);
	}

	if(m_cache.size() > 0)
		if(!m_simpleFontDataPruneTimer.isActive())
			m_simpleFontDataPruneTimer.startOneShot(0.25); // every 250ms
}

FontCustomPlatformData* createFontCustomPlatformData(SharedBuffer* buffer)
{
    ASSERT_ARG(buffer, buffer);

    RefPtr<SharedBuffer> sfntBuffer;
    if (isWOFF(buffer)) {
        Vector<char> sfnt;
        if (!convertWOFFToSfnt(buffer, sfnt))
            return 0;

        sfntBuffer = SharedBuffer::adoptVector(sfnt);
        buffer = sfntBuffer.get();
    }
    else if (isWOFF2(buffer)) {
		Vector<char> sfnt;
		if (!convertWOFF2ToSfnt(buffer, sfnt))
			return 0;

		sfntBuffer = SharedBuffer::adoptVector(sfnt);
		buffer = sfntBuffer.get();
	}

    static FT_Library library = 0;
    if (!library && FT_Init_FreeType(&library)) {
        library = 0;
        return 0;
    }

    FT_Face freeTypeFace;
    if (FT_New_Memory_Face(library, reinterpret_cast<const FT_Byte*>(buffer->data()), buffer->size(), 0, &freeTypeFace))
        return 0;
    return new FontCustomPlatformData(freeTypeFace, buffer);
}

bool FontCustomPlatformData::supportsFormat(const String& format)
{
    return
    	equalIgnoringCase(format, ASCIILiteral("truetype")) ||
		equalIgnoringCase(format, ASCIILiteral("opentype")) ||
		equalIgnoringCase(format, ASCIILiteral("woff")) ||
		equalIgnoringCase(format, ASCIILiteral("woff2"));
}

}
