
#include "config.h"
#include "FontBaseline.h"
#include <wtf/text/AtomicString.h>
#include "FontPlatformData.h"
#include "PlatformString.h"
#include "Font.h"
#include "FontCache.h"
#include "FontDescription.h"
#include "SimpleFontData.h"
#include "TextEncoding.h"
#include "UTF16UChar32Iterator.h"
#include <wtf/HashSet.h>
#include <wtf/unicode/UTF8.h>
//#include <freetype/tttables.h>

#ifndef CAIRO_HAS_GLESV2_SURFACE
#define CAIRO_HAS_GLESV2_SURFACE 0
#endif

namespace WebCore {

static bool sFontAntialiased = false;

DLL_PUBLIC void setFontAntialiased(bool antialias)
{
	sFontAntialiased = antialias;
}

static HashSet<FontPlatformData*>& fontPlatformDataCache()
{
	static HashSet<FontPlatformData*> cache;
	return cache;
}

static cairo_font_options_t* getDefaultFontOptions()
{
#if 0 && PLATFORM(GTK)
    if (GdkScreen* screen = gdk_screen_get_default()) {
        const cairo_font_options_t* screenOptions = gdk_screen_get_font_options(screen);
        if (screenOptions)
            return cairo_font_options_copy(screenOptions);
    }
#endif
    return cairo_font_options_create();
}

static void rotateCairoMatrixForVerticalOrientation(cairo_matrix_t* matrix)
{
    // The resulting transformation matrix for vertical glyphs (V) is a
    // combination of rotation (R) and translation (T) applied on the
    // horizontal matrix (H). V = H . R . T, where R rotates by -90 degrees
    // and T translates by font size towards y axis.
    cairo_matrix_rotate(matrix, -M_PI_2);
    cairo_matrix_translate(matrix, 0.0, 1.0);
}

DLL_PUBLIC FontPlatformData::FontPlatformData(const FontDescription& fontDescription, FcPattern* fcPattern, const String& familyName)
: m_pattern(fcPattern)
, m_familyName(familyName)
, m_size(fontDescription.computedPixelSize())
, m_syntheticBold(false)
, m_syntheticOblique(false)
, m_spaceWidth(0)
, m_cachePruneTimer(this, &FontPlatformData::cachePruneTimerCallback)
, m_orientation(fontDescription.orientation())
{
	if(!m_pattern)
		return;

    // Disable bitmap strikes in all local fonts if outlines are available.
    // Bitmap strikes are 1-bit black-and-white images, so properly hinted
    // anti-aliased glyphs are usually better even at very small sizes.
    FcPatternDel(m_pattern.get(), FC_EMBEDDED_BITMAP);
    FcPatternAddBool(m_pattern.get(), FC_EMBEDDED_BITMAP, false);

    RefPtr<cairo_font_face_t> fontFace = adoptRef(cairo_ft_font_face_create_for_pattern(m_pattern.get()));
    initializeWithFontFace(fontFace.get(), fontDescription);
    
    if (fontDescription.weight() >= FontWeightBold) {
        // The FC_EMBOLDEN property instructs us to fake the boldness of the font.
        FcBool fontConfigEmbolden;
        if (FcPatternGetBool(m_pattern.get(), FC_EMBOLDEN, 0, &fontConfigEmbolden) == FcResultMatch)
            m_syntheticBold = fontConfigEmbolden;
    }
}

FontPlatformData::FontPlatformData(float size, bool bold, bool italic)
    : m_size(size)
    , m_syntheticBold(bold)
    , m_syntheticOblique(italic)
    , m_spaceWidth(0)
, m_cachePruneTimer(this, &FontPlatformData::cachePruneTimerCallback)
, m_orientation(Horizontal)
{
}

FontPlatformData::FontPlatformData(cairo_font_face_t* fontFace, int size, bool bold, bool italic, FontOrientation orientation)
    : m_size(size)
    , m_syntheticBold(bold)
    , m_syntheticOblique(italic)
    , m_spaceWidth(0)
, m_cachePruneTimer(this, &FontPlatformData::cachePruneTimerCallback)
, m_orientation(orientation)
{
	if(fontFace)
		initializeWithFontFace(fontFace, FontDescription());
}

FontPlatformData::FontPlatformData(const FontPlatformData& other)
: m_cachePruneTimer(this, &FontPlatformData::cachePruneTimerCallback)
{
    *this = other;
}

FontPlatformData& FontPlatformData::operator=(const FontPlatformData& other)
{
    // Check for self-assignment.
    if (this == &other)
        return *this;

    clearCachedScaledFonts();

    if(m_cachePruneTimer.isActive())
    	m_cachePruneTimer.stop();

    m_size = other.m_size;
    m_syntheticBold = other.m_syntheticBold;
    m_syntheticOblique = other.m_syntheticOblique;
    m_spaceWidth = other.m_spaceWidth;
    m_charReplaceMap = other.m_charReplaceMap;
    m_familyName = other.m_familyName;
    m_scaledFont = other.m_scaledFont;
    m_pattern = other.m_pattern;
    m_orientation = other.m_orientation;
    m_horizontalOrientationMatrix = other.m_horizontalOrientationMatrix;

	// This will be re-created on demand.
	m_fallbacks.clear();

    return *this;
}

DLL_PUBLIC FontPlatformData::~FontPlatformData()
{
	fontPlatformDataCache().remove(this);

    clearCachedScaledFonts();
}

FontPlatformData FontPlatformData::smallCapsFontData(const FontDescription& fontDescription) const
{
	return FontPlatformData(fontDescription, m_pattern.get());
}

bool FontPlatformData::isFixedPitch()
{
    // TODO: Support isFixedPitch() for custom fonts.
    if (m_pattern){
		int spacing;
		if (FcPatternGetInteger(m_pattern.get(), FC_SPACING, 0, &spacing) == FcResultMatch)
			return spacing == FC_MONO;
    }
    
    return false;
}

bool FontPlatformData::containsCharacters(const UChar* characters, int bufferLength) const
{
	if(!m_scaledFont)
		return false;

	FT_Face face = cairo_ft_scaled_font_lock_face(m_scaledFont.get());
	if(!face)
		return false;

	UTF16UChar32Iterator iterator(characters, bufferLength);
	UChar32 character = iterator.next();
	while (character != iterator.end()) {
		if (getFreeTypeCharIndex(face, (FT_ULong)character) == 0) {
			cairo_ft_scaled_font_unlock_face(m_scaledFont.get());
			return false;
		}
		character = iterator.next();
	}

	cairo_ft_scaled_font_unlock_face(m_scaledFont.get());
	return true;
}

DLL_PUBLIC int FontPlatformData::totalCachedScaledFonts()
{
	int total = 0;
	HashSet<FontPlatformData*>& cache = fontPlatformDataCache();

	for(HashSet<FontPlatformData*>::iterator it = cache.begin(); it != cache.end(); ++it) {
		FontPlatformData* data = *it;
		total += data->m_cachedScaledFonts.size();
	}

	return total;
}

void FontPlatformData::cacheScaledFont(cairo_scaled_font_t* scaledFont)
{
    if(!m_cachePruneTimer.isActive())
    	m_cachePruneTimer.startOneShot(0.5);

	for(unsigned int i = 0; i < m_cachedScaledFonts.size(); ++i) {
		if(m_cachedScaledFonts[i].first == scaledFont) {
			m_cachedScaledFonts[i].second = WTF::monotonicallyIncreasingTime();
			return;
		}
	}

	HashSet<FontPlatformData*>& cache = fontPlatformDataCache();

	FontPlatformData* candidateData = 0;
	double candidateTime = 0;
	int candidateIndex = 0;
	int totalCachedScaledFonts = 0;

	for(HashSet<FontPlatformData*>::iterator it = cache.begin(); it != cache.end(); ++it) {
		FontPlatformData* data = *it;
		totalCachedScaledFonts += data->m_cachedScaledFonts.size();

		for(int j = 0; j < data->m_cachedScaledFonts.size(); ++j) {
			if(!candidateData || (candidateTime > data->m_cachedScaledFonts[j].second)) {
				candidateData = data;
				candidateTime = data->m_cachedScaledFonts[j].second;
				candidateIndex = j;
			}
		}
	}

	if(candidateData && totalCachedScaledFonts >= 128) {
		cairo_scaled_font_destroy(candidateData->m_cachedScaledFonts[candidateIndex].first);
		candidateData->m_cachedScaledFonts.remove(candidateIndex);
	}

	m_cachedScaledFonts.append(std::pair<cairo_scaled_font_t*, double>(cairo_scaled_font_reference(scaledFont), WTF::monotonicallyIncreasingTime()));

	if(!cache.contains(this))
		cache.add(this);
}

void FontPlatformData::cachePruneTimerCallback(Timer<FontPlatformData>*)
{
	double now = WTF::monotonicallyIncreasingTime();

	for(int i = m_cachedScaledFonts.size() - 1; i >= 0; --i) {
		if((now - m_cachedScaledFonts[i].second) > 30) { // not used in 30 seconds
			cairo_scaled_font_destroy(m_cachedScaledFonts[i].first);
			m_cachedScaledFonts.remove(i);
		}
	}

	if(m_cachedScaledFonts.size() > 0)
		m_cachePruneTimer.startOneShot(0.5);
	else
		fontPlatformDataCache().remove(this);
}

void FontPlatformData::clearCachedScaledFonts()
{
    if(m_cachedScaledFonts.size() == 0)
    	return;

	for(unsigned int i = 0; i < m_cachedScaledFonts.size(); ++i)
		cairo_scaled_font_destroy(m_cachedScaledFonts[i].first);

	m_cachedScaledFonts.clear();

	fontPlatformDataCache().remove(this);
}

void FontPlatformData::initializeWithFontFace(cairo_font_face_t* fontFace, const FontDescription& fontDescription)
{
	if(m_size <= 0)
		return;

	cairo_font_options_t* font_options = getDefaultFontOptions();

	cairo_matrix_t ctm;
	cairo_matrix_init_identity(&ctm);

	cairo_matrix_t fontMatrix;
	cairo_matrix_init_scale(&fontMatrix, m_size, m_size);

	// for BML
	//cairo_font_options_set_hint_style(font_options, CAIRO_HINT_STYLE_NONE);
	//cairo_font_options_set_hint_metrics(font_options, CAIRO_HINT_METRICS_OFF);

#if CAIRO_HAS_GLESV2_SURFACE
	if(sFontAntialiased)
		cairo_font_options_set_antialias(font_options, CAIRO_ANTIALIAS_GOOD);
#endif

	if(m_pattern) {
        // We requested an italic font, but Fontconfig gave us one that was neither oblique nor italic.
        int actualFontSlant;
        if (fontDescription.italic() && FcPatternGetInteger(m_pattern.get(), FC_SLANT, 0, &actualFontSlant) == FcResultMatch)
            m_syntheticOblique = actualFontSlant == FC_SLANT_ROMAN;
	}

    if (m_syntheticOblique) {
        static const float syntheticObliqueSkew = -tanf(14 * acosf(0) / 90);
        static const cairo_matrix_t skew = {1, 0, syntheticObliqueSkew, 1, 0, 0};
		static const cairo_matrix_t verticalSkew = {1, -syntheticObliqueSkew, 0, 1, 0, 0};
		cairo_matrix_multiply(&fontMatrix, m_orientation == Vertical ? &verticalSkew : &skew, &fontMatrix);
	}

    m_horizontalOrientationMatrix = fontMatrix;
    if (m_orientation == Vertical)
        rotateCairoMatrixForVerticalOrientation(&fontMatrix);

    m_scaledFont = adoptRef(cairo_scaled_font_create(fontFace, &fontMatrix, &ctm, font_options));

	cairo_font_options_destroy(font_options);
}

bool FontPlatformData::init()
{
    static bool initialized = false;
    if (initialized)
        return true;

    initialized = true;

    if (!FcInit())
        return false;

    return true;
}

float FontPlatformData::platformWidthForGlyph(Glyph glyph) const
{
	if(!m_scaledFont)
		return m_spaceWidth;

    if (cairo_scaled_font_status(scaledFont()) != CAIRO_STATUS_SUCCESS)
        return m_spaceWidth;

	cairo_glyph_t cglyph = { glyph, 0, 0 };
	cairo_text_extents_t extents;
	cairo_scaled_font_glyph_extents(m_scaledFont.get(), &cglyph, 1, &extents);

	float width = orientation() == Horizontal ? extents.x_advance : -extents.y_advance;
	return width ? width : m_spaceWidth;
}

static void addCharactersToFontConfigPattern(const UChar* characters, int length, FcPattern* pattern)
{
	FcCharSet* fontConfigCharSet = FcCharSetCreate();
	for (int i = 0; i < length; ++i) {
		if (U16_IS_SURROGATE(characters[i]) && U16_IS_SURROGATE_LEAD(characters[i])
				&& i != length - 1 && U16_IS_TRAIL(characters[i + 1])) {
			FcCharSetAddChar(fontConfigCharSet, U16_GET_SUPPLEMENTARY(characters[i], characters[i+1]));
			i++;
		} else
			FcCharSetAddChar(fontConfigCharSet, characters[i]);
	}
	FcPatternAddCharSet(pattern, FC_CHARSET, fontConfigCharSet);
	FcCharSetDestroy(fontConfigCharSet);
}

static RefPtr<FcPattern> findBestFontGivenFallbacks(const FontPlatformData& fontData, FcPattern* pattern, FcResult& fontConfigResult)
{
    if (!fontData.fcPattern())
        return 0;

    if (!fontData.m_fallbacks) {
        FcResult fontConfigResult;
        fontData.m_fallbacks = adoptPtr(FcFontSort(0, fontData.fcPattern(), FcTrue, 0, &fontConfigResult));
    }

    if (!fontData.m_fallbacks)
        return 0;

    FcFontSet* sets[] = { fontData.m_fallbacks.get() };
    return adoptRef(FcFontSetMatch(0, sets, 1, pattern, &fontConfigResult));
}

PassRefPtr<SimpleFontData> FontCache::getFontDataForCharacters(const Font& font, const UChar* characters, int length)
{
	if(m_perCharactersPattern) {
		addCharactersToFontConfigPattern(characters, length, m_perCharactersPattern.get());
	}
	else {
		m_perCharactersPattern = adoptRef(FcPatternCreate());

		addCharactersToFontConfigPattern(characters, length, m_perCharactersPattern.get());
	    FcPatternAddBool(m_perCharactersPattern.get(), FC_SCALABLE, FcTrue);

	    // the following two functions are very slow
	    FcConfigSubstitute(0, m_perCharactersPattern.get(), FcMatchPattern);
	    FcDefaultSubstitute(m_perCharactersPattern.get());
	}

	FcResult fontConfigResult;
    if(RefPtr<FcPattern> fallbackPattern = findBestFontGivenFallbacks(font.primaryFont()->platformData(), m_perCharactersPattern.get(), fontConfigResult)) {
    	FcChar8* fontConfigFamilyNameAfterMatching;
		FcPatternGetString(fallbackPattern.get(), FC_FAMILY, 0, &fontConfigFamilyNameAfterMatching);
		String familyNameAfterMatching = String::fromUTF8(reinterpret_cast<char*>(fontConfigFamilyNameAfterMatching));
		if(RefPtr<SimpleFontData> simpleFontData = getCachedFontData(font.fontDescription(), familyNameAfterMatching)) {
		    FcPatternDel(m_perCharactersPattern.get(), FC_CHARSET);
			return simpleFontData.release();
		}
    }

	if(RefPtr<FcPattern> resultPattern = adoptRef(FcFontMatch(0, m_perCharactersPattern.get(), &fontConfigResult))) {
		FcChar8* fontConfigFamilyNameAfterMatching;
		FcPatternGetString(resultPattern.get(), FC_FAMILY, 0, &fontConfigFamilyNameAfterMatching);
		String familyNameAfterMatching = String::fromUTF8(reinterpret_cast<char*>(fontConfigFamilyNameAfterMatching));
		if(RefPtr<SimpleFontData> simpleFontData = getCachedFontData(font.fontDescription(), familyNameAfterMatching)) {
		    FcPatternDel(m_perCharactersPattern.get(), FC_CHARSET);
			return simpleFontData.release();
		}
	}

    FcPatternDel(m_perCharactersPattern.get(), FC_CHARSET);
    return 0;
}

void FontPlatformData::setOrientation(FontOrientation orientation)
{
    ASSERT(m_scaledFont);

    if (!m_scaledFont || (m_orientation == orientation))
        return;

    cairo_matrix_t transformationMatrix;
    cairo_matrix_init_identity(&transformationMatrix);

    cairo_matrix_t fontMatrix;
    cairo_scaled_font_get_font_matrix(scaledFont(), &fontMatrix);

    cairo_font_options_t* options = getDefaultFontOptions();

    // In case of vertical orientation, rotate the transformation matrix.
    // Otherwise restore the horizontal orientation matrix.
    if (orientation == Vertical)
        rotateCairoMatrixForVerticalOrientation(&fontMatrix);
    else
        fontMatrix = m_horizontalOrientationMatrix;

    cairo_font_face_t* fontFace = cairo_scaled_font_get_font_face(scaledFont());
    m_scaledFont = adoptRef(cairo_scaled_font_create(fontFace, &fontMatrix, &transformationMatrix, options));
    cairo_font_options_destroy(options);
    m_orientation = orientation;
}

}
