///// Yunfeng: This is the 2rd version of SocketStreamHandle for RFC6455

#include "config.h"
#include "SocketStreamHandle.h"

#include "KURL.h"
#include "Logging.h"
#include "NotImplemented.h"
#include "SocketStreamHandleClient.h"

#include "SocketStreamError.h"
#include "SocketStreamErrorBase.h"
#include <wtf/HexNumber.h>
#include <wtf/MainThread.h>
#include <wtf/PassOwnPtr.h>
#include <wtf/RefCounted.h>
#include <wtf/text/StringBuilder.h>
#include <wtf/text/WTFString.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <unistd.h>
#include <fcntl.h>

namespace WebCore {

DLL_PUBLIC int gSocketStreamHandleCount;

#define sPollInterval 0.025

static const int ReadBufSize = 64 * 1024;

void SocketStreamHandle::ReadtimerEvent(Timer<SocketStreamHandle>* timer)
{
	RefPtr<SocketStreamHandle> protector(this);

    if (m_state != Open)
        return;
    if (m_socket == -1)
        return;

    fd_set read_flags; // the flag sets to be used

    struct timeval waitd;          // the max wait time for an event
    waitd.tv_sec = 0;
    waitd.tv_usec = 0;

    FD_ZERO(&read_flags);
    FD_SET(m_socket, &read_flags);

    int stat = select(m_socket+1, &read_flags, 0, 0, &waitd);

    if(stat < 0)
    {
        m_client->didFailSocketStream(this, SocketStreamError());
        return;
    }

    if (stat == 0)
    {
        timer->startOneShot(sPollInterval);
        return;
    }

    if (FD_ISSET(m_socket, &read_flags))
    {
        if(!m_readBuffer)
        	m_readBuffer = static_cast<char*>(fastMalloc(ReadBufSize * sizeof(char)));

        ssize_t sz;
        if (m_isSecure)
            sz = SSL_read(m_sslhandle, m_readBuffer, ReadBufSize);
        else
            sz = ::read(m_socket, m_readBuffer, ReadBufSize);

        if(sz == 0)
            return;

        if (sz < 0)
        {
            m_client->didFailSocketStream(this, SocketStreamError());
            return;
        }

        m_client->didReceiveSocketStreamData(this, m_readBuffer, sz);
    }

    timer->startOneShot(sPollInterval);
}

static Vector<char> StringToVector(String string)
{
    Vector<char> buffer;
    StringImpl* impl = string.impl();
    if (!impl)
        return buffer;


    unsigned length = impl->length();
    const UChar* characters = impl->characters();

    buffer.resize(length + 1);
    for (unsigned i = 0; i < length; ++i) {
        UChar ch = characters[i];
        buffer[i] = ch && (ch < 0x20 || ch > 0x7f) ? '?' : ch;
    }
    buffer[length] = '\0';

    return buffer;
}

void SocketStreamHandle::OpentimerEvent(Timer<SocketStreamHandle>* timer)
{
	RefPtr<SocketStreamHandle> protector(this);

	if (m_socket < 0) {
		struct addrinfo *ai;
		int aiErr;
		int rc;
		struct addrinfo *aiHead;
		struct addrinfo hints;
		struct sockaddr_in6 *pSadrIn6;

		char host[512];  // assuming host name will not be larger than 512 bytes
		char service[512];

		bzero(host, 512);
		bzero(service, 512);

		m_isSecure = m_url.protocolIs("wss");

		Vector<char> h = StringToVector(m_url.host());

		memcpy(host, h.data(), h.size());
		unsigned int port = m_url.hasPort() ? m_url.port() : (m_isSecure ? 443 : 80);
		sprintf(service, "%d", port);

		memset(&hints, 0, sizeof(hints));
		hints.ai_family = AF_UNSPEC;
		hints.ai_socktype = SOCK_STREAM;
		hints.ai_protocol = IPPROTO_TCP;

		int olen = strlen(host);
		if (host[0] == '[') {
			::memcpy(host, &host[1], olen - 2);
			host[olen - 2] = 0;
			host[olen - 1] = 0;
		}

		if ((aiErr = getaddrinfo(host, service, &hints, &aiHead)) != 0) {
			m_client->didFailSocketStream(this, SocketStreamError());
			return;
		}

		for (ai = aiHead, m_socket = -1; (ai != NULL) && (m_socket == -1); ai = ai->ai_next) {
			if (ai->ai_family == PF_INET6) {
				pSadrIn6 = (struct sockaddr_in6*) ai->ai_addr;
				if (pSadrIn6->sin6_scope_id == 0)
					pSadrIn6->sin6_scope_id = if_nametoindex("eth0");
			}

			if ((m_socket = ::socket(ai->ai_family, ai->ai_socktype, ai->ai_protocol)) < 0) {
				m_socket = -1;
				continue;
			}

			int x = fcntl(m_socket, F_GETFL, 0);
			fcntl(m_socket, F_SETFL, x | O_NONBLOCK);

			rc = ::connect(m_socket, ai->ai_addr, ai->ai_addrlen);
			if ((rc < 0) && (errno != EINPROGRESS)) {
				StringBuilder msg;

				msg.append("SocketStreamHandle::OpentimerEvent connect [");
				msg.append(m_url.string());
				msg.append("] error ");
				msg.appendNumber(rc);
				msg.append(" errno ");
				msg.appendNumber(errno);

				WTF::logMessage(msg.toString());

				::close(m_socket);
				m_socket = -1;
				continue;
			}
			else if (rc == 0) { // connection has succeeded immediately
				StringBuilder msg;
				msg.append("SocketStreamHandle::OpentimerEvent connect [");
				msg.append(m_url.string());
				msg.append("] success immediately");
				WTF::logMessage(msg.toString());
				break;
			}
			else {
				fd_set fdread;
				fd_set fdwrite;
				fd_set fdexcep;
				int maxfd = 0;

				long startTime = WTF::monotonicallyIncreasingTime();

				int selectTimeout = 15;

				struct timeval timeout;
				timeout.tv_sec = selectTimeout;
				timeout.tv_usec = 0;

				FD_ZERO(&fdread);
				FD_ZERO(&fdwrite);
				FD_ZERO(&fdexcep);

			    FD_SET(m_socket, &fdread);
			    FD_SET(m_socket, &fdwrite);
			    FD_SET(m_socket, &fdexcep);

				rc = ::select(m_socket + 1, &fdread, &fdwrite, &fdexcep, &timeout);

				// The normal return value from select is the total number of ready file descriptors in all of the sets.
				if (rc > 0) {
					long endTime = WTF::monotonicallyIncreasingTime();

					StringBuilder msg;
					msg.append("SocketStreamHandle::OpentimerEvent connect [");
					msg.append(m_url.string());
					msg.append("] ready after ");
					msg.appendNumber((int)(endTime - startTime));
					msg.append(" seconds");
					WTF::logMessage(msg.toString());

					break;
				} else {
					StringBuilder msg;
					msg.append("SocketStreamHandle::OpentimerEvent connect [");
					msg.append(m_url.string());
					msg.append("] timeout after ");
					msg.appendNumber(selectTimeout);
					msg.append(" seconds, remain ");
					msg.appendNumber(timeout.tv_sec);
					msg.append(" seconds");
					WTF::logMessage(msg.toString());

					::close(m_socket);
					m_socket = -1;
					continue;
				}
			}
		}  // end for

		freeaddrinfo(aiHead);

		if (m_socket == -1) {
			m_client->didFailSocketStream(this, SocketStreamError());
			return;
		}

		// setup secure socket instead.
		if (m_isSecure) {
			int err;

			//SSL_load_error_strings();
			//SSL_library_init();
#if (OPENSSL_VERSION_NUMBER >= 0x10100000L) && !defined(LIBRESSL_VERSION_NUMBER)
			m_sslctx = SSL_CTX_new(TLS_client_method());
#else
			m_sslctx = SSL_CTX_new(SSLv23_client_method());
#endif
			if (!m_sslctx) {
				::close(m_socket);
				m_socket = -1;
				m_client->didFailSocketStream(this, SocketStreamError());
				return;
			}
			m_sslhandle = SSL_new(m_sslctx);
			if (!m_sslhandle) {
				::close(m_socket);
				m_socket = -1;
				m_client->didFailSocketStream(this, SocketStreamError());
				return;
			}
			if (!SSL_set_fd(m_sslhandle, m_socket)) {
				::close(m_socket);
				m_socket = -1;
				m_client->didFailSocketStream(this, SocketStreamError());
				return;
			}

			err = SSL_connect(m_sslhandle);
			 /* 1  is fine
			    0  is "not successful but was shut down controlled"
			    <0 is "handshake was not successful, because a fatal error occurred" */
			 if(1 != err) {
				int detail = SSL_get_error(m_sslhandle, err);

			    if(SSL_ERROR_WANT_READ == detail) {
			    	//
			    }
			    else if(SSL_ERROR_WANT_WRITE == detail) {
			    	//
			    }
			    else {
			    	/* untreated error */
					StringBuilder msg;
					msg.append("SocketStreamHandle::OpentimerEvent connect [");
					msg.append(m_url.string());
					msg.append("] SSL error ");
					msg.appendNumber(detail);
					WTF::logMessage(msg.toString());

					::close(m_socket);
					m_socket = -1;
					m_client->didFailSocketStream(this, SocketStreamError());
					return;
			    }
			}
		}
	}

	m_state = SocketStreamHandleBase::Open;
	m_client->didOpenSocketStream(this);

	m_rdtimer.startOneShot(sPollInterval);
}

SocketStreamHandle::SocketStreamHandle(const KURL &url, SocketStreamHandleClient *cli)
    :SocketStreamHandleBase(url, cli)
    , m_socket(-1)
    , m_rdtimer(this, &SocketStreamHandle::ReadtimerEvent)
    , m_contimer(this, &SocketStreamHandle::OpentimerEvent)
    , m_isSecure(false)
    , m_sslhandle(NULL)
    , m_sslctx(NULL)
, m_readBuffer(0)
{
    m_contimer.startOneShot(sPollInterval);

    ++gSocketStreamHandleCount;
}

DLL_PUBLIC SocketStreamHandle::SocketStreamHandle(int socketfd, SocketStreamHandleClient *cli)
    :SocketStreamHandleBase(KURL(), cli)
    , m_socket(socketfd)
    , m_rdtimer(this, &SocketStreamHandle::ReadtimerEvent)
    , m_contimer(this, &SocketStreamHandle::OpentimerEvent)
    , m_isSecure(false)
    , m_sslhandle(NULL)
    , m_sslctx(NULL)
, m_readBuffer(0)
{
	m_state = Open;
    m_contimer.startOneShot(sPollInterval);

    ++gSocketStreamHandleCount;
}

SocketStreamHandle::~SocketStreamHandle()
{
    if (m_sslhandle)
    {
        SSL_shutdown(m_sslhandle);
        SSL_free(m_sslhandle);
        m_sslhandle = NULL;
    }

    if (m_sslctx)
    {
        SSL_CTX_free(m_sslctx);
        m_sslctx = NULL;
    }

    if(m_socket >= 0){
    	::close(m_socket);
    	m_socket = -1;
    }

    if(m_readBuffer){
    	fastFree(m_readBuffer);
    	m_readBuffer = 0;
    }

    --gSocketStreamHandleCount;
}

int SocketStreamHandle::platformSend(const char* data, int size)
{
    if (m_state == Closed || m_socket == -1)
    	return size;

    if (m_isSecure && m_sslhandle && m_sslctx) {
    	int rc = SSL_write(m_sslhandle, data, size);

    	if(rc <= 0) {
			int err = SSL_get_error(m_sslhandle, rc);

			switch(err) {
			case SSL_ERROR_WANT_READ:
			case SSL_ERROR_WANT_WRITE:
			  /* The operation did not complete; the same TLS/SSL I/O function
				 should be called again later. This is basically an EWOULDBLOCK
				 equivalent. */
			  return 0;

			default:
				StringBuilder msg;
				msg.append("SocketStreamHandle::platformSend connect [");
				msg.append(m_url.string());
				msg.append("] SSL error ");
				msg.appendNumber(err);
				WTF::logMessage(msg.toString());

				return -1;
			}
    	}

    	return rc;
    }
    else
        return ::write(m_socket, data, size);
}

void SocketStreamHandle::platformClose()
{
    m_rdtimer.stop();
    m_contimer.stop();

    if (m_socket == -1) {
        m_client->didCloseSocketStream(this);
        return;
    }

    if (m_isSecure && m_sslhandle && m_sslctx)
    {
        SSL_shutdown(m_sslhandle);
        SSL_free(m_sslhandle);
        m_sslhandle = NULL;
        SSL_CTX_free(m_sslctx);
        m_sslctx = NULL;
    }

    ::close(m_socket);
    m_socket = -1;

    m_client->didCloseSocketStream(this);
}

void SocketStreamHandle::didReceiveAuthenticationChallenge(const AuthenticationChallenge&)
{
}

void SocketStreamHandle::receivedCredential(const AuthenticationChallenge&, const Credential&)
{
}

void SocketStreamHandle::receivedRequestToContinueWithoutCredential(const AuthenticationChallenge&)
{
}

void SocketStreamHandle::receivedCancellation(const AuthenticationChallenge&)
{
}


}
