/*
 * Copyright (C) 1999 Lars Knoll (knoll@kde.org)
 *           (C) 1999 Antti Koivisto (koivisto@kde.org)
 *           (C) 2005 Allan Sandfeld Jensen (kde@carewolf.com)
 *           (C) 2005, 2006 Samuel Weinig (sam.weinig@gmail.com)
 * Copyright (C) 2005, 2006, 2007, 2008, 2009, 2013 Apple Inc. All rights reserved.
 * Copyright (C) 2010, 2012 Google Inc. All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */

#include "config.h"
#include "RenderElement.h"

#include "ContentData.h"
#include "RenderDeprecatedFlexibleBox.h"
#include "RenderFlexibleBox.h"
#include "RenderGrid.h"
#include "RenderImage.h"
#include "RenderImageResourceStyleImage.h"
#include "RenderListItem.h"
#include "RenderMultiColumnBlock.h"
#include "RenderRegion.h"
#include "RenderRuby.h"
#include "RenderRubyText.h"
#include "RenderTableCaption.h"
#include "RenderTableCell.h"
#include "RenderTableCol.h"
#include "RenderTableRow.h"

namespace WebCore {

RenderElement::RenderElement(Element* element)
    : RenderObject(element)
{
}

RenderElement::~RenderElement()
{
}

RenderElement* RenderElement::createFor(Element& element, RenderStyle& style)
{
    Document* doc = element.document();

    // Minimal support for content properties replacing an entire element.
    // Works only if we have exactly one piece of content and it's a URL.
    // Otherwise acts as if we didn't support this feature.
    const ContentData* contentData = style.contentData();
    if (contentData && !contentData->next() && contentData->isImage() && !element.isPseudoElement()) {
        RenderImage* image = new (std::nothrow) RenderImage(&element);
        if(!image)
        	return 0;
        // RenderImageResourceStyleImage requires a style being present on the image but we don't want to
		// trigger a style change now as the element is not fully attached. Moving this code to style change
		// doesn't make sense as it should be run once at renderer creation.
		image->setStyleInternal(&style);
        if (const StyleImage* styleImage = static_cast<const ImageContentData*>(contentData)->image()) {
            image->setImageResource(RenderImageResourceStyleImage::create(const_cast<StyleImage*>(styleImage)));
            image->setIsGeneratedContent();
        } else
            image->setImageResource(RenderImageResource::create());
        image->setStyleInternal(0);
        return image;
    }

    if (element.hasTagName(HTMLNames::rubyTag)) {
        if (style.display() == INLINE)
            return new (std::nothrow) RenderRubyAsInline(&element);
        else if (style.display() == BLOCK)
            return new (std::nothrow) RenderRubyAsBlock(&element);
    }
    // treat <rt> as ruby text ONLY if it still has its default treatment of block
    if (element.hasTagName(HTMLNames::rtTag) && style.display() == BLOCK)
        return new (std::nothrow) RenderRubyText(&element);
    if (doc->cssRegionsEnabled() && style.isDisplayRegionType() && !style.regionThread().isEmpty() && doc->renderView())
        return new (std::nothrow) RenderRegion(&element, 0);
    switch (style.display()) {
    case NONE:
        return 0;
    case INLINE:
        return new (std::nothrow) RenderInline(&element);
    case BLOCK:
    case INLINE_BLOCK:
    case RUN_IN:
    case COMPACT:
        if ((!style.hasAutoColumnCount() || !style.hasAutoColumnWidth()) && doc->regionBasedColumnsEnabled())
            return new (std::nothrow) RenderMultiColumnBlock(&element);
        return new (std::nothrow) RenderBlockFlow(&element);
    case LIST_ITEM:
        return new (std::nothrow) RenderListItem(&element);
    case TABLE:
    case INLINE_TABLE:
        return new (std::nothrow) RenderTable(&element);
    case TABLE_ROW_GROUP:
    case TABLE_HEADER_GROUP:
    case TABLE_FOOTER_GROUP:
        return new (std::nothrow) RenderTableSection(&element);
    case TABLE_ROW:
        return new (std::nothrow) RenderTableRow(&element);
    case TABLE_COLUMN_GROUP:
    case TABLE_COLUMN:
        return new (std::nothrow) RenderTableCol(&element);
    case TABLE_CELL:
        return new (std::nothrow) RenderTableCell(&element);
    case TABLE_CAPTION:
        return new (std::nothrow) RenderTableCaption(&element);
    case BOX:
    case INLINE_BOX:
        return new (std::nothrow) RenderDeprecatedFlexibleBox(&element);
#if ENABLE(CSS3_FLEXBOX)
    case FLEX:
    case INLINE_FLEX:
    case WEBKIT_FLEX:
    case WEBKIT_INLINE_FLEX:
        return new (std::nothrow) RenderFlexibleBox(&element);
#endif
    case GRID:
    case INLINE_GRID:
        return new (std::nothrow) RenderGrid(&element);
    }

    return 0;
}

}
